/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log$
 */

package tuwien.auto.calimero.buffer;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import tuwien.auto.calimero.CloseEvent;
import tuwien.auto.calimero.FrameEvent;
import tuwien.auto.calimero.GroupAddress;
import tuwien.auto.calimero.KNXAddress;
import tuwien.auto.calimero.Priority;
import tuwien.auto.calimero.cache.Cache;
import tuwien.auto.calimero.cache.CacheObject;
import tuwien.auto.calimero.cache.LDataCacheObject;
import tuwien.auto.calimero.cache.LDataObjectQueue;
import tuwien.auto.calimero.cache.LFUCache;
import tuwien.auto.calimero.cache.LDataObjectQueue.QueueItem;
import tuwien.auto.calimero.cemi.CEMI;
import tuwien.auto.calimero.cemi.CEMILData;
import tuwien.auto.calimero.datapoint.Datapoint;
import tuwien.auto.calimero.datapoint.DatapointModel;
import tuwien.auto.calimero.exception.KNXIllegalStateException;
import tuwien.auto.calimero.exception.KNXTimeoutException;
import tuwien.auto.calimero.link.KNXLinkClosedException;
import tuwien.auto.calimero.link.KNXNetworkLink;
import tuwien.auto.calimero.link.event.NetworkLinkListener;
import tuwien.auto.calimero.link.medium.KNXMediumSettings;
import tuwien.auto.calimero.log.LogManager;
import tuwien.auto.calimero.log.LogService;
import tuwien.auto.calimero.process.DataUnitBuilder;

/**
 * @author bm
 */
public final class NetworkBuffer
{
	public interface Configuration
	{
		void setCache(Cache c);

		/**
		 * Returns the used cache strategy.
		 * <p>
		 * The returned cache object does not change ownership.
		 * 
		 * @return the cache, or <code>null</code> if no cache in use
		 */
		Cache getCache();

		void activate(boolean activate);

		/**
		 * Returns whether caching is active.
		 * <p>
		 * 
		 * @return <code>true</code> if cache is active
		 */
		boolean isActive();

		KNXNetworkLink getBaseLink();

		KNXNetworkLink getBufferedLink();

		DatapointModel getDatapointModel();

		void setDatapointModel(DatapointModel datapoints);

		void setFilter(NetworkFilter nf, RequestFilter uf);

		NetworkFilter getNetworkFilter();
	}

	public static interface NetworkFilter
	{
		void create(CEMI frame, Configuration c);
	}

	public interface RequestFilter
	{

	}

	public static class StateFilter implements NetworkFilter
	{
		public StateFilter()
		{}

		// criteria for accept:
		// group-address, no model, or model and state-based DP,
		// message is A-Group.write or A-Group.res
		public void create(CEMI frame, Configuration c)
		{
			final CEMILData f = (CEMILData) frame;
			if (!(f.getDestination() instanceof GroupAddress))
				return;
			final DatapointModel m = c.getDatapointModel();
			Datapoint dp = null;
			if (m != null && (dp = m.get((GroupAddress) f.getDestination())) == null)
				return;
			if (dp != null && !dp.isStateBased())
				return;
			final byte[] data = f.getData();
			// filter for A-Group.write and A-Group.res services
			final int low = data[1] & 0xC0;
			if ((data[0] & 0x03) != 0 || (low != 0x40 && low != 0x80))
				return;
			CacheObject co = c.getCache().get(LDataCacheObject.createKey(f));
			CEMILData bufferFrame;
			if (low == 0x40) {
				bufferFrame = f;
			}
			else { 
				data[1] = (byte) ((data[1] & 0x3f) | 0x40);
				bufferFrame =
					new CEMILData(CEMILData.MC_LDATA_IND, f.getSource(), f
						.getDestination(), data, f.getPriority(), f.isRepetition(), f
						.getHopCount());
			}
			if (co != null) {
				((LDataCacheObject) co).setFrame(bufferFrame);
			}
			else {
				co = new LDataCacheObject(bufferFrame);
			}
			c.getCache().put(co);
		}
	}

	public static class QueueFilter implements NetworkFilter, RequestFilter
	{
		public interface QueueListener extends LDataObjectQueue.QueueListener
		{
			// void fullQueue(CEMIObjectQueue queue);
		}

		private final class QueueListenerImpl implements QueueListener
		{
			QueueListenerImpl()
			{}

			public void queueFilled(LDataObjectQueue queue)
			{
				if (userQListener != null)
					try {
						userQListener.queueFilled(queue);
					}
					catch (final RuntimeException e) {
						logger.error("L-Data queue listener unexpected behavior", e);
					}
			}
		}

		// stores LDataObjectQueues objects
		private final List indicationKeys = new LinkedList();
		private final QueueListener ql = new QueueListenerImpl();
		private QueueListener userQListener;

		/**
		 * 
		 */
		public QueueFilter()
		{}

		public void setQueueListener(QueueListener l)
		{
			userQListener = l;
		}

		public boolean hasNewIndications()
		{
			return !indicationKeys.isEmpty();
		}

		// might return a queue item with nothing in it, if queue was emptied in between
		public QueueItem getNextIndication()
		{
			if (indicationKeys.isEmpty())
				throw new KNXIllegalStateException("no indications");
			return ((LDataObjectQueue) indicationKeys.remove(0)).getItem();
		}

		public void create(CEMI frame, Configuration c)
		{
			final CEMILData f = (CEMILData) frame;
			if (!(f.getDestination() instanceof GroupAddress))
				return;
			final GroupAddress dst = (GroupAddress) f.getDestination();
			// check if we have a datapoint model, whether it contains the address,
			// and datapoint is command based
			final DatapointModel m = c.getDatapointModel();
			Datapoint dp = null;
			if (m != null && (dp = m.get(dst)) == null || dp != null && dp.isStateBased())
				return;
			final byte[] data = f.getData();
			// filter for A-Group.write and A-Group.res services
			final int bits = data[1] & 0xC0;
			if ((data[0] & 0x03) != 0 || (bits != 0x40 && bits != 0x80))
				return;
			CacheObject co = c.getCache().get(LDataCacheObject.createKey(f));
			if (co == null)
				co = new LDataObjectQueue(dst, true, 10, false, ql);
			c.getCache().put(co);
			indicationKeys.add(co);
			// this might invoke queue listener, so do it after put and add
			((LDataObjectQueue) co).setFrame(f);
		}
	}

	// this is for network link only, for now
	private static final class ConfigImpl implements Configuration
	{
		private final SquirrelLink lnk;
		private final NetworkLinkListener ll;
		private Cache cache;
		private NetworkFilter nwFilter;
		private RequestFilter qFilter;
		private volatile boolean active;
		private DatapointModel model;

		private final class SquirrelListener implements NetworkLinkListener
		{
			SquirrelListener()
			{}

			public void confirmation(FrameEvent e)
			{
				if (!active)
					return;
				final CEMI frame = e.getFrame();
				if (((CEMILData) frame).isPositiveConfirmation()) {
					updateBuffer(frame);
				}
			}

			public void indication(FrameEvent e)
			{
				if (!active)
					return;
				final CEMI frame = e.getFrame();
				updateBuffer(frame);
			}

			public void linkClosed(CloseEvent e)
			{
				activate(false);
			}

			private void updateBuffer(CEMI frame)
			{
				if (nwFilter != null)
					nwFilter.create(frame, ConfigImpl.this);
			}
		}

		private final class SquirrelLink implements KNXNetworkLink
		{
			private final KNXNetworkLink base;
			private final List listeners = new Vector();

			public SquirrelLink(KNXNetworkLink baseLink)
			{
				base = baseLink;
			}

			public void addLinkListener(NetworkLinkListener l)
			{
				base.addLinkListener(l);
				listeners.add(l);
			}

			public void close()
			{
				activate(false);
				base.close();
			}

			public byte getHopCount()
			{
				return base.getHopCount();
			}

			public KNXMediumSettings getKNXMedium()
			{
				return base.getKNXMedium();
			}

			public void removeLinkListener(NetworkLinkListener l)
			{
				base.removeLinkListener(l);
				listeners.remove(l);
			}

			public void sendRequest(KNXAddress dst, Priority p, byte[] nsdu)
				throws KNXLinkClosedException, KNXTimeoutException
			{
				if (!queryResponse(dst, nsdu))
					base.sendRequest(dst, p, nsdu);
			}

			public void sendRequestWait(KNXAddress dst, Priority p, byte[] nsdu)
				throws KNXTimeoutException, KNXLinkClosedException
			{
				if (!queryResponse(dst, nsdu))
					base.sendRequestWait(dst, p, nsdu);
			}

			public void setHopCount(int count)
			{
				base.setHopCount(count);
			}

			public void setKNXMedium(KNXMediumSettings a)
			{
				base.setKNXMedium(a);
			}

			public String getName()
			{
				return null;
			}

			public void sendRequest(CEMILData msg) throws KNXLinkClosedException,
				KNXTimeoutException
			{
				if (!queryResponse(msg.getDestination(), msg.getTPDU()))
					base.sendRequest(msg);
			}

			public void sendRequestWait(CEMILData msg) throws KNXTimeoutException,
				KNXLinkClosedException
			{
				if (!queryResponse(msg.getDestination(), msg.getTPDU()))
					base.sendRequestWait(msg);
			}

			public boolean isOpen()
			{
				return base.isOpen();
			}

			private boolean queryResponse(KNXAddress dst, byte[] nsdu)
			{
				if (dst instanceof GroupAddress) {
					// A-group.read
					if (DataUnitBuilder.getService(nsdu) == 0) {
						final LDataCacheObject co =
							(LDataCacheObject) cache.get(LDataCacheObject.createKey(dst));
						if (co != null) {
							fireIndication(co.getFrame());
							return true;
						}
					}
				}
				return false;
			}

			private void fireIndication(CEMILData frame)
			{
				final FrameEvent e = new FrameEvent(this, frame);
				synchronized (listeners) {
					for (final Iterator i = listeners.iterator(); i.hasNext();) {
						final NetworkLinkListener l = (NetworkLinkListener) i.next();
						try {
							l.indication(e);
						}
						catch (final RuntimeException rte) {
							removeLinkListener(l);
						}
					}
				}
			}
		}

		public ConfigImpl(KNXNetworkLink link)
		{
			lnk = new SquirrelLink(link);
			ll = new SquirrelListener();
			link.addLinkListener(ll);
		}

		public void activate(boolean activate)
		{
			active = activate;
			if (active && cache == null)
				setCache(new LFUCache(0, 0));
		}

		public KNXNetworkLink getBaseLink()
		{
			return lnk.base;
		}

		public KNXNetworkLink getBufferedLink()
		{
			return lnk;
		}

		public Cache getCache()
		{
			return cache;
		}

		public boolean isActive()
		{
			return active;
		}

		public void setCache(Cache c)
		{
			if (c == null && active)
				activate(false);
			cache = c;
		}

		void unregister()
		{
			getBaseLink().removeLinkListener(ll);
		}

		public DatapointModel getDatapointModel()
		{
			return model;
		}

		public void setDatapointModel(DatapointModel datapoints)
		{
			model = datapoints;
		}

		public void setFilter(NetworkFilter nf, RequestFilter uf)
		{
			nwFilter = nf;
			qFilter = uf;
		}

		public NetworkFilter getNetworkFilter()
		{
			return nwFilter;
		}
	}

	// all network buffers currently in use
	private static final List buffers = new ArrayList();
	private static int uniqueInstID;

	private static final LogService logger =
		LogManager.getManager().getLogger("Network-Buffer");

	private final List configs = new ArrayList();
	private final String inst;

	private NetworkBuffer(String installation)
	{
		inst = installation;
	}

	public static NetworkBuffer createBuffer(String installationID)
	{
		final NetworkBuffer b = new NetworkBuffer(validateInstID(installationID));
		buffers.add(b);
		return b;
	}

	public static NetworkBuffer getBuffer(String installationID)
	{
		for (final Iterator i = buffers.iterator(); i.hasNext();) {
			final NetworkBuffer db = (NetworkBuffer) i.next();
			if (db.inst.equals(installationID))
				return db;
		}
		return null;
	}

	public static NetworkBuffer[] getAllBuffers()
	{
		return (NetworkBuffer[]) buffers.toArray(new NetworkBuffer[buffers.size()]);
	}

	public static void removeBuffer(String installationID)
	{
		for (final Iterator i = buffers.iterator(); i.hasNext();) {
			final NetworkBuffer b = (NetworkBuffer) i.next();
			if (b.inst.equals(installationID)) {
				i.remove();
				while (!b.configs.isEmpty())
					b.removeConfiguration((Configuration) b.configs
						.get(b.configs.size() - 1));
				return;
			}
		}
	}

	// a close to the returned network link automatically unregisters the link
	/**
	 * <p>
	 * If <code>link</code> closes, the created and returned configuration will get
	 * deactivated (see {@link Configuration#activate(boolean)}), and the buffered link
	 * of the configuration, obtained with {@link Configuration#getBufferedLink()}, will
	 * get closed as well.
	 * 
	 * @param link
	 * @param installationID
	 */
	public static Configuration createConfiguration(KNXNetworkLink link,
		String installationID)
	{
		final NetworkBuffer b = createBuffer(installationID);
		return b.createConfiguration(link);
	}

	public static void removeConfiguration(Configuration c, String installationID)
	{
		for (final Iterator i = buffers.iterator(); i.hasNext();) {
			final NetworkBuffer b = (NetworkBuffer) i.next();
			if (b.inst.equals(installationID)
				|| (installationID == null && b.configs.contains(c))) {
				b.removeConfiguration(c);
				return;
			}
		}
	}

	// a close to the returned network link automatically unregisters the link
	public Configuration createConfiguration(KNXNetworkLink link)
	{
		final ConfigImpl c = new ConfigImpl(link);
		configs.add(c);
		return c;
	}

	public Configuration getConfiguration(KNXNetworkLink link)
	{
		for (final Iterator i = configs.iterator(); i.hasNext();) {
			final ConfigImpl lc = (ConfigImpl) i.next();
			if (lc.getBufferedLink() == link)
				return lc;
		}
		return null;
	}

	public String getInstallationID()
	{
		return inst;
	}

	public void removeConfiguration(Configuration c)
	{
		if (configs.remove(c))
			((ConfigImpl) c).unregister();
	}

	private static String validateInstID(String instID)
	{
		if (instID == null || instID.length() == 0)
			return "Installation " + ++uniqueInstID;
		return instID;
	}
}
