/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: Cache.java,v $
 Revision 1.3  2007/07/14 21:03:50  bm
 added API specification comments

 Revision 1.2  2007/04/13 13:15:46  bm
 added API specification comments
 code cleanup

 Revision 1.1  2007/03/01 20:11:08  bm
 adding cache interface and dummy impl. for lfucache
 
 */

package tuwien.auto.calimero.cache;

/**
 * Cache interface for use by different caching policies.
 * <p>
 * Cache entries are unambiguously identified through a key, as defined by
 * {@link CacheObject}.<br>
 * For comparison of keys, {@link Object#hashCode()} and
 * {@link Object#equals(Object)} are used.
 * 
 * @author bm
 * @see CacheObject
 */
public interface Cache
{
	/**
	 * Offers information about data counting and cache performance.
	 * <p>
	 * 
	 * @author bm
	 */
	public interface Statistic
	{
		/**
		 * Returns the total number of successful requests with
		 * {@link Cache#get(Object)} for a {@link CacheObject}.
		 * <p>
		 * 
		 * @return total hits
		 */
		long hits();

		/**
		 * Returns the total number of {@link CacheObject} not found with
		 * {@link Cache#get(Object)}.
		 * <p>
		 * 
		 * @return total misses
		 */
		long misses();

		/**
		 * Returns the ratio between found {@link CacheObject}s to total
		 * requests through {@link Cache#get(Object)}.
		 * <p>
		 * This value is obtained through
		 * <code>ratio = (hits / (hits + misses))</code>.
		 * 
		 * @return the hit ratio in the range [0,1]
		 */
		double hitRatio();
	}
	
	/**
	 * Gets the {@link CacheObject} associated with <code>key</code> from the
	 * cache.
	 * <p>
	 * If found, the access count of the CacheObject is incremented by 1.
	 * 
	 * @param key key to search
	 * @return the CacheObject or <code>null</code> if <code>key</code> does
	 *         not exist in the cache
	 */
	CacheObject get(Object key);
	
	/**
	 * Inserts the {@link CacheObject} <code>obj</code> into the cache.
	 * <p>
	 * If a {@link CacheObject} with an equal key (delivered by
	 * {@link CacheObject#getKey()}) to <code>obj.getKey()</code> is already
	 * in the cache, it will be replaced by <code>obj</code>.<br>
	 * {@link CacheObject#resetTimestamp()} is invoked on <code>obj</code> after 
	 * <code>obj</code> was inserted successfully.
	 * 
	 * @param obj CacheObject to put into the cache
	 */
	void put(CacheObject obj);
	
	/**
	 * Removes the CacheObject associated with <code>key</code> from the cache,
	 * if found.
	 * <p>
	 * 
	 * @param key key of CacheObject to remove
	 */
	void remove(Object key);
	
	/**
	 * Empties the cache of all {@link CacheObject}s.
	 * <p>
	 */
	void clear();
	
	/**
	 * Removes all {@link CacheObject}s which are not valid anymore, as defined
	 * by a caching policy, from the cache.
	 * <p>
	 */
	void removeExpired();
	
	/**
	 * Returns information collected by this cache since its creation.
	 * <p>
	 * 
	 * @return a {@link Statistic} object
	 */
	Statistic statistic();
}
