/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: CacheObject.java,v $
 Revision 1.3  2007/07/14 21:03:51  bm
 added API specification comments

 Revision 1.2  2007/04/13 13:15:46  bm
 added API specification comments
 code cleanup

 Revision 1.1  2007/03/02 15:05:24  bm
 added wrapper class for a cache element

 */

package tuwien.auto.calimero.cache;

import tuwien.auto.calimero.exception.KNXIllegalArgumentException;

/**
 * Holds a key and a value entry and is used by a {@link Cache}.
 * <p>
 * The key object is not allowed to change after creation of this cache object.<br>
 * An access count is maintained to tell the number of requests for this
 * CacheObject by a cache.<br>
 * The usage value might be used by a caching policy to store a different way of
 * access counting.<br>
 * A timestamp is stored of the most recent assignment of a value object hold by
 * this CacheObject. This timestamp is created by
 * {@link System#currentTimeMillis()}.
 * 
 * @author bm
 * @see Cache
 * @see System#currentTimeMillis()
 */
public class CacheObject
{
	/**
	 * The value object hold by this cache object.
	 */
	protected Object value;
	
	private final Object key;
	// hit count (increment count) of this object
	private int count;
	// timestamp of last value assignment
	private volatile long timestamp;
	// client usage indicator of this object (cache specific)
	private int usage;

	/**
	 * Creates a CacheObject associated with <code>key</code> holding a
	 * <code>value</code> entry.
	 * <p>
	 * 
	 * @param key key of this {@link CacheObject}
	 * @param value value of this {@link CacheObject}
	 */
	public CacheObject(Object key, Object value)
	{
		if (key == null || value == null)
			throw new KNXIllegalArgumentException(
				"key and value must not be null");
		this.key = key;
		this.value = value;
		resetTimestamp();
	}

	/**
	 * Returns the key associated with this cache object.
	 * <p>
	 * 
	 * @return the key object
	 */
	public Object getKey()
	{
		return key;
	}

	/**
	 * Gets the value entry hold by this cache object.
	 * <p>
	 * Note that the value is guaranteed to be always a non <code>null</code>
	 * reference.
	 * 
	 * @return the value object
	 */
	public Object getValue()
	{
		return value;
	}

	/**
	 * Returns the most up to date timestamp of the moment the current value
	 * object was assigned to this cache object, or the moment this object was
	 * {@link Cache#put(CacheObject)} into a cache.
	 * <p>
	 * 
	 * @return the timestamp in milliseconds of type long
	 * @see System#currentTimeMillis()
	 */
	public final long getTimestamp()
	{
		return timestamp;
	}

	/**
	 * Sets the timestamp to time = now.
	 * <p>
	 * The timestamp is obtained by {@link System#currentTimeMillis()}.
	 */
	public final void resetTimestamp()
	{
		timestamp = System.currentTimeMillis();
	}
	
	/**
	 * Returns the access count this cache object was queried from the cache.
	 * <p>
	 * 
	 * @return the access count
	 */
	public final int getCount()
	{
		return count;
	}

	/**
	 * Returns the usage value of this cache object as defined by the caching
	 * policy.
	 * <p>
	 * 
	 * @return the usage value
	 */
	public final int getUsage()
	{
		return usage;
	}

	/**
	 * Increments the access count by 1.
	 * <p>
	 * It is invoked by a {@link Cache} implementation to record a client access to
	 * this object.
	 */
	public final synchronized void incCount()
	{
		++count;
	}

	/**
	 * Sets the access count to 0.
	 * <p>
	 */
	protected final synchronized void resetCount()
	{
		count = 0;
	}

	/**
	 * Sets a new usage value.
	 * <p>
	 * The usage value might be used by a caching policy to store a different
	 * way of access counting.<br>
	 * 
	 * @param newUsage new usage value
	 */
	protected final void setUsage(int newUsage)
	{
		usage = newUsage;
	}
}
