/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: ExpiringCache.java,v $
 Revision 1.2  2007/04/13 13:15:46  bm
 added API specification comments
 code cleanup

 Revision 1.1  2007/04/09 19:06:34  bm
 Extract supertype for expiring cache objects

 */

package tuwien.auto.calimero.cache;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Implements a cache expiring mechanism for {@link CacheObject}s.
 * <p>
 * A time span is specified, how long a value is considered valid. After that
 * time for expiring the cache object is removed on the next call of
 * {@link #removeExpired()} by an internal {@link CacheSweeper}.<br>
 * The timestamp of {@link CacheObject#getTimestamp()} is used to determine if a
 * cache object value has expired.<br>
 * Note that if the timestamp of a {@link CacheObject} changes after it was put
 * into the cache, the object has to be reinserted (see
 * {@link #put(CacheObject)}) to keep the cache in a consistent state. If no
 * expiring is used, this may be omitted.<br>
 * The default sweep interval used by a {@link CacheSweeper} is
 * {@value ExpiringCache#defaultSweepInterval}
 * seconds.
 * 
 * @author bm
 */
public abstract class ExpiringCache implements Cache
{
	/**
	 * Default sweep interval in seconds used for a {@link CacheSweeper}.
	 * <p>
	 */
	protected static final int defaultSweepInterval = 60;
	/**
	 * Sweep interval in seconds used for the {@link CacheSweeper}.
	 * <p>
	 * It defaults to {@value #defaultSweepInterval} seconds. For a new value to
	 * take effect, it has to be assigned either before the first
	 * {@link #startSweeper()} call, or the cache sweeper has to be restarted.
	 */
	protected int sweepInterval = defaultSweepInterval;
	/**
	 * The map holding the {@link CacheObject}s.
	 * <p>
	 */
	protected Map map;
	private CacheSweeper sweeper;
	private int timeToExpire;
	
	/**
	 * Creates an {@link ExpiringCache}.
	 * <p>
	 * Note that for the actual begin of cache sweeping {@link #startSweeper()}
	 * has to be invoked.
	 * 
	 * @param timeToExpire timespan > 0 in seconds for cache entries to stay
	 *        valid, otherwise no expiring will occur
	 */
	public ExpiringCache(int timeToExpire)
	{
		if (timeToExpire > 0) {
			this.timeToExpire = timeToExpire;
			map = new LinkedHashMap();
		}
		else
			map = new HashMap();
	}

	/**
	 * Removes all {@link CacheObject}s, where<br>
	 * <code>{@link CacheObject#getTimestamp()} + timeToExpire &lt;= now</code>,
	 * with <code>timeToExpire &gt; 0</code> and <code>now</code> is the
	 * point of time {@link #removeExpired()} is invoked. <br>
	 * If no expiring time was specified at creation of cache, no cache object
	 * will be expired.
	 */
	public void removeExpired()
	{
		if (timeToExpire == 0 || !(map instanceof LinkedHashMap))
			return;
		final long now = System.currentTimeMillis();
		final long duration = timeToExpire * 1000;
		CacheObject o = null;
		synchronized (this) {
			for (final Iterator i = map.values().iterator(); i.hasNext();) {
				o = (CacheObject) i.next();
				if (now >= o.getTimestamp() + duration) {
					i.remove();
					notifyRemoved(o);
				}
				else
					break;
			}
		}
	}

	/**
	 * Override this method to get notified when {@link #removeExpired()}
	 * removed a {@link CacheObject} from the {@link #map}.
	 * <p>
	 * 
	 * @param obj removed {@link CacheObject}
	 */
	protected void notifyRemoved(CacheObject obj)
	{}
	
	/**
	 * Starts a new {@link CacheSweeper}, if not already running, and if an
	 * expiring time for {@link CacheObject} was specified.
	 */
	protected final void startSweeper()
	{
		if (timeToExpire > 0)
			synchronized (this) {
				if (sweeper == null)
					(sweeper = new CacheSweeper(this, sweepInterval)).start();
			}
	}

	/**
	 * Stops the {@link CacheSweeper}, if any.
	 */
	protected final void stopSweeper()
	{
		final CacheSweeper s = sweeper;
		if (s != null) {
			s.stopSweeper();
			sweeper = null;
		}
	}
}
