/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: CEMICacheObject.java,v $
 Revision 1.3  2007/07/14 21:03:50  bm
 added API specification comments

 Revision 1.2  2007/04/13 13:15:46  bm
 added API specification comments
 code cleanup

 Revision 1.1  2007/03/02 15:04:04  bm
 added cache object impl. for cEMI frames

 */

package tuwien.auto.calimero.cache;

import tuwien.auto.calimero.KNXAddress;
import tuwien.auto.calimero.cemi.CEMILData;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;

/**
 * Used for holding {@link CEMILData} frames.
 * <p>
 * 
 * @author bm
 */
public class LDataCacheObject extends CacheObject
{
	/**
	 * Creates a {@link LDataCacheObject} holding the <code>frame</code> argument.
	 * <p>
	 * The cache object key is generated using {@link #createKey(CEMILData)}.
	 * 
	 * @param frame {@link CEMILData} frame
	 */
	public LDataCacheObject(CEMILData frame)
	{
		super(createKey(frame), frame);
	}

	/**
	 * Creates a {@link LDataCacheObject} set to key <code>addrKey</code> and holding
	 * <code>value</code>.
	 * <p>
	 * If <code>value</code> is not of type {@link CEMILData}, the methods
	 * {@link #setFrame(CEMILData)} and {@link #getFrame()} have to be overridden.
	 * 
	 * @param addrKey key of this cache object
	 * @param value value hold by this cache object
	 */
	protected LDataCacheObject(String addrKey, Object value)
	{
		super(addrKey, value);
	}

	/**
	 * Creates the key used for the {@link CEMILData} <code>frame</code> by a
	 * {@link LDataCacheObject}.
	 * <p>
	 * Note that for key creation just the KNX destination address is of relevance. So
	 * this method is equal to {@link #createKey(KNXAddress)} with argument
	 * frame.getDestinationAddress().
	 * 
	 * @param frame {@link CEMILData} frame to which key is requested
	 * @return the created key
	 */
	public static String createKey(CEMILData frame)
	{
		return frame.getDestination().toString() + "-" + frame.getDestination().getType();
	}

	/**
	 * Creates the key used for a {@link CEMILData} frame with KNX address
	 * <code>addr</code>.
	 * <p>
	 * 
	 * @param addr KNX address to which key is requested
	 * @return the created key
	 */
	public static String createKey(KNXAddress addr)
	{
		return addr.toString() + "-" + addr.getType();
	}

	/**
	 * Returns a {@link CEMILData} frame contained in this cache object.
	 * <p>
	 * On no frame available, <code>null</code> is returned.
	 * 
	 * @return the {@link CEMILData} frame, or <code>null</code>
	 */
	public CEMILData getFrame()
	{
		return (CEMILData) value;
	}

	/**
	 * Sets a new {@link CEMILData} <code>frame</code> for this {@link LDataCacheObject}.
	 * <p>
	 * The key generated out of <code>frame</code> (i.e. out of the KNX address of
	 * <code>frame</code>) has to be equal to the key of this {@link LDataCacheObject},
	 * as returned from {@link #getKey()}. Otherwise a
	 * {@link KNXIllegalArgumentException} will be thrown.<br>
	 * Note that on successful set the timestamp is renewed.
	 * 
	 * @param frame the new {@link CEMILData} frame to set
	 */
	public void setFrame(CEMILData frame)
	{
		if (!LDataCacheObject.createKey(frame).equals(getKey()))
			throw new KNXIllegalArgumentException("frame key differs from cache key");
		value = frame;
		resetTimestamp();
	}
}
