/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: LFUCache.java,v $
 Revision 1.3  2007/07/14 21:03:50  bm
 added API specification comments

 Revision 1.2  2007/04/13 13:15:46  bm
 added API specification comments
 code cleanup

 Revision 1.1  2007/03/01 20:11:08  bm
 adding cache interface and dummy impl. for lfucache

 */

package tuwien.auto.calimero.cache;

import java.util.Comparator;
import java.util.TreeMap;

/**
 * A {@link Cache} using a LFU strategy.
 * <p>
 * The usage value of {@link CacheObject#getUsage()} equals the access count,
 * {@link CacheObject#getCount()}.
 * 
 * @author bm
 */
public class LFUCache extends ExpiringCache
{
	private TreeMap tree;
	private int maxSize;
	private long hits;
	private long misses;
	
	/**
	 * Creates a new LFU cache.
	 * <p>
	 * Optionally, a maximum cache size and an expiring time can be specified.
	 * 
	 * @param cacheSize maximum number of {@link CacheObject}s in the cache, or
	 *        0 for no maximum
	 * @param timeToExpire timespan in seconds for cache objects to stay valid,
	 *        or 0 for no expiring
	 */
	public LFUCache(int cacheSize, int timeToExpire)
	{
		super(timeToExpire);
		if (cacheSize > 0)
			maxSize = cacheSize;
		tree = new TreeMap(new LFUObjectCompare());
	}

	/* (non-Javadoc)
	 * @see tuwien.auto.calimero.cache.Cache#clear()
	 */
	public synchronized void clear()
	{
		stopSweeper();
		map.clear();
		tree.clear();
	}

	/* (non-Javadoc)
	 * @see tuwien.auto.calimero.cache.Cache#get()
	 */
	public synchronized CacheObject get(Object key)
	{
		final CacheObject o = (CacheObject) map.get(key);
		if (o != null) {
			tree.remove(o);
			updateAccess(o);
			tree.put(o, o);
			++hits;
		}
		else
			++misses;
		return o;
	}

	/**
	 * {@inheritDoc}<br>
	 * If expiring of cache objects is set, and the timestamp of a
	 * {@link CacheObject} is renewed after it has been put into the cache, a
	 * new {@link #put(CacheObject)} is required for that object to apply the
	 * timestamp and keep the cache in a consistent state.
	 */
	public synchronized void put(CacheObject obj)
	{
		// ensure sweeping is on if we have expiring objects
		startSweeper();
		final Object old = map.remove(obj.getKey());
		if (old != null)
			tree.remove(old);
		else
			ensureSizeLimits();
		obj.resetTimestamp();
		map.put(obj.getKey(), obj);
		tree.put(obj, obj);
	}

	/* (non-Javadoc)
	 * @see tuwien.auto.calimero.cache.Cache#remove()
	 */
	public synchronized void remove(Object key)
	{
		final Object o = map.remove(key);
		if (o != null)
			tree.remove(o);
	}
	
	/* (non-Javadoc)
	 * @see tuwien.auto.calimero.cache.Cache#statistic()
	 */
	public synchronized Statistic statistic()
	{
		return new StatisticImpl(hits, misses);
	}

	/* (non-Javadoc)
	 * @see tuwien.auto.calimero.cache.ExpiringCache#notifyRemoved
	 * (tuwien.auto.calimero.cache.CacheObject)
	 */
	protected final void notifyRemoved(CacheObject obj)
	{
		tree.remove(obj);
	}	
	
	private static void updateAccess(CacheObject obj)
	{
		obj.incCount();
		obj.setUsage(obj.getCount());
		// obj.setUsage(obj.getUsage() >> 1 | (1 << 30));
	}

	private void ensureSizeLimits()
	{
		if (maxSize > 0)
			while (map.size() >= maxSize)
				remove(((CacheObject) tree.firstKey()).getKey());
	}

	private static class LFUObjectCompare implements Comparator
	{
		public LFUObjectCompare() {}
		public int compare(Object o1, Object o2)
		{
			final CacheObject cmp1 = (CacheObject) o1;
			final CacheObject cmp2 = (CacheObject) o2;
			if (cmp1.getUsage() > cmp2.getUsage())
				return 1;
			if (cmp1.getUsage() < cmp2.getUsage())
				return -1;
			if (cmp1.getCount() > cmp2.getCount())
				return 1;
			if (cmp1.getCount() < cmp2.getCount())
				return -1;
			return 0;
		}
	}
}
