/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log$
 */

package tuwien.auto.calimero.datapoint;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import tuwien.auto.calimero.GroupAddress;
import tuwien.auto.calimero.Priority;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;
import tuwien.auto.calimero.exception.KNXIllegalStateException;
import tuwien.auto.calimero.xml.Attribute;
import tuwien.auto.calimero.xml.Element;
import tuwien.auto.calimero.xml.KNXMLException;
import tuwien.auto.calimero.xml.XMLReader;
import tuwien.auto.calimero.xml.XMLWriter;

/**
 * Represents a KNX datapoint.
 * <p>
 * It stores knowledge about a datapoint in the KNX network, used for communication within
 * Calimero, the KNX network and the user.<br>
 * The datapoint is identified through a KNX group address. A name is supplied to allow a
 * more friendly interaction with the user. Information exchanged between datapoints
 * consists of a certain encoding, defined by a datapoint type (DPT). This information
 * exchange is done through messages, which are sent with a {@link Priority} associated
 * with the respective datapoint.<br>
 * A KNX datapoint is either state-based or event-based. State-based datapoint interaction
 * leads to transition between states (so a datapoint value associated to a certain
 * datapoint represents a particular state over an amount of time), while event-based
 * datapoint interaction is done solely with events (i.e. "commands"), not telling
 * anything about a datapoint state. A state-based datapoint's state can be
 * updated/invalidated by different KNX group addresses.
 * <p>
 * Note, no datapoint values or events, nor any datapoint messages are stored within this
 * type.
 * 
 * @author bm
 */
public class Datapoint
{
	private static final String TAG_DATAPOINT = "datapoint";
	private static final String ATTR_NAME = "name";
	private static final String ATTR_STATEBASED = "stateBased";
	private static final String ATTR_MAINNUMBER = "mainNumber";
	private static final String ATTR_DPTID = "dptID";
	private static final String ATTR_PRIORITY = "priority";
	private static final String TAG_UPDATING = "updatingAddresses";
	private static final String TAG_INVALIDATING = "invalidatingAddresses";

	private final GroupAddress main;
	private String name;

	// DPT translation, a not set DPT is identified through mainNo=0 and dptID=null
	private int mainNo;
	private String dptID;

	// message priority for this datapoint
	private Priority priority = Priority.LOW;

	private final boolean stateBased;

	// members below only available for state-based data points

	// list of group addresses, whose .ind messages invalidate the data point
	private List invalidating;
	// list of group addresses, whose .ind and .res messages update the data point
	private List updating;

	/**
	 * 
	 */
	public Datapoint(GroupAddress main, String name, boolean stateBased)
	{
		this.main = main;
		this.name = name;
		this.stateBased = stateBased;
		if (stateBased) {
			invalidating = Collections.synchronizedList(new ArrayList());
			updating = Collections.synchronizedList(new ArrayList());
		}
	}

	public Datapoint(GroupAddress main, String name, boolean stateBased, int mainNumber,
		String dptID)
	{
		this(main, name, stateBased);
		setDPT(mainNumber, dptID);
	}

	// state based only
	public Datapoint(GroupAddress point, String name, Collection invalidatingAddresses,
		Collection updatingAddresses)
	{
		this(point, name, true);
		invalidating = Collections.synchronizedList(new ArrayList(invalidatingAddresses));
		checkGAs(invalidating);
		updating = Collections.synchronizedList(new ArrayList(updatingAddresses));
		checkGAs(updating);
	}

	public Datapoint(XMLReader r) throws KNXMLException
	{
		if (r.getPosition() != XMLReader.START_TAG)
			r.read();
		final Element e = r.getCurrent();
		final int line = r.getLineNumber();
		if (r.getPosition() != XMLReader.START_TAG || !e.getName().equals(TAG_DATAPOINT))
			throw new KNXMLException("no KNX datapoint element", e != null ? e.getName()
				: null, line);
		if ((name = e.getAttribute(ATTR_NAME)) == null)
			throw new KNXMLException("missing attribute " + ATTR_NAME, null, line);
		String a = e.getAttribute(ATTR_STATEBASED);
		if ("false".equalsIgnoreCase(a))
			stateBased = false;
		else if ("true".equalsIgnoreCase(a)) {
			stateBased = true;
			invalidating = Collections.synchronizedList(new ArrayList());
			updating = Collections.synchronizedList(new ArrayList());
		}
		else
			throw new KNXMLException("malformed attribute " + ATTR_STATEBASED, a, line);

		if ((dptID = e.getAttribute(ATTR_DPTID)) == null)
			throw new KNXMLException("missing attribute " + ATTR_DPTID, null, line);
		if (dptID.length() == 0)
			dptID = null;
		try {
			mainNo = Integer.decode(a = e.getAttribute(ATTR_MAINNUMBER)).intValue();
			priority = Priority.get(a = e.getAttribute(ATTR_PRIORITY));
		}
		catch (final RuntimeException re) {
			throw new KNXMLException("malformed attibute " + ATTR_MAINNUMBER, a, line);
		}
		GroupAddress initMain = null;
		while (r.read() == XMLReader.START_TAG) {
			final String tag = r.getCurrent().getName();
			if (stateBased && tag.equals(TAG_UPDATING))
				while (r.read() == XMLReader.START_TAG)
					updating.add(new GroupAddress(r));
			else if (stateBased && tag.equals(TAG_INVALIDATING)) {
				while (r.read() == XMLReader.START_TAG)
					invalidating.add(new GroupAddress(r));
			}
			else if (initMain == null)
				initMain = new GroupAddress(r);
			else
				throw new KNXMLException("invalid element", tag, r.getLineNumber());
		}
		main = initMain;
		if (main == null)
			throw new KNXMLException("DP has no main address", null, r.getLineNumber());
	}

	public final String getName()
	{
		return name;
	}

	public final void setName(String friendlyName)
	{
		name = friendlyName;
	}

	public final GroupAddress getMainAddress()
	{
		return main;
	}

	public final boolean isStateBased()
	{
		return stateBased;
	}

	public final Priority getPriority()
	{
		return priority;
	}

	public final void setPriority(Priority p)
	{
		this.priority = p;
	}

	public final void setDPT(int mainNumber, String dptID)
	{
		mainNo = mainNumber;
		this.dptID = dptID;
	}

	public final int getMainNumber()
	{
		return mainNo;
	}

	public final String getDPT()
	{
		return dptID;
	}

	public Collection getAddresses(boolean updatingAddresses)
	{
		if (!stateBased)
			return null;
		return Collections.unmodifiableCollection(updatingAddresses ? updating
			: invalidating);
	}

	public final boolean isInvalidating(GroupAddress a)
	{
		return invalidating != null && invalidating.contains(a);
	}

	public final boolean isUpdating(GroupAddress a)
	{
		return updating != null && updating.contains(a);
	}

	public final void add(GroupAddress a, boolean toUpdating)
	{
		if (!stateBased)
			throw new KNXIllegalStateException("no state-based datapoint");
		if (toUpdating)
			updating.add(a);
		else
			invalidating.add(a);
	}

	public final void remove(GroupAddress a, boolean fromUpdating)
	{
		if (!stateBased)
			throw new KNXIllegalStateException("no state-based datapoint");
		if (fromUpdating)
			updating.remove(a);
		else
			invalidating.remove(a);
	}

	public void save(XMLWriter w) throws KNXMLException
	{
		final List att = new ArrayList();
		att.add(new Attribute(ATTR_NAME, name));
		att.add(new Attribute(ATTR_STATEBASED, Boolean.toString(stateBased)));
		att.add(new Attribute(ATTR_MAINNUMBER, Integer.toString(mainNo)));
		att.add(new Attribute(ATTR_DPTID, dptID == null ? "" : dptID));
		att.add(new Attribute(ATTR_PRIORITY, priority.toString()));
		w.writeElement(TAG_DATAPOINT, att, null);
		main.save(w);
		if (stateBased) {
			w.writeElement(TAG_UPDATING, Collections.EMPTY_LIST, null);
			for (final Iterator i = updating.iterator(); i.hasNext();)
				((GroupAddress) i.next()).save(w);
			w.endElement();
			w.writeElement(TAG_INVALIDATING, Collections.EMPTY_LIST, null);
			for (final Iterator i = invalidating.iterator(); i.hasNext();)
				((GroupAddress) i.next()).save(w);
			w.endElement();
		}
		w.endElement();
	}

	// iteration not synchronized
	private void checkGAs(List l)
	{
		for (final Iterator i = l.iterator(); i.hasNext();)
			if (!(i.next() instanceof GroupAddress))
				throw new KNXIllegalArgumentException("not a group address list");
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString()
	{
		return main.toString() + " " + name + ", DPT main " + mainNo + " id "
			+ (dptID == null ? "-" : dptID) + ", priority " + priority.toString()
			+ (stateBased ? ", state-based" : ", event-based");
	}
}
