/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log$
 */

package tuwien.auto.calimero.datapoint;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import tuwien.auto.calimero.GroupAddress;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;
import tuwien.auto.calimero.xml.Element;
import tuwien.auto.calimero.xml.KNXMLException;
import tuwien.auto.calimero.xml.XMLReader;
import tuwien.auto.calimero.xml.XMLWriter;

/**
 * @author bm
 */
public class DatapointMap implements DatapointModel
{
	private static final String TAG_DATAPOINTS = "datapoints";

	private final Map points;

	/**
	 * 
	 */
	public DatapointMap()
	{
		points = new HashMap(20);
	}

	/**
	 * 
	 */
	public DatapointMap(Collection datapoints)
	{
		// not all HashSets put additional capacity in HashSet(Collection) ctor
		final Map m = new HashMap(Math.max(2 * datapoints.size(), 11));
		for (final Iterator i = datapoints.iterator(); i.hasNext();) {
			final Datapoint dp = (Datapoint) i.next();
			if (m.containsKey(dp.getMainAddress()))
				throw new KNXIllegalArgumentException("duplicate datapoint "
					+ dp.getMainAddress());
			m.put(dp.getMainAddress(), dp);
		}
		points = Collections.synchronizedMap(m);
	}

	public void add(Datapoint dp)
	{
		synchronized (points) {
			if (points.containsKey(dp.getMainAddress()))
				throw new KNXIllegalArgumentException("duplicate datapoint "
					+ dp.getMainAddress());
			points.put(dp.getMainAddress(), dp);
		}
	}

	public void remove(Datapoint dp)
	{
		points.remove(dp.getMainAddress());
	}

	public boolean contains(GroupAddress main)
	{
		return points.containsKey(main);
	}

	public boolean contains(Datapoint dp)
	{
		return points.containsKey(dp.getMainAddress());
	}

	public Datapoint get(GroupAddress main)
	{
		return (Datapoint) points.get(main);
	}

	public void removeAll()
	{
		points.clear();
	}

	public Collection getDatapoints()
	{
		return Collections.unmodifiableCollection(points.values());
	}

	public void load(XMLReader r) throws KNXMLException
	{
		if (r.getPosition() != XMLReader.START_TAG)
			r.read();
		final Element e = r.getCurrent();
		if (r.getPosition() != XMLReader.START_TAG || !e.getName().equals(TAG_DATAPOINTS))
			throw new KNXMLException(TAG_DATAPOINTS + " element not found", e != null ? e
				.getName() : null, r.getLineNumber());
		synchronized (points) {
			while (r.read() == XMLReader.START_TAG) {
				final Datapoint dp = new Datapoint(r);
				if (points.containsKey(dp.getMainAddress()))
					throw new KNXMLException(TAG_DATAPOINTS + " element contains "
						+ "duplicate datapoint", dp.getMainAddress().toString(), r
						.getLineNumber());
				points.put(dp.getMainAddress(), dp);
			}
		}
	}

	public void save(XMLWriter w) throws KNXMLException
	{
		w.writeElement(TAG_DATAPOINTS, Collections.EMPTY_LIST, null);
		synchronized (points) {
			for (final Iterator i = points.values().iterator(); i.hasNext();)
				((Datapoint) i.next()).save(w);
		}
		w.endElement();
	}
}
