/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log$
 */

package tuwien.auto.calimero.dptxlator;

import java.util.HashMap;
import java.util.Map;

import tuwien.auto.calimero.exception.KNXFormatException;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;

/**
 * Translator for KNX DPT 7, type <b>2 byte unsigned value</b>.
 * <p>
 * The KNX data type width is 2 bytes.<br>
 * The default return value after creation is 0.<br>
 */
public class DPTXlator2ByteUnsigned extends DPTXlator
{
	/**
	 * DPT ID 7.001, Unsigned count; values from <b>0</b> to <b>65535</b> pulses.
	 * <p>
	 */
	public static final DPT DPT_VALUE_2_UCOUNT =
		new DPT("7.001", "Unsigned count", "0", "65535", "pulses");

	/**
	 * DPT ID 7.002, Time period in ms; values from <b>0</b> to <b>65535</b> ms.
	 * <p>
	 */
	public static final DPT DPT_TIMEPERIOD =
		new DPT("7.002", "Time period in ms", "0", "65535", "ms");

	/**
	 * DPT ID 7.003, Time period (resolution 10 ms); values from <b>0</b> to <b>655350</b>
	 * ms.
	 * <p>
	 */
	public static final DPT DPT_TIMEPERIOD_10 =
		new DPT("7.003", "Time period (resolution 10 ms)", "0", "655350", "ms");

	/**
	 * DPT ID 7.004, Time period (resolution 100 ms); values from <b>0</b> to <b>6553500</b>
	 * ms.
	 * <p>
	 */
	public static final DPT DPT_TIMEPERIOD_100 =
		new DPT("7.004", "Time period (resolution 100 ms)", "0", "6553500", "ms");

	/**
	 * DPT ID 7.005, Time period in seconds; values from <b>0</b> to <b>65535</b> s
	 * (~18,2 hours).
	 * <p>
	 */
	public static final DPT DPT_TIMEPERIOD_SEC =
		new DPT("7.005", "Time period in seconds", "0", "65535", "s");

	/**
	 * DPT ID 7.006, Time period in minutes; values from <b>0</b> to <b>65535</b> min
	 * (~45,5 days).
	 * <p>
	 */
	public static final DPT DPT_TIMEPERIOD_MIN =
		new DPT("7.006", "Time period in minutes", "0", "65535", "min");

	/**
	 * DPT ID 7.007, Time period in hours; values from <b>0</b> to <b>65535</b> h (~7,4
	 * years).
	 * <p>
	 */
	public static final DPT DPT_TIMEPERIOD_HOURS =
		new DPT("7.007", "Time period in hours", "0", "65535", "h");

	/**
	 * DPT ID 7.010, Interface object property ID; values from <b>0</b> to <b>65535</b>.
	 * <p>
	 */
	public static final DPT DPT_PROP_DATATYPE =
		new DPT("7.010", "Interface object property ID", "0", "65535", "");

	/**
	 * DPT ID 7.012, Electrical current; values from <b>0</b> to <b>65535</b> mA.
	 * <p>
	 * A value of 0 indicates no bus power supply functionality available, 1 to 65535 is
	 * the current in mA.
	 */
	public static final DPT DPT_ELECTRICAL_CURRENT =
		new DPT("7.012", "Electrical current", "0", "65535", "mA");

	/**
	 * DPT ID 7.013, Brightness; values from <b>0</b> to <b>65535</b> lux.
	 * <p>
	 */
	public static final DPT DPT_BRIGHTNESS =
		new DPT("7.013", "Brightness", "0", "65535", "lux");

	private static final Map types;

	private final int min;
	private final int max;

	static {
		types = new HashMap(15);
		types.put(DPT_VALUE_2_UCOUNT.getID(), DPT_VALUE_2_UCOUNT);
		types.put(DPT_PROP_DATATYPE.getID(), DPT_PROP_DATATYPE);
		types.put(DPT_TIMEPERIOD.getID(), DPT_TIMEPERIOD);
		types.put(DPT_TIMEPERIOD_10.getID(), DPT_TIMEPERIOD_10);
		types.put(DPT_TIMEPERIOD_100.getID(), DPT_TIMEPERIOD_100);
		types.put(DPT_TIMEPERIOD_SEC.getID(), DPT_TIMEPERIOD_SEC);
		types.put(DPT_TIMEPERIOD_MIN.getID(), DPT_TIMEPERIOD_MIN);
		types.put(DPT_TIMEPERIOD_HOURS.getID(), DPT_TIMEPERIOD_HOURS);
		types.put(DPT_ELECTRICAL_CURRENT.getID(), DPT_ELECTRICAL_CURRENT);
		types.put(DPT_BRIGHTNESS.getID(), DPT_BRIGHTNESS);
	}

	/**
	 * Creates a translator for the given datapoint type.
	 * <p>
	 * 
	 * @param dpt the requested datapoint type
	 * @throws KNXFormatException on not supported or not available DPT
	 */
	public DPTXlator2ByteUnsigned(DPT dpt) throws KNXFormatException
	{
		this(dpt.getID());
	}

	/**
	 * Creates a translator for the given datapoint type ID.
	 * <p>
	 * 
	 * @param dptID available implemented datapoint type ID
	 * @throws KNXFormatException on wrong formatted or not expected (available)
	 *         <code>dptID</code>
	 */
	public DPTXlator2ByteUnsigned(String dptID) throws KNXFormatException
	{
		super(2);
		setTypeID(types, dptID);
		min = getLimit(dpt.getLowerValue());
		max = getLimit(dpt.getUpperValue());
		data = new short[2];
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.dptxlator.DPTXlator#getAllValues()
	 */
	public String[] getAllValues()
	{
		final String[] s = new String[data.length / 2];
		for (int i = 0; i < s.length; ++i)
			s[i] = makeString(i);
		return s;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.dptxlator.DPTXlator#getData(byte[], int)
	 */
	public byte[] getData(byte[] dst, int offset)
	{
		final int end = Math.min(data.length, dst.length - offset) & ~0x01;
		for (int i = 0; i < end; ++i)
			dst[offset + i] = (byte) data[i];
		return dst;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.dptxlator.DPTXlator#getSubTypes()
	 */
	public final Map getSubTypes()
	{
		return types;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.dptxlator.DPTXlator#getValue()
	 */
	public String getValue()
	{
		return makeString(0);
	}

	/**
	 * Returns the first translation item as int.
	 * <p>
	 * A value of DPT {@link #DPT_TIMEPERIOD_10} or {@link #DPT_TIMEPERIOD_100} is
	 * returned in unit millisecond, i.e. a KNX DPT_TIMEPERIOD_10 data value is multiplied
	 * with 10, DPT_TIMEPERIOD_100 with 100.<br>
	 * On any other DPT the value is returned according to its unit.
	 * 
	 * @return value as unsigned int
	 * @see #getType()
	 */
	public final int getValueInt()
	{
		return fromDPT(0);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.dptxlator.DPTXlator#setData(byte[], int)
	 */
	public void setData(byte[] data, int offset)
	{
		if (offset < 0 || offset > data.length)
			throw new KNXIllegalArgumentException("illegal offset " + offset);
		final int size = (data.length - offset) & ~0x01;
		if (size > 0) {
			this.data = new short[size];
			for (int i = 0; i < size; ++i)
				this.data[i] = ubyte(data[offset + i]);
		}
	}

	/**
	 * Sets the value of the first translation item.
	 * <p>
	 * A value of DPT {@link #DPT_TIMEPERIOD_10} or {@link #DPT_TIMEPERIOD_100} is
	 * expected in unit millisecond, i.e. a value of DPT_TIMEPERIOD_10 gets divided by 10,
	 * a value of DPT_TIMEPERIOD_100 by 100. The result is rounded to the nearest
	 * representable value (with 0.5 rounded up). On any other DPT the value is expected
	 * according to its unit.
	 * 
	 * @param value unsigned value, 0 &lt;= value &lt;= max, with
	 *        <ul>
	 *        <li>max = 655350 on DPT {@link #DPT_TIMEPERIOD_10}</li>
	 *        <li>max = 6553500 on DPT {@link #DPT_TIMEPERIOD_100}</li>
	 *        <li>max = 65535 otherwise</li>
	 *        </ul>
	 * @throws KNXFormatException on input value out of range for DPT
	 * @see #getType()
	 */
	public final void setValue(int value) throws KNXFormatException
	{
		data = toDPT(value);
	}

	/**
	 * {@inheritDoc} The <code>value</code> might be formatted using decimal,
	 * hexadecimal, and octal numbers, distinguished by using these prefixes:
	 * <dd>no prefix for decimal numeral
	 * <dd><code>0x</code>, <code>0X</code> or <code>#</code> for hexadecimal
	 * <dd><code>0</code> for octal numeral
	 */
	public void setValue(String value) throws KNXFormatException
	{
		data = toDPT(value);
	}

	/**
	 * {@inheritDoc} The same numeral formats as for {@link #setValue(String)} are
	 * allowed.
	 */
	public void setValues(String[] values) throws KNXFormatException
	{
		if (values.length > 0) {
			final short[] buf = new short[2 * values.length];
			for (int i = 0; i < values.length; ++i) {
				final short[] d = toDPT(values[i]);
				buf[2 * i] = d[0];
				buf[2 * i + 1] = d[1];
			}
			data = buf;
		}
	}

	/**
	 * @return the subtypes of the 2-byte unsigned translator type
	 * @see DPTXlator#getSubTypesStatic()
	 */
	protected static Map getSubTypesStatic()
	{
		return types;
	}

	/**
	 * Sets the first translation item value from input of unit millisecond.
	 * <p>
	 * The method is for DPTs dealing with periods of time, in particular
	 * {@link #DPT_TIMEPERIOD}, {@link #DPT_TIMEPERIOD_10}, {@link #DPT_TIMEPERIOD_100}
	 * {@link #DPT_TIMEPERIOD_SEC}, {@link #DPT_TIMEPERIOD_MIN} and
	 * {@link #DPT_TIMEPERIOD_HOURS}. The milliseconds are converted to the unit of the
	 * set DPT, with the result rounded to the nearest representable value (with 0.5
	 * rounded up).<br>
	 * On any other DPT, the input is treated equal to {@link #setValue(int)}.
	 * 
	 * @param milliseconds the value in milliseconds, 0 &lt;= <code>milliseconds</code>
	 * @throws KNXFormatException on milliseconds out of range for DPT
	 */
	public final void setTimePeriod(long milliseconds) throws KNXFormatException
	{
		data = toDPT(milliseconds);
	}

	private int fromDPT(int index)
	{
		final int v = (data[2 * index] << 8) | data[2 * index + 1];
		if (dpt.equals(DPT_TIMEPERIOD_10))
			return v * 10;
		else if (dpt.equals(DPT_TIMEPERIOD_100))
			return v * 100;
		return v;
	}

	private String makeString(int index)
	{
		return appendUnit(String.valueOf(fromDPT(index)));
	}

	private short[] toDPT(String value) throws KNXFormatException
	{
		try {
			return toDPT(Integer.decode(removeUnit(value)).intValue());
		}
		catch (final NumberFormatException e) {
			throw new KNXFormatException("wrong value format", value);
		}
	}

	private short[] toDPT(long ms) throws KNXFormatException
	{
		// prevent round up to 0 from negative milliseconds
		if (ms < 0)
			throw new KNXFormatException("negative input value", Long.toString(ms));
		long v = ms;
		if (dpt.equals(DPT_TIMEPERIOD_SEC))
			v = Math.round(ms / 1000.0);
		else if (dpt.equals(DPT_TIMEPERIOD_MIN))
			v = Math.round(ms / 1000.0 / 60.0);
		else if (dpt.equals(DPT_TIMEPERIOD_HOURS))
			v = Math.round(ms / 1000.0 / 60.0 / 60.0);
		return toDPT((int) v);
	}

	private short[] toDPT(int value) throws KNXFormatException
	{
		if (value < min || value > max)
			throw new KNXFormatException("input value out of range ["
				+ dpt.getLowerValue() + ".." + dpt.getUpperValue() + "]", value);
		int v = value;
		if (dpt.equals(DPT_TIMEPERIOD_10))
			v = Math.round(value / 10.0f);
		else if (dpt.equals(DPT_TIMEPERIOD_100))
			v = Math.round(value / 100.0f);
		return new short[] { ubyte(v >> 8), ubyte(v) };
	}

	private int getLimit(String limit) throws KNXFormatException
	{
		try {
			final int i = Integer.parseInt(limit);
			final int upper =
				dpt.equals(DPT_TIMEPERIOD_10) ? 655350 : dpt.equals(DPT_TIMEPERIOD_100)
					? 6553500 : 65535;
			if (i >= 0 && i <= upper)
				return i;
		}
		catch (final NumberFormatException e) {}
		throw new KNXFormatException("invalid DPT range", limit);
	}
}
