/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log$
 */

package tuwien.auto.calimero.dptxlator;

import java.util.HashMap;
import java.util.Map;

import tuwien.auto.calimero.exception.KNXFormatException;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;

/**
 * Translator for KNX DPT 12, type <b>4 byte unsigned value</b>.
 * <p>
 * The KNX data type width is 4 bytes.<br>
 * The default return value after creation is 0.<br>
 */
public class DPTXlator4ByteUnsigned extends DPTXlator
{
	/**
	 * DPT ID 12.001, Unsigned count; values from <b>0</b> to <b>4294967295</b> counter
	 * pulses.
	 * <p>
	 */
	public static final DPT DPT_VALUE_4_UCOUNT =
		new DPT("12.001", "Unsigned count", "0", "4294967295", "counter pulses");

	private static final Map types;

	static {
		types = new HashMap(2);
		types.put(DPT_VALUE_4_UCOUNT.getID(), DPT_VALUE_4_UCOUNT);
	}

	/**
	 * Creates a translator for the given datapoint type.
	 * <p>
	 * 
	 * @param dpt the requested datapoint type
	 * @throws KNXFormatException on not supported or not available DPT
	 */
	public DPTXlator4ByteUnsigned(DPT dpt) throws KNXFormatException
	{
		this(dpt.getID());
	}

	/**
	 * Creates a translator for the given datapoint type ID.
	 * <p>
	 * 
	 * @param dptID available implemented datapoint type ID
	 * @throws KNXFormatException on wrong formatted or not expected (available)
	 *         <code>dptID</code>
	 */
	public DPTXlator4ByteUnsigned(String dptID) throws KNXFormatException
	{
		super(4);
		setTypeID(types, dptID);
		data = new short[4];
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.dptxlator.DPTXlator#getAllValues()
	 */
	public String[] getAllValues()
	{
		final String[] s = new String[data.length / 4];
		for (int i = 0; i < s.length; ++i)
			s[i] = makeString(i);
		return s;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.dptxlator.DPTXlator#getData(byte[], int)
	 */
	public byte[] getData(byte[] dst, int offset)
	{
		final int end = Math.min(data.length, dst.length - offset) & ~0x03;
		for (int i = 0; i < end; ++i)
			dst[offset + i] = (byte) data[i];
		return dst;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.dptxlator.DPTXlator#getSubTypes()
	 */
	public final Map getSubTypes()
	{
		return types;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.dptxlator.DPTXlator#getValue()
	 */
	public String getValue()
	{
		return makeString(0);
	}

	/**
	 * Returns the first translation item.
	 * <p>
	 * 
	 * @return unsigned 32 bit value
	 * @see #getType()
	 */
	public final long getValueLong()
	{
		return fromDPT(0);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.dptxlator.DPTXlator#setData(byte[], int)
	 */
	public void setData(byte[] data, int offset)
	{
		if (offset < 0 || offset > data.length)
			throw new KNXIllegalArgumentException("illegal offset " + offset);
		final int size = (data.length - offset) & ~0x03;
		if (size > 0) {
			this.data = new short[size];
			for (int i = 0; i < size; ++i)
				this.data[i] = ubyte(data[offset + i]);
		}
	}

	/**
	 * Sets the value of the first translation item.
	 * <p>
	 * 
	 * @param value unsigned value, 0 &lt;= value &lt;= 0xFFFFFFFF
	 * @throws KNXFormatException on input value out of range for DPT
	 * @see #getType()
	 */
	public final void setValue(long value) throws KNXFormatException
	{
		data = toDPT(value, new short[4], 0);
	}

	/**
	 * {@inheritDoc} The <code>value</code> might be formatted using decimal,
	 * hexadecimal, and octal numbers, distinguished by using these prefixes:
	 * <dd>no prefix for decimal numeral
	 * <dd><code>0x</code>, <code>0X</code> or <code>#</code> for hexadecimal
	 * <dd><code>0</code> for octal numeral
	 */
	public void setValue(String value) throws KNXFormatException
	{
		data = toDPT(value, new short[4], 0);
	}

	/**
	 * {@inheritDoc} The same numeral formats as for {@link #setValue(String)} are
	 * allowed.
	 */
	public void setValues(String[] values) throws KNXFormatException
	{
		if (values.length > 0) {
			final short[] buf = new short[4 * values.length];
			for (int i = 0; i < values.length; ++i)
				toDPT(values[i], buf, i);
			data = buf;
		}
	}

	/**
	 * @return the subtypes of the 4-byte unsigned translator type
	 * @see DPTXlator#getSubTypesStatic()
	 */
	protected static Map getSubTypesStatic()
	{
		return types;
	}

	private long fromDPT(int index)
	{
		final int i = 4 * index;
		return (long) data[i] << 24 | data[i + 1] << 16 | data[i + 2] << 8 | data[i + 3];
	}

	private String makeString(int index)
	{
		return appendUnit(Long.toString(fromDPT(index)));
	}

	private short[] toDPT(String value, short[] dst, int index) throws KNXFormatException
	{
		try {
			return toDPT(Long.decode(removeUnit(value)).longValue(), dst, index);
		}
		catch (final NumberFormatException e) {
			throw new KNXFormatException("wrong value format", value);
		}
	}

	private short[] toDPT(long value, short[] dst, int index) throws KNXFormatException
	{
		if (value < 0 || value > 0xFFFFFFFFL)
			throw new KNXFormatException("input value out of range", Long.toString(value));
		final int i = 4 * index;
		dst[i] = (short) ((value >> 24) & 0xFF);
		dst[i + 1] = (short) ((value >> 16) & 0xFF);
		dst[i + 2] = (short) ((value >> 8) & 0xFF);
		dst[i + 3] = (short) (value & 0xFF);
		return dst;
	}
}
