/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: DPTXlatorString.java,v $
 Revision 1.1  2007/07/15 00:53:28  bm
 added API specification comments

 Revision 1.2  2007/02/07 23:03:01  bm
 manual character set encoding / decoding (no import of java.nio.* classes)
 for port to cdc foundation profile

 Revision 1.1  2007/02/07 01:59:28  bm
 initial commit

 */

package tuwien.auto.calimero.dptxlator;

import java.util.HashMap;
import java.util.Map;

import tuwien.auto.calimero.exception.KNXFormatException;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;

/**
 * Translator for KNX DPT 16 <b>String</b>.
 * <p>
 * The KNX data type width is 14 bytes.<br>
 * The default return value after creation is the empty string "" or 14 bytes filled with
 * <i>NULL</i> (ASCII code 0).<br>
 * Characters not supported by the selected character set are silently replaced with the
 * question mark character, '?'.<br>
 */
public class DPTXlatorString extends DPTXlator
{
	/**
	 * DPT ID 16.000, ASCII string; 7 Bit character set encoding.
	 * <p>
	 */
	public static final DPT DPT_STRING_ASCII = new DPT("16.000", "ASCII string", "", "");

	/**
	 * DPT ID 16.001, ISO-8859-1 string (Latin 1); 8 Bit character set encoding.
	 * <p>
	 */
	public static final DPT DPT_STRING_8859_1 =
		new DPT("16.001", "ISO-8859-1 string (latin 1)", "", "");

	private static final Map types;

	// default replacement for unsupported characters
	private static final short replacement = '?';

	// standard string length of this type
	private static final int stringLength = 14;

	static {
		types = new HashMap(5);
		types.put(DPT_STRING_ASCII.getID(), DPT_STRING_ASCII);
		types.put(DPT_STRING_8859_1.getID(), DPT_STRING_8859_1);
	}

	/**
	 * Creates a translator for the given datapoint type.
	 * <p>
	 * 
	 * @param dpt the requested datapoint type
	 * @throws KNXFormatException on not supported or not available DPT
	 */
	public DPTXlatorString(DPT dpt) throws KNXFormatException
	{
		this(dpt.getID());
	}

	/**
	 * Creates a translator for the given datapoint type ID.
	 * <p>
	 * 
	 * @param dptID available implemented datapoint type ID
	 * @throws KNXFormatException on wrong formatted or not expected (available)
	 *         <code>dptID</code>
	 */
	public DPTXlatorString(String dptID) throws KNXFormatException
	{
		super(14);
		setTypeID(types, dptID);
		data = new short[stringLength];
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.dptxlator.DPTXlator#getAllValues()
	 */
	public String[] getAllValues()
	{
		final String[] buf = new String[getItems()];
		for (int i = 0; i < buf.length; ++i)
			buf[i] = fromDPT(i);
		return buf;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.dptxlator.DPTXlator#getData(byte[], int)
	 */
	public byte[] getData(byte[] dst, int offset)
	{
		final int items = Math.min(data.length, dst.length - offset) / stringLength;
		final int end = items * stringLength;
		if (end == 0)
			logger.error("destination range has to be multiple of 14 bytes "
				+ "for KNX strings");
		for (int i = 0; i < end; ++i)
			dst[offset + i] = (byte) data[i];
		return dst;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.dptxlator.DPTXlator#getSubTypes()
	 */
	public final Map getSubTypes()
	{
		return types;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.dptxlator.DPTXlator#setData(byte[], int)
	 */
	public void setData(byte[] data, int offset)
	{
		if (offset < 0 || offset > data.length)
			throw new KNXIllegalArgumentException("illegal offset " + offset);
		if (data.length - offset > 0)
			toDPT(data, offset);
	}

	/**
	 * {@inheritDoc}<br>
	 * 
	 * @param value string for translation, <code>value.length</code> &lt;= 14
	 */
	public void setValue(String value) throws KNXFormatException
	{
		if (value.length() > stringLength)
			throw new KNXFormatException("maximum string length is 14", value);
		data = toDPT(value, new short[stringLength], 0);
	}

	/**
	 * {@inheritDoc}<br>
	 * 
	 * @param values strings for translation, <code>values[index].length</code> &lt;= 14
	 */
	public void setValues(String[] values) throws KNXFormatException
	{
		if (values.length > 0) {
			final short[] buf = new short[stringLength * values.length];
			for (int i = 0; i < values.length; ++i)
				toDPT(values[i], buf, i);
			data = buf;
		}
	}

	/**
	 * @return the subtypes of the string translator type
	 * @see DPTXlator#getSubTypesStatic()
	 */
	protected static Map getSubTypesStatic()
	{
		return types;
	}

	private String fromDPT(int index)
	{
		final int offset = index * stringLength;
		final char[] output = new char[stringLength];
		int strlen = 0;
		for (int i = offset; i < (offset + stringLength) && data[i] != 0; ++i) {
			output[strlen] = (char) data[i];
			++strlen;
		}
		return new String(output, 0, strlen);
	}

	private void toDPT(byte[] buf, int offset)
	{
		// check character set, default to ASCII 7 bit encoding
		int rangeMax = 0x7f;
		if (dpt.equals(DPT_STRING_8859_1))
			rangeMax = 0xff;
		final int items = (buf.length - offset) / stringLength;
		if (items == 0) {
			logger.error("source range has to be multiple of 14 bytes for KNX strings");
			return;
		}
		data = new short[items * stringLength];
		// 7 bit encoded characters of ISO-8859-1 are equal to ASCII
		// the lower 256 characters of UCS correspond to ISO-8859-1
		for (int i = 0; i < data.length; ++i) {
			final short c = ubyte(buf[offset + i]);
			data[i] = c <= rangeMax ? c : replacement;
		}
	}

	private short[] toDPT(String value, short[] dst, int index) throws KNXFormatException
	{
		// check character set, default to ASCII 7 bit encoding
		char rangeMax = '\u007f';
		if (dpt.equals(DPT_STRING_8859_1))
			rangeMax = '\u00ff';
		// 7 bit encoded characters of ISO-8859-1 are equal to ASCII
		// the lower 256 characters of UCS correspond to ISO-8859-1
		if (value.length() > stringLength)
			throw new KNXFormatException("KNX string length maximum is 14 chars", value);
		final int offset = index * stringLength;
		for (int i = 0; i < value.length(); ++i) {
			final char c = value.charAt(i);
			dst[offset + i] = c <= rangeMax ? (short) c : replacement;
		}
		return dst;
	}
}
