/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: XlatorTypes.java,v $
 Revision 1.1  2007/07/15 00:53:27  bm
 added API specification comments

 Revision 1.2  2007/03/01 22:51:16  bm
 minor stylistic changes

 Revision 1.1  2007/02/07 01:59:28  bm
 initial commit

 */

package tuwien.auto.calimero.dptxlator;

import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Map;

import tuwien.auto.calimero.dptxlator.DPTXlator.DPT;
import tuwien.auto.calimero.exception.KNXException;
import tuwien.auto.calimero.exception.KNXFormatException;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;

/**
 * Maintains available KNX datapoint main types and its DPT translators.
 * <p>
 * It stores all available, registered DP main types with the corresponding translator and
 * an optional description of the type.<br>
 * For more common used data types, the main types are declared as constants, although
 * this doesn't necessarily indicate a translator is actually available.<br>
 * All DPT translator implementations in this package are registered here and available by
 * default. Translators might be added or removed by the user.
 * <p>
 * A datapoint type consists of a data type and a dimension. The data type is referred to
 * through a main number, the existing dimensions of a data type are listed through sub
 * numbers. The data type specifies format and encoding, while dimension specifies the
 * range and unit.<br>
 * A datapoint identifier (dptID for short), stands for one particular datapoint type, and
 * is built up as expression "<i>main number</i>.<i>sub number</i>".<br>
 * In short, a datapoint type has a dptID and standardizes one combination of format,
 * encoding, range and unit.
 * <p>
 * Note: main and sub refer to the former used terms major / minor.
 * 
 * @author bm
 * @see tuwien.auto.calimero.dptxlator.DPTXlator
 */
public final class TranslatorTypes
{
	/**
	 * DPT main number for <b>B1 (boolean)</b>, number = {@value #TYPE_BOOLEAN}.
	 * <p>
	 */
	public static final int TYPE_BOOLEAN = 1;

	/**
	 * DPT main number for <b>B2 (1 Bit controlled)</b>, number =
	 * {@value #TYPE_1BIT_CONTROLLED}.
	 * <p>
	 */
	public static final int TYPE_1BIT_CONTROLLED = 2;

	/**
	 * DPT main number for <b>B1U3 (3 Bit controlled)</b>, number =
	 * {@value #TYPE_3BIT_CONTROLLED}.
	 * <p>
	 */
	public static final int TYPE_3BIT_CONTROLLED = 3;

	/**
	 * DPT main number for <b>character set</b>, number = {@value #TYPE_CHARACTER_SET}.
	 * <p>
	 */
	public static final int TYPE_CHARACTER_SET = 4;

	/**
	 * DPT main number for <b>8 Bit unsigned value</b>, number =
	 * {@value #TYPE_8BIT_UNSIGNED}.
	 * <p>
	 */
	public static final int TYPE_8BIT_UNSIGNED = 5;

	/**
	 * DPT main number for <b>V8 (8 Bit signed value)</b>, shares main number with
	 * "Status with mode", number = {@value #TYPE_8BIT_SIGNED}.
	 * <p>
	 */
	public static final int TYPE_8BIT_SIGNED = 6;

	/**
	 * DPT main number for <b>2-octet unsigned value</b>, number =
	 * {@value #TYPE_2OCTET_UNSIGNED}.
	 * <p>
	 */
	public static final int TYPE_2OCTET_UNSIGNED = 7;

	/**
	 * DPT main number for <b>2-octet signed value</b>, number =
	 * {@value #TYPE_2OCTET_SIGNED}.
	 * <p>
	 */
	public static final int TYPE_2OCTET_SIGNED = 8;

	/**
	 * DPT main number for <b>2-octet float value</b>, number =
	 * {@value #TYPE_2OCTET_FLOAT}.
	 * <p>
	 */
	public static final int TYPE_2OCTET_FLOAT = 9;

	/**
	 * DPT main number for <b>time</b>, number = {@value #TYPE_TIME}.
	 * <p>
	 */
	public static final int TYPE_TIME = 10;

	/**
	 * DPT main number for <b>date</b>, number = {@value #TYPE_DATE}.
	 * <p>
	 */
	public static final int TYPE_DATE = 11;

	/**
	 * DPT main number for <b>4-octet unsigned value</b>, number =
	 * {@value #TYPE_4OCTET_UNSIGNED}.
	 * <p>
	 */
	public static final int TYPE_4OCTET_UNSIGNED = 12;

	/**
	 * DPT main number for <b>4-octet signed value</b>, number =
	 * {@value #TYPE_4OCTET_SIGNED}.
	 * <p>
	 */
	public static final int TYPE_4OCTET_SIGNED = 13;

	/**
	 * DPT main number for <b>4-octet float value</b>, number =
	 * {@value #TYPE_4OCTET_FLOAT}.
	 * <p>
	 */
	public static final int TYPE_4OCTET_FLOAT = 14;

	/**
	 * DPT main number for <b>access data</b>, number = {@value #TYPE_ACCESS}.
	 * <p>
	 */
	public static final int TYPE_ACCESS = 15;

	/**
	 * DPT main number for <b>string</b>, number = {@value #TYPE_STRING}.
	 * <p>
	 */
	public static final int TYPE_STRING = 16;

	/**
	 * DPT main number for <b>date/time</b>, number = {@value #TYPE_DATE_TIME}.
	 * <p>
	 */
	public static final int TYPE_DATE_TIME = 19;

	/**
	 * Maps a main data type to a corresponding translator class doing the DPT
	 * translations.
	 * <p>
	 * 
	 * @author bm
	 */
	public static class MainType
	{
		private final Class xlator;
		private final String desc;
		private final int main;

		/**
		 * Creates a data main type to translator mapping.
		 * <p>
		 * 
		 * @param mainNumber main number assigned to the data type
		 * @param translator represents a translator class of type {@link DPTXlator}
		 * @param description textual information describing this data type to a user, use
		 *        <code>null</code> for no description
		 */
		public MainType(int mainNumber, Class translator, String description)
		{
			if (mainNumber <= 0)
				throw new KNXIllegalArgumentException("invalid main number");
			if (!DPTXlator.class.isAssignableFrom(translator)
				|| DPTXlator.class.equals(translator))
				throw new KNXIllegalArgumentException(translator.getName()
					+ " is not a valid DPT translator type");
			main = mainNumber;
			xlator = translator;
			desc = description == null ? "" : description;
		}

		/**
		 * Returns the data type main number.
		 * <p>
		 * 
		 * @return main number as int
		 */
		public final int getMainNumber()
		{
			return main;
		}

		/**
		 * Creates a new translator for the given datapoint type.
		 * <p>
		 * 
		 * @param dpt datapoint type specifying the particular translation behavior; if
		 *        the datapoint type is not a subtype of this main type, a
		 *        {@link KNXFormatException} is thrown
		 * @return the new {@link DPTXlator} instance
		 * @throws KNXFormatException to forward all target exceptions thrown in the
		 *         constructor of the translator
		 * @throws KNXException thrown on translator class creation errors (e.g. security /
		 *         access problems)
		 */
		public final DPTXlator createTranslator(DPT dpt) throws KNXException
		{
			return createTranslator(dpt.getID());
		}

		/**
		 * Creates a new instance of the translator for the given datapoint type ID.
		 * <p>
		 * 
		 * @param dptID datapoint type ID for selecting a particular kind of value
		 *        translation; if the datapoint type ID is not assigned to a subtype of
		 *        this main type, a {@link KNXFormatException} is thrown
		 * @return the new {@link DPTXlator} instance
		 * @throws KNXFormatException to forward all target exceptions thrown in the
		 *         constructor of the translator
		 * @throws KNXException thrown on translator class creation errors (e.g. security /
		 *         access problems)
		 */
		public DPTXlator createTranslator(String dptID) throws KNXException
		{
			try {
				return (DPTXlator) xlator.getConstructor(new Class[] { String.class })
					.newInstance(new String[] { dptID });
			}
			catch (final InvocationTargetException e) {
				// try to forward encapsulated target exception
				if (e.getTargetException() instanceof KNXFormatException)
					throw (KNXFormatException) e.getTargetException();
				// throw generic message
				throw new KNXFormatException("failed to init translator", dptID);
			}
			catch (final NoSuchMethodException e) {
				DPTXlator.logger
					.fatal("DPT translator is required to have a public ctor(String dptID)");
				throw new KNXException("interface specification error at translator");
			}
			catch (final Exception e) {
				// for SecurityException, InstantiationException, IllegalAccessException
				throw new KNXException("failed to create translator, " + e.getMessage());
			}
		}

		/**
		 * Returns the translator class used for this main type.
		 * <p>
		 * 
		 * @return the translator class as {@link Class} object
		 */
		public Class getTranslator()
		{
			return xlator;
		}

		/**
		 * Returns the description to this main type, if any.
		 * <p>
		 * 
		 * @return description as String, or the empty string if no description set
		 */
		public final String getDescription()
		{
			return desc;
		}

		/**
		 * Returns a map containing all implemented subtypes.
		 * <p>
		 * 
		 * @return available subtypes as {@link Map}
		 * @throws KNXException thrown on problems accessing the translator while
		 *         retrieving sub types (e.g. security / access problem) for external,
		 *         client-supplied translators
		 */
		public Map getSubTypes() throws KNXException
		{
			try {
				return (Map) xlator.getDeclaredMethod("getSubTypesStatic", null).invoke(
					null, null);
			}
			catch (final NoSuchMethodException e) {
				throw new KNXException("no method to get subtypes, " + e.getMessage());
			}
			catch (final Exception e) {
				// for SecurityException and IllegalAccessException
				// no reason for InvocationTargetException
				throw new KNXException("security / access problem, " + e.getMessage());
			}
		}
	}

	private static Map map;

	private TranslatorTypes()
	{}

	static {
		map = new HashMap(20);
		map.put(new Integer(TYPE_BOOLEAN), new MainType(TYPE_BOOLEAN,
			DPTXlatorBoolean.class, "Boolean (main type 1)"));
		map.put(new Integer(TYPE_3BIT_CONTROLLED), new MainType(TYPE_3BIT_CONTROLLED,
			DPTXlator3BitControlled.class, "3 Bit controlled (main type 3)"));
		map.put(new Integer(TYPE_8BIT_UNSIGNED), new MainType(TYPE_8BIT_UNSIGNED,
			DPTXlator8BitUnsigned.class, "8 Bit unsigned value (main type 5)"));
		map.put(new Integer(TYPE_2OCTET_UNSIGNED), new MainType(TYPE_2OCTET_UNSIGNED,
			DPTXlator2ByteUnsigned.class, "2-octet unsigned value (main type 7)"));
		map.put(new Integer(TYPE_2OCTET_FLOAT), new MainType(TYPE_2OCTET_FLOAT,
			DPTXlator2ByteFloat.class, "2-octet float value (main type 9)"));
		map.put(new Integer(TYPE_4OCTET_UNSIGNED), new MainType(TYPE_4OCTET_UNSIGNED,
			DPTXlator4ByteUnsigned.class, "4-octet unsigned value (main type 12)"));
		map.put(new Integer(TYPE_STRING), new MainType(TYPE_STRING,
			DPTXlatorString.class, "String (main type 16)"));
		map.put(new Integer(TYPE_DATE_TIME), new MainType(TYPE_DATE_TIME,
			DPTXlatorDateTime.class, "Date with time (main type 19)"));
	}

	/**
	 * Returns the {@link MainType} object assigned the given data type main number.
	 * <p>
	 * 
	 * @param mainNumber main type to lookup
	 * @return the main type information found, or <code>null</code> if main number not
	 *         listed
	 */
	public static MainType getMainType(int mainNumber)
	{
		return (MainType) map.get(new Integer(mainNumber));
	}

	/**
	 * Returns all available data types which have a DPT translator implementation
	 * assigned.
	 * <p>
	 * The map returned is the same used by this class for type lookup. Map entries can be
	 * added, likewise entries might be removed to change future lookup results.
	 * 
	 * @return a {@link Map} containing all data types as {@link MainType} objects
	 */
	public static Map getAllMainTypes()
	{
		return map;
	}

	/**
	 * Creates a DPT translator for the given datapoint type ID.
	 * <p>
	 * The translation behavior of a DPT translator instance is uniquely defined by the
	 * supplied datapoint type ID.
	 * <p>
	 * If the <code>dptID</code> argument is built up the recommended way, that is "<i>main
	 * number</i>.<i>sub number</i>", the <code>mainNumber</code> argument might be
	 * left 0 to use the datapoint type ID only.<br>
	 * Note, that we don't enforce any particular or standardized format on the dptID
	 * structure, so using a different formatted dptID solely without main number argument
	 * results in undefined behavior.
	 * 
	 * @param mainNumber data type main number, number >= 0; use 0 to infer translator
	 *        type from <code>dptID</code> argument only
	 * @param dptID datapoint type ID for selecting a particular kind of value translation
	 * @return the new {@link DPTXlator} object
	 * @throws KNXException on main type not found or creation failed (refer to
	 *         {@link MainType#createTranslator(String)})
	 */
	public static DPTXlator createTranslator(int mainNumber, String dptID)
		throws KNXException
	{
		final int main =
			mainNumber != 0 ? mainNumber : Integer.parseInt(dptID.substring(0, dptID
				.indexOf('.')));
		final MainType type = (MainType) map.get(new Integer(main));
		if (type == null)
			throw new KNXException("main number not listed");
		return type.createTranslator(dptID);
	}
}
