/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log$
 */

package tuwien.auto.calimero.knxnetip.util;

import tuwien.auto.calimero.exception.KNXFormatException;

/**
 * Container for a connection response data (CRD).
 * <p>
 * The CRD structure is returned by a connection response, and contains information
 * regarding the established communication.<br>
 * It contains description data of a specific connection type. A CRD is built up of host
 * protocol independent data and host protocol dependent data, both optional. Refer to the
 * available subtypes for more specific type information.
 * <p>
 * For now, a plain CRD is returned for management connections, since this connection type
 * doesn't require any additional host protocol data.
 * <p>
 * Factory methods are provided for creation of CRD objects.
 * 
 * @author bm
 * @see tuwien.auto.calimero.knxnetip.servicetype.ConnectResponse
 */
public class CRD extends CRBase
{
	/**
	 * Creates a new CRD out of a byte array.
	 * <p>
	 * 
	 * @param data byte array containing a CRD structure
	 * @param offset start offset of CRD in <code>data</code>
	 * @throws KNXFormatException if no CRD found or invalid structure
	 */
	public CRD(byte[] data, int offset) throws KNXFormatException
	{
		super(data, offset);
	}

	/**
	 * Creates a new CRD for the given connection type.
	 * <p>
	 * The array of <code>optionalData</code> is not copied for internal storage. No
	 * additional checks regarding content are done.
	 * 
	 * @param connectionType connection type the description is intended for
	 * @param optionalData byte array containing optional host protocol independent and
	 *        dependent data, this information is located starting at offset 2 in the CRD
	 *        structure, <code>optionalData.length</code> < 254
	 */
	public CRD(short connectionType, byte[] optionalData)
	{
		super(connectionType, optionalData);
	}

	/**
	 * Creates a new CRD out of a byte array.
	 * <p>
	 * If possible, a matching, more specific, CRD subtype is returned. Note, that CRD for
	 * specific communication types might expect certain characteristics on
	 * <code>data</code> (regarding contained data).<br>
	 * 
	 * @param data byte array containing the CRD structure
	 * @param offset start offset of CRD in <code>data</code>
	 * @return the new CRD object
	 * @throws KNXFormatException if no CRD found or invalid structure
	 */
	public static CRD createResponse(byte[] data, int offset)
		throws KNXFormatException
	{
		return (CRD) create(false, data, offset);
	}

	/**
	 * Creates a CRD for the given connection type.
	 * <p>
	 * If possible, a matching, more specific, CRD subtype is returned. Note, that CRD for
	 * specific communication types might expect certain characteristics on
	 * <code>optionalData</code> (regarding length and/or content).<br>
	 * 
	 * @param connectionType connection type this CRD is used for (e.g. tunnel connection)
	 * @param optionalData byte array containing optional host protocol independent and
	 *        dependent data, this information is located starting at offset 2 in the CRD
	 *        structure, <code>optionalData.length</code> < 254, may be
	 *        <code>null</code> for no optional data
	 * @return the new CRD object
	 */
	public static CRD createResponse(short connectionType, byte[] optionalData)
	{
		return (CRD) create(false, connectionType, optionalData);
	}
}
