/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: ManufacturerDIB.java,v $
 Revision 1.1  2007/07/20 00:52:50  bm
 added API specification comments

 */

package tuwien.auto.calimero.knxnetip.util;

import tuwien.auto.calimero.exception.KNXFormatException;

/**
 * Represents a manufacturer data description information block.
 * <p>
 * Since the data in this DIB is dependent on the manufacturer and might contain any
 * information, no specific content parsing is done.
 * 
 * @author bm
 */
public class ManufacturerDIB extends DIB
{
	private final int id;
	private final byte[] mfrData;

	/**
	 * Creates a manufacturer data DIB out of a byte array.
	 * <p>
	 * 
	 * @param data byte array containing manufacturer data DIB structure
	 * @param offset start offset of DIB in <code>data</code>
	 * @throws KNXFormatException if no DIB found or invalid structure
	 */
	public ManufacturerDIB(byte[] data, int offset) throws KNXFormatException
	{
		super(data, offset);
		if (type != MFR_DATA)
			throw new KNXFormatException("DIB is not of type manufacturer data", type);
		if (size < 4)
			throw new KNXFormatException("MFR DIB too short");
		id = ((data[2] & 0xFF) << 8) | (data[3] & 0xFF);
		mfrData = new byte[data.length - offset - 4];
		for (int i = 0; i < mfrData.length; ++i)
			mfrData[i] = data[2 + offset + i];
	}

	/**
	 * Returns the KNX manufacturer ID.
	 * <p>
	 * The ID clearly identifies the manufacturer who created this DIB structure.
	 * 
	 * @return ID as unsigned short
	 */
	public final int getID()
	{
		return id;
	}

	/**
	 * Returns the manufacturer specific description data.
	 * <p>
	 * This data block starts at byte offset 4 in the DIB structure.
	 * 
	 * @return byte array with manufacturer data
	 */
	public final byte[] getData()
	{
		return (byte[]) mfrData.clone();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.knxnetip.util.DIB#toByteArray()
	 */
	public byte[] toByteArray()
	{
		final byte[] buf = super.toByteArray();
		buf[2] = (byte) (id >> 8);
		buf[3] = (byte) id;
		for (int i = 0; i < mfrData.length; ++i)
			buf[4 + i] = mfrData[i];
		return buf;
	}
}
