/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: SuppFamiliesDIB.java,v $
 Revision 1.1  2007/07/20 00:52:50  bm
 added API specification comments

 Revision 1.1  2007/02/07 01:59:28  bm
 initial commit

 */

package tuwien.auto.calimero.knxnetip.util;

import java.io.ByteArrayInputStream;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import tuwien.auto.calimero.exception.KNXFormatException;

/**
 * Supported service families description information block.
 * <p>
 * It informs about the service families supported by the device.<br>
 * 
 * @author bm
 * @see tuwien.auto.calimero.knxnetip.servicetype.DescriptionResponse
 */
public class SuppFamiliesDIB extends DIB
{
	private static final String[] familyNames =
		{ null, null, "Core", "Device Management", "Tunneling", "Routing",
			"Remote Logging", "Remote Configuration/Diagnosis", "Object Server" };

	private final Map map = new HashMap();

	/**
	 * Creates a supported families DIB out of a byte array.
	 * <p>
	 * 
	 * @param data byte array containing the service families DIB structure
	 * @param offset start offset of DIB in <code>data</code>
	 * @throws KNXFormatException if no DIB found or invalid structure
	 */
	public SuppFamiliesDIB(byte[] data, int offset) throws KNXFormatException
	{
		super(data, offset);
		if (type != SUPP_SVC_FAMILIES)
			throw new KNXFormatException("not a supported service families DIB", type);
		final ByteArrayInputStream is =
			new ByteArrayInputStream(data, offset + 2, data.length - offset - 2);
		for (int i = 2; i < size; i += 2) {
			final short family = (short) is.read();
			map.put(new Short(family), new Short((short) is.read()));
		}
	}

	/**
	 * Returns all supported service families, each family together with the version it is
	 * implemented and supported up to.
	 * <p>
	 * The returned set holds {@link Map.Entry} items, with the service family of type
	 * {@link Short} being the key, and the version of type {@link Short} being the value.
	 * 
	 * @return an unmodifiable set containing supported entries (family-version pair)
	 */
	public final Set getFamilies()
	{
		return Collections.unmodifiableSet(map.entrySet());
	}

	/**
	 * Returns the version associated to a given supported service family.
	 * <p>
	 * If the service family is not supported, 0 is returned.
	 * 
	 * @param family supported service family to lookup
	 * @return version as unsigned byte, or 0
	 */
	public final short getVersion(short family)
	{
		return ((Short) map.get(new Short(family))).shortValue();
	}

	/**
	 * Returns the service family name for the supplied family ID.
	 * <p>
	 * 
	 * @param family service family ID to get name for
	 * @return family name as string, or <code>null</code> on no name available
	 */
	public final String getFamilyName(short family)
	{
		return family < familyNames.length ? familyNames[family] : null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.knxnetip.util.DIB#toByteArray()
	 */
	public byte[] toByteArray()
	{
		final byte[] buf = super.toByteArray();
		int i = 2;
		for (final Iterator it = map.entrySet().iterator(); it.hasNext();) {
			final Map.Entry e = (Map.Entry) it.next();
			buf[i++] = ((Short) e.getKey()).byteValue();
			buf[i++] = ((Short) e.getValue()).byteValue();
		}
		return buf;
	}
}
