/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: TunnelCRD.java,v $
 Revision 1.1  2007/07/20 00:52:51  bm
 added API specification comments

 */

package tuwien.auto.calimero.knxnetip.util;

import tuwien.auto.calimero.IndividualAddress;
import tuwien.auto.calimero.exception.KNXFormatException;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;
import tuwien.auto.calimero.knxnetip.KNXnetIPTunnel;

/**
 * Connection response data used for KNX tunneling connection.
 * <p>
 * 
 * @author bm
 */
public class TunnelCRD extends CRD
{
	/**
	 * Creates a new CRD for tunnel connection type out of a byte array.
	 * <p>
	 * The CRD structure has a length of 4 bytes.<br>
	 * 
	 * @param data byte array containing a CRD structure,
	 *        <code>data.length - offset = 4</code>
	 * @param offset start offset of CRD in <code>data</code>
	 * @throws KNXFormatException if no CRD found or invalid structure
	 */
	public TunnelCRD(byte[] data, int offset) throws KNXFormatException
	{
		super(data, offset);
		if (getConnectionType() != KNXnetIPTunnel.TUNNEL_CONNECTION)
			throw new KNXFormatException("not a tunneling CRD", getConnectionType());
		if (getStructLength() != 4)
			throw new KNXFormatException("wrong length for tunneling CRD");
	}

	/**
	 * Creates a new CRD for tunnel connection type containing the given address
	 * information.
	 * <p>
	 * 
	 * @param addr individual address assigned to the tunneling connection
	 */
	public TunnelCRD(IndividualAddress addr)
	{
		super(KNXnetIPTunnel.TUNNEL_CONNECTION, addr.toByteArray());
	}

	/**
	 * Creates a CRD for tunnel connection type containing optional data.
	 * <p>
	 * Note, the optional data field contains the assigned individual address.<br>
	 * 
	 * @param optionalData byte array containing tunneling host protocol data, this
	 *        information is located starting at offset 2 in the CRD structure,
	 *        <code>optionalData.length</code> = 2
	 */
	TunnelCRD(byte[] optionalData)
	{
		super(KNXnetIPTunnel.TUNNEL_CONNECTION, (byte[]) optionalData.clone());
		if (getStructLength() != 4)
			throw new KNXIllegalArgumentException("wrong length for tunneling CRD");
	}

	/**
	 * Returns the assigned address for the tunneling connection.
	 * <p>
	 * 
	 * @return individual address as {@link IndividualAddress}
	 */
	public final IndividualAddress getAssignedAddress()
	{
		return new IndividualAddress(opt);
	}
}
