/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log$
 */

package tuwien.auto.calimero.link;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import tuwien.auto.calimero.CloseEvent;
import tuwien.auto.calimero.FrameEvent;
import tuwien.auto.calimero.KNXListener;
import tuwien.auto.calimero.knxnetip.ConnectionCloseEvent;
import tuwien.auto.calimero.link.event.LinkListener;
import tuwien.auto.calimero.link.event.NetworkLinkListener;
import tuwien.auto.calimero.log.LogService;

/**
 * Threaded event notifier for network link and monitor.
 * <p>
 * 
 * @author bm
 */
abstract class EventNotifier extends Thread implements KNXListener
{
	private static interface EventCallback
	{
		void invoke(LinkListener l);
	}

	static final class Indication implements EventCallback
	{
		private final FrameEvent event;

		Indication(FrameEvent e)
		{
			event = e;
		}

		public void invoke(LinkListener l)
		{
			l.indication(event);
		}
	}

	static final class Confirmation implements EventCallback
	{
		private final FrameEvent event;

		Confirmation(FrameEvent e)
		{
			event = e;
		}

		public void invoke(LinkListener l)
		{
			((NetworkLinkListener) l).confirmation(event);
		}
	}

	static final class Closed implements EventCallback
	{
		private final CloseEvent event;

		Closed(CloseEvent e)
		{
			event = e;
		}

		public void invoke(LinkListener l)
		{
			l.linkClosed(event);
		}
	}

	final LogService logger;
	final Object source;

	// event listeners
	private final List listeners = new ArrayList();
	private List listenersCopy = new ArrayList();

	private final List events = new LinkedList();
	private volatile boolean stop;

	EventNotifier(Object source, LogService logger)
	{
		super("Link notifier");
		this.logger = logger;
		this.source = source;
		setDaemon(true);
		start();
	}

	final void quit()
	{
		if (stop)
			return;
		stop = true;
		interrupt();
		if (currentThread() != this) {
			try {
				join();
			}
			catch (final InterruptedException e) {}
		}
	}

	final void addEvent(EventCallback ec)
	{
		if (!stop) {
			synchronized (events) {
				events.add(ec);
				events.notify();
			}
		}
	}

	final void addListener(LinkListener l)
	{
		if (stop || l == null)
			return;
		synchronized (listeners) {
			if (!listeners.contains(l)) {
				listeners.add(l);
				listenersCopy = new ArrayList(listeners);
			}
			else
				logger.warn("event listener already registered");
		}
	}

	final void removeListener(LinkListener l)
	{
		synchronized (listeners) {
			if (listeners.remove(l))
				listenersCopy = new ArrayList(listeners);
		}
	}

	public final void run()
	{
		while (!stop) {
			try {
				EventCallback ec;
				synchronized (events) {
					while (events.isEmpty())
						events.wait();
					ec = (EventCallback) events.remove(0);
				}
				fire(ec);
			}
			catch (final InterruptedException ignore) {}
		}
		// empty event queue
		synchronized (events) {
			while (!events.isEmpty())
				fire((EventCallback) events.remove(0));
		}
	}

	public abstract void frameReceived(FrameEvent e);

	public void connectionClosed(ConnectionCloseEvent e)
	{
		final boolean cr = e.getInitiator() == ConnectionCloseEvent.SERVER ? false : true;
		addEvent(new Closed(new CloseEvent(source, cr, e.getReason())));
	}

	private void fire(EventCallback ec)
	{
		for (final Iterator i = listenersCopy.iterator(); i.hasNext();) {
			final LinkListener l = (LinkListener) i.next();
			try {
				ec.invoke(l);
			}
			catch (final RuntimeException rte) {
				removeListener(l);
				logger.error("removed event listener", rte);
			}
		}
	}
};
