/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: KNXNetworkLink.java,v $
 Revision 1.1  2007/03/02 15:13:18  bm
 introduced osi layer interfaces

 */

package tuwien.auto.calimero.link;

import tuwien.auto.calimero.KNXAddress;
import tuwien.auto.calimero.Priority;
import tuwien.auto.calimero.cemi.CEMILData;
import tuwien.auto.calimero.exception.KNXTimeoutException;
import tuwien.auto.calimero.knxnetip.KNXnetIPConnection;
import tuwien.auto.calimero.link.event.NetworkLinkListener;
import tuwien.auto.calimero.link.medium.KNXMediumSettings;
import tuwien.auto.calimero.log.LogManager;

/**
 * KNX data link layer, corresponds to layer 2 in the ISO OSI model.
 * <p>
 * This layer enables transparency of the type of connection used to join the KNX network,
 * and an abstraction of the particular KNX communication medium used (e.g. TP1).<br>
 * It's communication is done through cEMI frames. The data link layer relies on an
 * underlying intermediate connection technology and protocol (e.g. KNXnet/IP,
 * {@link KNXnetIPConnection}) to access KNX communication.<br>
 * 
 * @author bm
 */
public interface KNXNetworkLink
{
	/**
	 * Configures and supplies information necessary for KNX communication.
	 * <p>
	 * These informations are differing between KNX media and depend on the KNX network
	 * this link is communicating with.<br>
	 * A new <code>settings</code> object type has to match the medium type supplied to
	 * the opened link in the first place.<br>
	 * The <code>settings</code> object is not copied internally to allow changes to
	 * settings by the user which should take effect immediately.
	 * 
	 * @param settings medium settings to use, the expected subtype is according to the
	 *        KNX network medium
	 */
	void setKNXMedium(KNXMediumSettings settings);

	/**
	 * Returns the KNX medium settings used by this network link.
	 * <p>
	 * The returned object is a reference to the one used by this link and not a copy.
	 * 
	 * @return medium settings for KNX network
	 */
	KNXMediumSettings getKNXMedium();

	/**
	 * Adds the specified event listener <code>l</code> to receive events from this
	 * {@link KNXNetworkLink}.
	 * <p>
	 * If <code>l</code> was already added as listener, no action is performed.
	 * 
	 * @param l the listener to add
	 */
	void addLinkListener(NetworkLinkListener l);

	/**
	 * Removes the specified event listener <code>l</code>, so it does no longer
	 * receive events from this {@link KNXNetworkLink}.
	 * <p>
	 * If <code>l</code> was not added in the first place, no action is performed.
	 * 
	 * @param l the listener to remove
	 */
	void removeLinkListener(NetworkLinkListener l);

	/**
	 * Sets the hop count used as default in KNX frames.
	 * <p>
	 * It denotes how many sub networks a frame is allowed to travel.<br>
	 * A frames hop count is decremented by KNX routers to limit distance and avoid
	 * looping. On hop count value 0, the frame is discarded from the network. A hop count
	 * of 7 never gets decremented.<br>
	 * By default, a hop count of 6 is specified.
	 * <p>
	 * 
	 * @param count hop count value, 0 &lt;= value &lt;= 7
	 */
	void setHopCount(int count);

	/**
	 * Returns the hop count used as default for KNX frames.
	 * <p>
	 * 
	 * @return hop count as 3 bit unsigned value with the range 0 to 7
	 * @see #setHopCount(int)
	 */
	byte getHopCount();

	/**
	 * Sends a link layer request frame to the given destination.
	 * <p>
	 * 
	 * @param dst KNX destination address
	 * @param p priority this frame is assigned to
	 * @param nsdu link layer service data unit
	 * @throws KNXLinkClosedException
	 * @throws KNXTimeoutException
	 */
	void sendRequest(KNXAddress dst, Priority p, byte[] nsdu)
		throws KNXLinkClosedException, KNXTimeoutException;

	/**
	 * @param msg
	 * @throws KNXLinkClosedException
	 * @throws KNXTimeoutException
	 */
	void sendRequest(CEMILData msg) throws KNXLinkClosedException, KNXTimeoutException;

	/**
	 * Sends a link layer request frame to the given destination, and waits for the
	 * corresponding link layer confirmation.
	 * <p>
	 * Depending on the address, the request is either point-to-point, multicast or
	 * broadcast.
	 * 
	 * @param dst KNX destination address
	 * @param p priority this frame is assigned to
	 * @param nsdu link layer service data unit
	 * @throws KNXTimeoutException on timeout while waiting for the confirmation to the
	 *         sent request
	 * @throws KNXLinkClosedException
	 */
	void sendRequestWait(KNXAddress dst, Priority p, byte[] nsdu)
		throws KNXTimeoutException, KNXLinkClosedException;

	/**
	 * @param msg
	 * @throws KNXTimeoutException
	 * @throws KNXLinkClosedException
	 */
	void sendRequestWait(CEMILData msg) throws KNXTimeoutException,
		KNXLinkClosedException;

	/**
	 * Returns the name of the link, a short textual representation to identify a link.
	 * <p>
	 * The name is unique for links with different remote endpoints.<br>
	 * The returned name is used by this link for the name of its log service. Supply
	 * {@link #getName()} on {@link LogManager#getLogger(String)} for example to
	 * get the log service of this link.
	 * <p>
	 * By default, "Link " + address/ID of the remote control endpoint is returned (e.g.
	 * "Link 192.168.0.10:3671" for IP link).<br>
	 * After closing the link, the name might differ, e.g. get reset to some arbitrary
	 * name.
	 * 
	 * @return link name as string
	 */
	String getName();

	/**
	 * Ends communication with the KNX network.
	 * <p>
	 * All registered event listeners get notified.<br>
	 * If no communication access was established in the first place, no action is
	 * performed.
	 */
	void close();

	/**
	 * Checks for open network link for communication.
	 * <p>
	 * After a call to {@link #close()} or after the underlying protocol initiated the end
	 * of the communication, this method always returns <code>false</code>.
	 * 
	 * @return <code>true</code> if this network link is open, <code>false</code> on
	 *         closed
	 */
	boolean isOpen();
}
