/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: KNXNetworkLinkIP.java,v $
 Revision 1.1  2007/09/03 17:53:53  bm
 initial commit, cache access interface implemented

 Revision 1.1  2007/03/18 11:38:17  bm
 added class for implementation of datalink layer

 */

package tuwien.auto.calimero.link;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.UnknownHostException;

import tuwien.auto.calimero.FrameEvent;
import tuwien.auto.calimero.IndividualAddress;
import tuwien.auto.calimero.KNXAddress;
import tuwien.auto.calimero.Priority;
import tuwien.auto.calimero.cemi.CEMILData;
import tuwien.auto.calimero.cemi.CEMILDataEx;
import tuwien.auto.calimero.exception.KNXException;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;
import tuwien.auto.calimero.exception.KNXTimeoutException;
import tuwien.auto.calimero.knxnetip.ConnectionCloseEvent;
import tuwien.auto.calimero.knxnetip.KNXConnectionClosedException;
import tuwien.auto.calimero.knxnetip.KNXnetIPConnection;
import tuwien.auto.calimero.knxnetip.KNXnetIPRouter;
import tuwien.auto.calimero.knxnetip.KNXnetIPTunnel;
import tuwien.auto.calimero.link.event.NetworkLinkListener;
import tuwien.auto.calimero.link.medium.KNXMediumSettings;
import tuwien.auto.calimero.link.medium.PLSettings;
import tuwien.auto.calimero.link.medium.RFSettings;
import tuwien.auto.calimero.link.medium.TPSettings;
import tuwien.auto.calimero.log.LogManager;
import tuwien.auto.calimero.log.LogService;

/**
 * Implementation of {@link KNXNetworkLink}, with support for KNX frame caching.
 * <p>
 * Caching is an optional feature and not enabled by default, i.e. no caching activity.
 * <br>
 * The cache is used for L-data request and indication frames only. On occurrence of such
 * a L-Data frame, a factory is invoked for the user's opportunity to filter particular
 * frames and put them into the cache.<br>
 * The default cache object factory used by this layer filters for frames containing
 * application group object values. The default cache strategy is a LFU cache.
 * 
 * @author bm
 */
public class KNXNetworkLinkIP implements KNXNetworkLink
{
	/**
	 * Service mode for link layer tunneling.
	 * <p>
	 */
	public static final int TUNNEL = 1;

	/**
	 * Service mode for routing.
	 * <p>
	 */
	public static final int ROUTER = 2;

	private static final class LinkNotifier extends EventNotifier
	{
		LinkNotifier(Object source, LogService logger)
		{
			super(source, logger);
		}

		public void frameReceived(FrameEvent e)
		{
			final int mc = e.getFrame().getMessageCode();
			if (mc == CEMILData.MC_LDATA_IND)
				addEvent(new Indication(new FrameEvent(source, e.getFrame())));
			else if (mc == CEMILData.MC_LDATA_CON)
				addEvent(new Confirmation(new FrameEvent(source, e.getFrame())));
			else
				logger.warn("unspecified frame event - ignored, msg code = " + mc);
		}

		public void connectionClosed(ConnectionCloseEvent e)
		{
			((KNXNetworkLinkIP) source).closed = true;
			super.connectionClosed(e);
		}
	};

	private final int mode;
	private volatile boolean closed;
	private final KNXnetIPConnection conn;
	private volatile byte hopCount = 6;
	private KNXMediumSettings medium;

	private final LogService logger;
	// our link connection event notifier
	private final EventNotifier notifier;

	/**
	 * Opens a new data link layer based on the KNXnet/IP protocol, using a
	 * {@link KNXnetIPConnection}.
	 * <p>
	 * For more details on KNXnet/IP connections, refer to the various implementations.<br>
	 * 
	 * @param serviceMode mode of communication to open, <code>serviceMode</code> is one
	 *        of the service mode constants (e.g. {@link #TUNNEL}); depending on the mode
	 *        set, the local / remote endpoint values expected might differ
	 * @param localEP the local endpoint of the data link layer to use;<br> - in a mode
	 *        with logical connection (point-to-point), this is the client control
	 *        endpoint, use <code>null</code> for the default local host and an
	 *        ephemeral port number<br> - in {@link #ROUTER} mode, specifies the
	 *        multicast interface, i.e. the local network interface is taken that has the
	 *        IP address bound to it (if IP address is bound more than once, it's
	 *        undefined which interface is returned), the port is not used; use
	 *        <code>null</code> for <code>localEP</code> or an unresolved IP address
	 *        to take the default multicast interface
	 * @param remoteEP the remote endpoint of the data link layer to communicate;<br> -
	 *        in a mode with logical connection (point-to-point), this is the server
	 *        control endpoint <br> - in {@link #ROUTER} mode, the IP address specifies
	 *        the multicast group to join, the port is not used; use <code>null</code>
	 *        for <code>remoteEP</code> or an unresolved IP address to take the default
	 *        multicast group
	 * @param useNAT <code>true</code> to consider network address translation in
	 *        tunneling service mode, <code>false</code> to use the default (non aware)
	 *        mode; parameter is ignored in other service modes
	 * @param settings medium settings defining the specific KNX medium, and needed for
	 *        communication
	 * @throws KNXException on failure establishing KNXnet/IP connection
	 */
	public KNXNetworkLinkIP(int serviceMode, InetSocketAddress localEP,
		InetSocketAddress remoteEP, boolean useNAT, KNXMediumSettings settings)
		throws KNXException
	{
		switch (serviceMode) {
		case TUNNEL:
			InetSocketAddress local = localEP;
			if (local == null)
				try {
					local = new InetSocketAddress(InetAddress.getLocalHost(), 0);
				}
				catch (final UnknownHostException e) {
					throw new KNXException("no local host available");
				}
			conn = new KNXnetIPTunnel(KNXnetIPTunnel.LINK_LAYER, local, remoteEP, useNAT);
			break;
		case ROUTER:
			NetworkInterface netIf = null;
			if (localEP != null && !localEP.isUnresolved())
				try {
					netIf = NetworkInterface.getByInetAddress(localEP.getAddress());
				}
				catch (final SocketException e) {
					throw new KNXException("error getting network interface: "
						+ e.getMessage());
				}
			final InetAddress mcast = remoteEP != null ? remoteEP.getAddress() : null;
			conn = new KNXnetIPRouter(netIf, mcast);
			break;
		default:
			throw new KNXIllegalArgumentException("unknown service mode");
		}
		// initialize our link with opened connection
		mode = serviceMode;
		logger = LogManager.getManager().getLogger(getName());
		notifier = new LinkNotifier(this, logger);
		conn.addConnectionListener(notifier);
		// configure KNX medium stuff
		setKNXMedium(settings);
	}

	// softy ctor: tunnel mode, default local host, default remote KNXnet/IP port, no NAT
	public KNXNetworkLinkIP(String remoteHost, KNXMediumSettings settings)
		throws KNXException
	{
		this(TUNNEL, null, new InetSocketAddress(remoteHost, KNXnetIPConnection.IP_PORT),
			false, settings);
	}

	public KNXNetworkLinkIP(NetworkInterface netIf, InetAddress mcGroup,
		KNXMediumSettings settings) throws KNXException
	{
		conn = new KNXnetIPRouter(netIf, mcGroup);
		// initialize our link
		mode = ROUTER;
		logger = LogManager.getManager().getLogger(getName());
		notifier = new LinkNotifier(this, logger);
		conn.addConnectionListener(notifier);
		// configure KNX medium stuff
		setKNXMedium(settings);
	}

	public void addLinkListener(NetworkLinkListener l)
	{
		notifier.addListener(l);
	}

	public void setKNXMedium(KNXMediumSettings settings)
	{
		if (settings == null)
			throw new KNXIllegalArgumentException("medium settings are mandatory");
		if (medium != null && !settings.getClass().isAssignableFrom(medium.getClass())
			&& !medium.getClass().isAssignableFrom(settings.getClass()))
			throw new KNXIllegalArgumentException("medium differs");
		medium = settings;
	}

	public KNXMediumSettings getKNXMedium()
	{
		return medium;
	}

	/**
	 * {@inheritDoc}<br>
	 * Note, that a cache set up before closing still stays available.
	 */
	public synchronized void close()
	{
		if (!closed) {
			closed = true;
			conn.close();
			notifier.quit();
		}
	}

	public final byte getHopCount()
	{
		return hopCount;
	}

	public void removeLinkListener(NetworkLinkListener l)
	{
		notifier.removeListener(l);
	}

	public void sendRequest(KNXAddress dst, Priority p, byte[] nsdu)
		throws KNXLinkClosedException, KNXTimeoutException
	{
		send(dst, p, nsdu, false);
	}

	public void sendRequest(CEMILData msg) throws KNXLinkClosedException,
		KNXTimeoutException
	{
		send(adjustMsgType(msg), false);
	}

	public void sendRequestWait(KNXAddress dst, Priority p, byte[] nsdu)
		throws KNXTimeoutException, KNXLinkClosedException
	{
		send(dst, p, nsdu, true);
	}

	public void sendRequestWait(CEMILData msg) throws KNXTimeoutException,
		KNXLinkClosedException
	{
		send(adjustMsgType(msg), true);
	}

	public final void setHopCount(int count)
	{
		if (count < 0 || count > 7)
			throw new KNXIllegalArgumentException("hop count out of range [0..7]");
		hopCount = (byte) count;
	}

	public String getName()
	{
		// do our own IP:port string, since InetAddress.toString() always prepends a '/'
		final InetSocketAddress a = conn.getRemoteAddress();
		return "Link " + a.getAddress().getHostAddress() + ":" + a.getPort();
	}

	private void send(KNXAddress dst, Priority p, byte[] nsdu, boolean confirm)
		throws KNXTimeoutException, KNXLinkClosedException
	{
		final CEMILData f;
		final short mc = mode == TUNNEL ? CEMILData.MC_LDATA_REQ : CEMILData.MC_LDATA_IND;
		final IndividualAddress src = medium.getDeviceAddress();
		if (nsdu.length <= 16 && medium instanceof TPSettings)
			f = new CEMILData(mc, src, dst, nsdu, p, true, hopCount);
		else
			f = new CEMILDataEx(mc, src, dst, nsdu, p, true, hopCount);
		send(f, confirm);
	}

	private void send(CEMILData msg, boolean confirm) throws KNXTimeoutException,
		KNXLinkClosedException
	{
		if (closed)
			throw new KNXLinkClosedException("link closed");
		if (medium instanceof PLSettings) {
			final CEMILDataEx f = (CEMILDataEx) msg;
			if (f.getAdditionalInfo(CEMILDataEx.ADDINFO_PLMEDIUM) == null)
				f.addAdditionalInfo(CEMILDataEx.ADDINFO_PLMEDIUM, ((PLSettings) medium)
					.getDomainAddress());
		}
		else if (medium.getMedium() == KNXMediumSettings.MEDIUM_RF) {
			final CEMILDataEx f = (CEMILDataEx) msg;
			final RFSettings rf = (RFSettings) medium;
			if (f.getAdditionalInfo(CEMILDataEx.ADDINFO_RFMEDIUM) == null) {
				final byte[] sn =
					f.isDomainBroadcast() ? rf.getDomainAddress() : rf.getSerialNumber();
				// add-info: rf-info (0 ignores a lot), sn (6 bytes), lfn (=255:void)
				f.addAdditionalInfo(CEMILDataEx.ADDINFO_RFMEDIUM, new byte[] { 0, sn[0],
					sn[1], sn[2], sn[3], sn[4], sn[5], (byte) 0xff });
			}
		}
		try {
			conn.send(msg, confirm ? KNXnetIPConnection.WAIT_FOR_CON
				: KNXnetIPConnection.WAIT_FOR_ACK);
		}
		catch (final KNXConnectionClosedException e) {
			close();
			throw new KNXLinkClosedException("link closed, " + e.getMessage());
		}
	}

	private CEMILData adjustMsgType(CEMILData msg)
	{
		if (medium instanceof TPSettings || msg instanceof CEMILDataEx)
			return msg;
		return new CEMILDataEx(msg.getMessageCode(), msg.getSource(), msg
			.getDestination(), msg.getTPDU(), msg.getPriority(), msg.isRepetition(),
			true, msg.isAckRequested(), msg.getHopCount());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.link.KNXNetworkLink#isOpen()
	 */
	public boolean isOpen()
	{
		return !closed;
	}
}
