/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: ErrorHandler.java,v $
 Revision 1.1  2007/08/06 18:15:23  bm
 added log error handler

 */

package tuwien.auto.calimero.log;

/**
 * An error handler is invoked on problems of a log writer not handled by the log writer
 * itself.
 * <p>
 * Such problems shouldn't get logged using another log writer, to prevent the possibility
 * of a fault / failure chain, and shouldn't lead to disruption of the log system at a
 * whole.<br>
 * Moreover, log writer errors most certainly can't be returned back to the owner of the
 * writer responsible for handling the errors. And handling errors directly in log
 * services can rarely be done in a satisfying way.<br>
 * So the associated error handler is used for such error events.<br>
 * The handler itself might be adapted to the application needs. For example, an
 * overridden {@link #error(LogWriter, String, Exception)} might close the log writer the
 * error originated from.
 * 
 * @author bm
 * @see LogWriter
 */
public class ErrorHandler
{
	/**
	 * Set the maximum number {@link #error(LogWriter, String, Exception)} should handle
	 * error events. This should be set reasonably low to prevent a continued handling of
	 * repeated errors of the same source.
	 */
	protected int maxInvocations = 5;

	/**
	 * Counter how often error() was called (without abort).
	 */
	protected int invocations;

	/**
	 * Creates a new error handler.
	 */
	public ErrorHandler()
	{}

	/**
	 * Invoked on a log writer error event.
	 * <p>
	 * The default behavior used here prints the logging source and the error message to
	 * the system standard error stream (System.err). If an exception object is given, the
	 * exception and the last method calls of the log writer leading to the error are also
	 * printed.<br>
	 * Only predefined maximum invocations are handled, subsequent calls are ignored.
	 * 
	 * @param source the log writer the error originated from
	 * @param msg message describing the error
	 * @param e exception related to the error, may be <code>null</code>
	 */
	public synchronized void error(LogWriter source, String msg, Exception e)
	{
		if (invocations >= maxInvocations)
			return;
		++invocations;

		String out = source + ": " + msg;
		StackTraceElement[] trace = null;
		if (e != null) {
			out += " (" + e.getMessage() + ")";
			trace = e.getStackTrace();
		}
		synchronized (System.err) {
			System.err.println(out);
			final String srcName = source.getClass().getName();
			if (trace != null)
				for (int i = 0; i < trace.length; ++i) {
					System.err.println("\t- " + trace[i]);
					if (trace[i].getClassName().equals(srcName))
						break;
				}
		}
	}
}
