/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: LogLevel.java,v $
 Revision 1.1  2007/07/15 00:59:13  bm
 initial commit after package renaming

 Revision 1.2  2007/04/13 13:18:22  bm
 added API specification comments
 code cleanup

 Revision 1.1  2007/03/01 19:10:23  bm
 adding package with logging functionality

 */

package tuwien.auto.calimero.log;

/**
 * Defines different log levels.
 * <p>
 * Defines several severity levels used to categorize logging information and
 * loggers. It also gives a clue about the chattiness of the respective object.
 * 
 * @author bm
 */
public class LogLevel
{ 
	private static final int LOG_OFF = 0;
	private static final int LOG_FATAL = 100;
	private static final int LOG_ERROR = 200;
	private static final int LOG_WARN = 300;
	private static final int LOG_INFO = 400;
	private static final int LOG_TRACE = 500;
	private static final int LOG_ALL = 1000;	

	/**
	 * The lowest log level, logging is turned off.
	 */
	public static final LogLevel OFF = new LogLevel(LOG_OFF);	
	/**
	 * This level identifies the attached object for fatal errors purposes.
	 */
	public static final LogLevel FATAL = new LogLevel(LOG_FATAL);
	/**
	 * This level identifies the attached object for errors purposes.
	 */
	public static final LogLevel ERROR = new LogLevel(LOG_ERROR);
	/**
	 * This level identifies the attached object for warning purposes.
	 */
	public static final LogLevel WARN = new LogLevel(LOG_WARN);
	/**
	 * This level identifies the attached object for informational purposes.
	 */
	public static final LogLevel INFO = new LogLevel(LOG_INFO);	
	/**
	 * This level identifies the attached object for tracing purposes.
	 */
	public static final LogLevel TRACE = new LogLevel(LOG_TRACE);
	/**
	 * The highest log level, no restrictions, everything will be logged.
	 */
	public static final LogLevel ALL = new LogLevel(LOG_ALL);

	final int level;

	/**
	 * Creates a new LogLevel object with <code>logLevel</code> assigned to it. 
	 * @param logLevel log level represented by this object 
	 */
	protected LogLevel(int logLevel)
	{
		level = logLevel;
	}

	/**
	 * Compares the log level with <code>obj</code>.
	 * <p>
	 * 
	 * @param obj {@inheritDoc}
	 * @return <code>true</code> if <code>o</code> is of type
	 *         <code>LogLevel</code> and has the same log level as this object,
	 *         <code>false</code> otherwise.
	 */
	public boolean equals(Object obj)
	{
		if (obj instanceof LogLevel)
			return ((LogLevel) obj).level == level;
		return false;
	}

	/**
	 * Compares for a higher log level than <code>l</code>.
	 * <p>
	 * 
	 * @param l level to compare
	 * @return true iff this object has a higher log level than <code>l</code>.
	 */
	public final boolean higher(LogLevel l)
	{
		return level > l.level;
	}

	/**
	 * Returns the log level in textual format, like "off", "trace",
	 * .&nbsp;.&nbsp;.&nbsp; or "unknown" if level is not defined.
	 * 
	 * @return the log level as string
	 */
	public String toString()
	{
		switch (level) {
		case LOG_OFF:
			return "off";
		case LOG_TRACE:
			return "trace";
		case LOG_INFO:
			return "info";
		case LOG_WARN:
			return "warn";
		case LOG_ERROR:
			return "error";
		case LOG_FATAL:
			return "fatal";
		case LOG_ALL:
			return "all";
		default:
			return "unknown";
		}
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode()
	{
		final int offset = 435273;
		return level + offset;
	}
}
