/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: LogNetWriter.java,v $
 Revision 1.4  2007/08/06 18:15:23  bm
 added log error handler

 Revision 1.3  2007/07/15 00:59:12  bm
 initial commit after package renaming

 Revision 1.2  2007/04/13 13:18:22  bm
 added API specification comments
 code cleanup

 Revision 1.1  2007/03/10 18:11:14  bm
 added LogNetWriter for logging over a network connection

 */

package tuwien.auto.calimero.log;

import java.io.IOException;
import java.net.Socket;


/**
 * A LogNetWriter is used to log information over a socket connection.
 * <p>
 * A destination host is supplied on creation of this log writer, a
 * {@link Socket} TCP connection is opened and used for further logging. After
 * {@link #close()}ing the log writer, it cannot be connected again.<br>
 * 
 * @author bm
 * @see Socket
 */
public class LogNetWriter extends LogStreamWriter
{
	/**
	 * Socket connection used as logging destination.<p>
	 */
	protected Socket s;

	/**
	 * Creates a log writer and open a socket connection to destination
	 * <code>host</code> and <code>port</code>.
	 * <p>
	 * 
	 * @param host destination host name; if <code>null</code> or an empty
	 *        string is specified, an address of the loopback interface is used
	 * @param port destination port
	 * @throws KNXLogException if IP address of host could not be determined or
	 *         socket binding / connecting failed
	 * @see Socket
	 */
	public LogNetWriter(String host, int port) throws KNXLogException
	{
		try {
			s = new Socket(host, port);
			createWriter(s.getOutputStream());
		}
		catch (final IOException e) {
			throw new KNXLogException(e.getMessage());
		}
	}

	/**
	 * Like {@link #LogNetWriter(String, int)} with the ability to set the
	 * filter log level for information logged by LogNetWriter.
	 * <p>
	 * 
	 * @param level log level used by this LogWriter to filter log information
	 * @param host destination host name; if <code>null</code> or an empty
	 *        string is specified, an address of the loopback interface is used
	 * @param port destination port
	 * @throws KNXLogException if IP address of host could not be determined or
	 *         socket binding / connecting failed
	 */
	public LogNetWriter(LogLevel level, String host, int port) throws KNXLogException
	{
		this(host, port);
		setLogLevel(level);
	}

	/* (non-Javadoc)
	 * @see tuwien.auto.calimero.log.LogStreamWriter#close()
	 */
	public synchronized void close()
	{
		if (s != null) {
			super.close();
			try {
				s.close();
			}
			catch (final IOException e) {}
			s = null;
		}
	}

	/**
	 * Returns the host IP address of this connection, or "" if the connection
	 * was closed.
	 * <p>
	 * 
	 * @return IP address as String
	 * @see java.net.InetAddress#getHostAddress()
	 */
	public final String getHostAddress()
	{
		final Socket socket = s;
		return socket == null ? "" : socket.getInetAddress().getHostAddress();
	}

	/**
	 * Returns the host name of this connection, or "" if the connection was
	 * closed.
	 * <p>
	 * Note that this might involve a reverse name lookup. It is possible that
	 * only the IP address string is returned.
	 * 
	 * @return host name (or IP address) as String
	 * @see java.net.InetAddress#getHostName()
	 */
	public final String getHostName()
	{
		final Socket socket = s;
		return socket == null ? "" : socket.getInetAddress().getHostName();
	}

	/**
	 * Returns the destination port of this connection, or 0 if the connection
	 * was closed.
	 * <p>
	 * 
	 * @return port number
	 */
	public final int getPort()
	{
		final Socket socket = s;
		return socket == null ? 0 : socket.getPort();
	}
}
