/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: LogWriter.java,v $
 Revision 1.4  2007/08/06 18:15:23  bm
 added log error handler

 Revision 1.3  2007/07/15 00:59:13  bm
 initial commit after package renaming

 Revision 1.2  2007/04/13 13:18:22  bm
 added API specification comments
 code cleanup

 Revision 1.1  2007/03/01 19:10:23  bm
 adding package with logging functionality

 */

package tuwien.auto.calimero.log;

/**
 * LogWriter is responsible for writing information to individual destinations.
 * <p>
 * A LogWriter has an assigned log level (by default {@link LogLevel#ALL}), which
 * specifies what information to write and what to ignore. A log level of
 * {@link LogLevel#OFF} will ignore any write requests. If a LogWriter is not used
 * anymore, {@link #close()} has to be called.
 * 
 * @author bm
 * @see LogLevel
 * @see LogService
 */
public abstract class LogWriter
{
	private static ErrorHandler errorHandler = new ErrorHandler();

	LogLevel logLevel = LogLevel.ALL;

	/**
	 * Sets the error handler used on all log writer errors.
	 * <p>
	 * 
	 * @param handler the new error handler
	 */
	public static synchronized void setErrorHandler(ErrorHandler handler)
	{
		if (handler != null)
			errorHandler = handler;
	}

	/**
	 * Returns the used error handler.
	 * <p>
	 * 
	 * @return an {@link ErrorHandler} object
	 */
	public static synchronized ErrorHandler getErrorHandler()
	{
		return errorHandler;
	}

	/**
	 * Returns the log level used for deciding which messages will be logged.
	 * <p>
	 * 
	 * @return log {@link LogLevel}
	 */
	public final LogLevel getLogLevel()
	{
		return logLevel;
	}

	/**
	 * Sets the log level for deciding which messages gets logged by this LogWriter.
	 * <p>
	 * 
	 * @param level new log level
	 */
	public final void setLogLevel(LogLevel level)
	{
		logLevel = level;
	}

	/**
	 * Writes a message out to this LogWriter.
	 * <p>
	 * The message has the associated log level <code>level</code>. It will only be
	 * written if the LogWriter logging level is not more restrictive than
	 * <code>level</code>. Otherwise the message is ignored. LogWriter is responsible
	 * for formatting the output.
	 * 
	 * @param logService log service name stating the source of the message
	 * @param level log level of message
	 * @param msg the message to write
	 */
	public abstract void write(String logService, LogLevel level, String msg);

	/**
	 * Like {@link #write(String, LogLevel, String)}, in addition a
	 * <code>Throwable</code> object is taken which will be added to the message.
	 * 
	 * @param logService log service name stating the source of the message
	 * @param level log level of message
	 * @param msg msg the message to write
	 * @param t Throwable object, might be <code>null</code>
	 */
	public abstract void write(String logService, LogLevel level, String msg, Throwable t);

	/**
	 * Flushes all buffered output.
	 */
	public abstract void flush();

	/**
	 * Closes the LogWriter and all its resources.
	 * <p>
	 * If necessary, all output is flushes before.
	 */
	public abstract void close();
}
