/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log$
 */

package tuwien.auto.calimero.mgmt;

import tuwien.auto.calimero.IndividualAddress;
import tuwien.auto.calimero.exception.KNXIllegalStateException;
import tuwien.auto.calimero.link.KNXLinkClosedException;

/**
 * @author bm
 */
// models state of a transport layer logical connection
public class Destination
{
	public static final class AggregatorProxy
	{
		private final TransportLayer aggr;
		private Destination d;

		/**
		 * 
		 */
		public AggregatorProxy(TransportLayer aggregator)
		{
			aggr = aggregator;
		}

		public Destination getDestination()
		{
			return d;
		}

		public synchronized int getSeqReceive()
		{
			return d.seqRcv;
		}

		public synchronized int getSeqSend()
		{
			return d.seqSend;
		}

		public synchronized void incSeqReceive()
		{
			d.seqRcv = ++d.seqRcv & 0x0F;
		}

		public synchronized void incSeqSend()
		{
			d.seqSend = ++d.seqSend & 0x0f;
		}

		public void restartTimeout()
		{
			d.restartTimer();
		}

		public void setState(byte newState)
		{
			d.setState(newState);
		}

		void setDestination(Destination d)
		{
			this.d = d;
		}
	}

	private final class ConnectionTimeout extends Thread
	{
		// idle timeout for connection in seconds
		private static final byte TIMEOUT = 6;
		private boolean stop;
		private boolean dormant;
		private boolean restart;

		public ConnectionTimeout()
		{
			super("Destination timeout");
			setDaemon(true);
			start();
		}

		public synchronized void quit()
		{
			stop = true;
			notify();
		}

		public synchronized void restart()
		{
			dormant = false;
			restart = true;
			notify();
		}

		public synchronized void dormant()
		{
			dormant = true;
			notify();
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Thread#run()
		 */
		public synchronized void run()
		{
			long remaining = 0;
			long end = 0;
			restart = true;
			while (!stop) {
				if (dormant) {
					doWait(0);
				}
				else if (restart) {
					restart = false;
					remaining = TIMEOUT * 1000;
					end = System.currentTimeMillis() + remaining;
				}
				else if (remaining <= 0) {
					dormant = true;
					connectionTimeout();
				}
				else {
					doWait(remaining);
					remaining = end - System.currentTimeMillis();
				}
			}
		}

		private void doWait(long timeout)
		{
			try {
				wait(timeout);
			}
			catch (final InterruptedException e) {}
		}
	}

	public static final byte DESTROYED = 0;
	public static final byte DISCONNECTED = 1;
	public static final byte CONNECTING = 2;
	public static final byte OPEN_IDLE = 3;
	public static final byte OPEN_WAIT = 4;

	private final TransportLayer tl;
	private final IndividualAddress addr;
	private volatile byte state = DISCONNECTED;
	private int seqRcv;
	private int seqSend;
	// we do lazy initialization on timer
	private ConnectionTimeout timer;
	private final boolean co;
	private final boolean alive;

	// see 03/05/01 Resources: Verify Mode Control
	private final boolean verify;

	// verify and keep alive default to false
	public Destination(AggregatorProxy aggregator, IndividualAddress remote,
		boolean connectionOriented)
	{
		this(aggregator, remote, connectionOriented, false, false);
	}

	public Destination(AggregatorProxy aggregator, IndividualAddress remote,
		boolean connectionOriented, boolean keepAlive, boolean verifyMode)
	{
		tl = aggregator.aggr;
		aggregator.setDestination(this);
		addr = remote;
		co = connectionOriented;
		alive = co ? keepAlive : false;
		verify = verifyMode;
	}

	/**
	 * 
	 */
	public synchronized void destroy()
	{
		if (state == DESTROYED)
			return;
		if (state != DISCONNECTED)
			try {
				tl.disconnect(this);
			}
			catch (final KNXLinkClosedException e) {
				// we already should've been destroyed on catching this exception
			}
		setState(DESTROYED);
		tl.destroyDestination(this);
	}

	/**
	 * @return the connection
	 */
	public IndividualAddress getAddress()
	{
		return addr;
	}

	/**
	 * @return the connection state
	 */
	public final byte getState()
	{
		return state;
	}

	public final boolean isConnectionOriented()
	{
		return co;
	}

	public final boolean isKeepAlive()
	{
		return alive;
	}

	public final boolean isVerifyMode()
	{
		return verify;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	public String toString()
	{
		final String s =
			"destination " + addr + " (" + tl.getName() + ") state=" + getStateString();
		if (state == DESTROYED)
			return s;
		return s + " conn.oriented=" + co + " keep alive=" + alive + " verify mode="
			+ verify;
	}

	private void connectionTimeout()
	{
		if (state == DISCONNECTED || state == DESTROYED)
			return;
		try {
			tl.disconnect(this);
		}
		catch (final KNXLinkClosedException e) {
			// we already should've been destroyed on catching this exception
		}
	}

	private String getStateString()
	{
		switch (state) {
		case DISCONNECTED:
			return "disconnected";
		case CONNECTING:
			return "connecting";
		case OPEN_IDLE:
			return "open idle";
		case OPEN_WAIT:
			return "open wait";
		case DESTROYED:
			return "destroyed";
		default:
			assert false;
		}
		return "unknown";
	}

	private synchronized void setState(byte newState)
	{
		if (state == DESTROYED)
			return;
		state = newState;
		if (state == CONNECTING) {
			seqSend = 0;
			seqRcv = 0;
		}
		else if (state == OPEN_IDLE) {
			restartTimer();
		}
		else if (state == OPEN_WAIT) {
			restartTimer();
		}
		else if (state == DISCONNECTED) {
			if (timer != null)
				timer.dormant();
		}
		else if (state == DESTROYED) {
			if (timer != null)
				timer.quit();
		}
	}

	private void restartTimer()
	{
		if (!co)
			throw new KNXIllegalStateException("no timer if not connection oriented");
		if (state == DESTROYED)
			return;
		if (timer != null)
			timer.restart();
		else
			timer = new ConnectionTimeout();
	}
}
