/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: KnIPDeviceMgmtAdapter.java,v $
 Revision 1.1  2007/08/17 22:34:06  bm
 initial commit

 */

package tuwien.auto.calimero.mgmt;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import tuwien.auto.calimero.CloseEvent;
import tuwien.auto.calimero.FrameEvent;
import tuwien.auto.calimero.KNXListener;
import tuwien.auto.calimero.cemi.CEMIDevMgmt;
import tuwien.auto.calimero.exception.KNXException;
import tuwien.auto.calimero.exception.KNXIllegalStateException;
import tuwien.auto.calimero.exception.KNXInvalidResponseException;
import tuwien.auto.calimero.exception.KNXRemoteException;
import tuwien.auto.calimero.exception.KNXTimeoutException;
import tuwien.auto.calimero.knxnetip.ConnectionCloseEvent;
import tuwien.auto.calimero.knxnetip.KNXConnectionClosedException;
import tuwien.auto.calimero.knxnetip.KNXnetIPConnection;
import tuwien.auto.calimero.knxnetip.KNXnetIPDevMgmt;

/**
 * The object instance used is always the first one, i.e. object instance 1.
 * 
 * @author bm
 */
public class KnIPDeviceMgmtAdapter implements PropertyAdapter
{
	private static final class Pair
	{
		final int oindex;
		final int otype;

		Pair(int objIndex, int objType)
		{
			oindex = objIndex;
			otype = objType;
		}
	}

	private final class KNXListenerImpl implements KNXListener
	{
		KNXListenerImpl()
		{}

		public void frameReceived(FrameEvent e)
		{
			frames.add(e.getFrame());
		}

		public void connectionClosed(ConnectionCloseEvent e)
		{
			final boolean user =
				serverReset ? false : e.getInitiator() == ConnectionCloseEvent.CLIENT
					? true : false;
			final String reason = serverReset ? "reset by server" : "requested by user";
			listener.adapterClosed(new CloseEvent(KnIPDeviceMgmtAdapter.this, user,
				reason));
		}
	}

	private static final int DEVICE_OBJECT = 0;
	private static final int PID_IO_LIST = 71;

	private final KNXnetIPConnection conn;
	private final PropertyAdapterListener listener;
	private final List frames = Collections.synchronizedList(new LinkedList());

	private final List interfaceObjects = new ArrayList();

	private volatile boolean serverReset;
	private volatile boolean closed;
	private final boolean checkRW;

	/**
	 * @param localEP
	 * @param serverCtrlEP
	 * @param useNAT
	 * @param l
	 * @throws KNXException
	 */
	public KnIPDeviceMgmtAdapter(InetSocketAddress localEP,
		InetSocketAddress serverCtrlEP, boolean useNAT, PropertyAdapterListener l,
		boolean queryWriteEnable) throws KNXException
	{
		InetSocketAddress local = localEP;
		if (local == null)
			try {
				local = new InetSocketAddress(InetAddress.getLocalHost(), 0);
			}
			catch (final UnknownHostException e) {
				throw new KNXException("no local host available");
			}
		conn = new KNXnetIPDevMgmt(local, serverCtrlEP, useNAT);
		conn.addConnectionListener(new KNXListenerImpl());
		listener = l;
		checkRW = queryWriteEnable;
		try {
			initInterfaceObjects();
		}
		catch (final KNXException e) {
			conn.close();
			throw e;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.qsl.PropertyAdapter#close()
	 */
	public void close()
	{
		if (!closed) {
			closed = true;
			conn.close();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.qsl.PropertyAdapter#getPropertyDescription
	 *      (tuwien.auto.calimero.IndividualAddress, int, int, int)
	 */
	public byte[] getDescription(int objIndex, int pid, int propIndex)
		throws KNXTimeoutException, KNXConnectionClosedException, KNXRemoteException
	{
		// imitate property description:
		// oindex: PID_IO_LIST
		// pindex: supplied, conditional
		// pid: supplied, conditional
		// pdt: -
		// elements: current number of elements
		// rLevel: -
		// wLevel: -
		// read-only: only with emulate write

		// get current number of elements
		// since we can't query by index, emulate with PID if necessary
		// works for now, even though we get O(n^2) in scans..
		byte[] ret = null;
		int p = pid;
		if (p == 0) {
			int i = -1;
			while (p < 0xff && i < propIndex)
				try {
					ret = getProperty(objIndex, ++p, 0, 1);
					++i;
				}
				catch (final KNXRemoteException e) {}
			if (i != propIndex)
				throw new KNXRemoteException("can't deduce property index in object");
		}
		else
			ret = getProperty(objIndex, p, 0, 1);
		int writeEnabled = 0;
		if (checkRW) {
			// check out if write enabled or read only
			try {
				setProperty(objIndex, p, 0, 1, ret);
				writeEnabled = 0x80;
			}
			catch (final KNXRemoteException e) {}
		}
		// int elements = 0;
		// for (int i = 0; i < ret.length; ++i)
		// 	elements = elements << 8 | ret[i] & 0xff;
		return new byte[] { (byte) objIndex, (byte) p, (byte) propIndex,
			(byte) writeEnabled, 0, 0, 0 };
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.qsl.PropertyAdapter#getProperty
	 *      (tuwien.auto.calimero.IndividualAddress, int, int, int, int)
	 */
	public byte[] getProperty(int objIndex, int pid, int start, int elements)
		throws KNXTimeoutException, KNXRemoteException, KNXConnectionClosedException
	{
		if (closed)
			throw new KNXIllegalStateException("adapter closed");
		conn.send(new CEMIDevMgmt(CEMIDevMgmt.MC_PROPREAD_REQ, getObjectType(objIndex),
			1, pid, start, elements), KNXnetIPConnection.WAIT_FOR_CON);
		return findFrame(CEMIDevMgmt.MC_PROPREAD_CON);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.qsl.PropertyAdapter#setProperty
	 *      (tuwien.auto.calimero.IndividualAddress, int, int, int, byte[])
	 */
	public void setProperty(int objIndex, int pid, int start, int elements, byte[] data)
		throws KNXTimeoutException, KNXRemoteException, KNXConnectionClosedException
	{
		if (closed)
			throw new KNXIllegalStateException("adapter closed");
		conn.send(new CEMIDevMgmt(CEMIDevMgmt.MC_PROPWRITE_REQ, getObjectType(objIndex),
			1, pid, start, elements, data), KNXnetIPConnection.WAIT_FOR_CON);
		findFrame(CEMIDevMgmt.MC_PROPWRITE_CON);
	}

	/**
	 * {@inheritDoc} The name for this KNXnet/IP local device management (DM) adapter
	 * starts with "local", allowing easier distinction of adapters.
	 */
	public String getName()
	{
		return "local DM " + conn.getRemoteAddress();
	}

	/* (non-Javadoc)
	 * @see tuwien.auto.calimero.mgmt.PropertyAdapter#isOpen()
	 */
	public boolean isOpen()
	{
		return !closed;
	}

	private byte[] findFrame(int messageCode) throws KNXRemoteException,
		KNXConnectionClosedException
	{
		while (frames.size() > 0) {
			final CEMIDevMgmt frame = (CEMIDevMgmt) frames.remove(0);
			final int mc = frame.getMessageCode();
			if (mc == CEMIDevMgmt.MC_RESET_IND) {
				serverReset = true;
				close();
				throw new KNXConnectionClosedException("reset by server");
			}
			else if (mc == messageCode) {
				if (frame.isNegativeResponse())
					throw new KNXRemoteException(frame.getErrorMessage());
				return frame.getData();
			}
		}
		throw new KNXInvalidResponseException("confirmation expected");
	}

	private int getObjectType(int objIndex) throws KNXRemoteException
	{
		for (final Iterator i = interfaceObjects.iterator(); i.hasNext();) {
			final Pair p = (Pair) i.next();
			if (p.oindex == objIndex)
				return p.otype;
		}
		throw new KNXRemoteException("object not listed");
	}

	/**
	 * @throws KNXTimeoutException
	 */
	private void initInterfaceObjects() throws KNXException
	{
		conn.send(new CEMIDevMgmt(CEMIDevMgmt.MC_PROPREAD_REQ, DEVICE_OBJECT, 1,
			PID_IO_LIST, 0, 1), KNXnetIPConnection.WAIT_FOR_CON);
		int objects = 0;
		try {
			final byte[] ret = findFrame(CEMIDevMgmt.MC_PROPREAD_CON);
			objects = (ret[0] & 0xff) << 8 | ret[1] & 0xff;
		}
		catch (final KNXRemoteException e) {
			// device only has device- and cEMI server-object
			interfaceObjects.add(new Pair(0, DEVICE_OBJECT));
			interfaceObjects.add(new Pair(1, 8));
			return;
		}
		conn.send(new CEMIDevMgmt(CEMIDevMgmt.MC_PROPREAD_REQ, DEVICE_OBJECT, 1,
			PID_IO_LIST, 1, objects), KNXnetIPConnection.WAIT_FOR_CON);
		final byte[] ret = findFrame(CEMIDevMgmt.MC_PROPREAD_CON);
		int lastObjType = -1;
		for (int i = 0; i < objects; ++i) {
			final int objType = (ret[2 * i] & 0xff) << 8 | ret[2 * i + 1] & 0xff;
			if (objType != lastObjType)
				interfaceObjects.add(new Pair(i, objType));
			lastObjType = objType;
		}
	}
}
