/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: TransportLayerImpl.java,v $
 Revision 1.1  2007/09/09 23:02:54  bm
 initial commit, transport layer encapsulates one logical connection
 with all necessary information

 Revision 1.1  2007/03/17 18:02:49  bm
 added implementation of transport layer,
 no appropriate exception handling yet

 */

package tuwien.auto.calimero.mgmt;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import tuwien.auto.calimero.CloseEvent;
import tuwien.auto.calimero.FrameEvent;
import tuwien.auto.calimero.GroupAddress;
import tuwien.auto.calimero.IndividualAddress;
import tuwien.auto.calimero.KNXAddress;
import tuwien.auto.calimero.Priority;
import tuwien.auto.calimero.cemi.CEMI;
import tuwien.auto.calimero.cemi.CEMILData;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;
import tuwien.auto.calimero.exception.KNXIllegalStateException;
import tuwien.auto.calimero.exception.KNXTimeoutException;
import tuwien.auto.calimero.link.KNXLinkClosedException;
import tuwien.auto.calimero.link.KNXNetworkLink;
import tuwien.auto.calimero.link.event.NetworkLinkListener;
import tuwien.auto.calimero.log.LogManager;
import tuwien.auto.calimero.log.LogService;
import tuwien.auto.calimero.mgmt.Destination.AggregatorProxy;
import tuwien.auto.calimero.process.DetachEvent;

/**
 * @author bm
 */
public class TransportLayerImpl implements TransportLayer
{
	private final class NLListener implements NetworkLinkListener
	{
		NLListener()
		{}

		public void confirmation(FrameEvent e)
		{}

		public void indication(FrameEvent e)
		{
			final CEMILData f = (CEMILData) e.getFrame();
			final int ctrl = f.getData()[0] & 0xfc;
			if (ctrl == 0) {
				final KNXAddress dst = f.getDestination();
				if (dst instanceof GroupAddress)
					// check for broadcast or group
					fireFrameType(f, dst.getRawAddress() == 0 ? 0 : 1);
				else
					// individual
					fireFrameType(f, 2);
			}
			else {
				final IndividualAddress src = f.getSource();
				// are we waiting for ack?
				synchronized (indications) {
					if (active != null
						&& active.getDestination().getAddress().equals(src)) {
						indications.add(e);
						indications.notify();
						return;
					}
				}
				try {
					AggregatorProxy ap = null;
					synchronized (proxies) {
						ap = (AggregatorProxy) proxies.get(f.getSource());
					}
					handleConnected(f, ap);
				}
				catch (final KNXLinkClosedException ignore) {
					// we get notified with link-closed event
				}
				catch (final KNXTimeoutException ignore) {
					// possible timeouts on sending ack
				}
			}
		}

		public void linkClosed(CloseEvent e)
		{
			closeDestinations(true);
			fireLinkClosed(e);
		}
	}

	private static final short CONNECT = 0x80;
	private static final short DISCONNECT = 0x81;
	private static final short ACK = 0xC2;
	private static final short NACK = 0xC3;
	private static final short DATA_CONNECTED = 0x40;
	// group, broadcast and individual .req service codes are 0

	// acknowledge timeout in seconds
	private static final short ACK_TIMEOUT = 3;
	// maximum repetitions of send in connected mode
	private static final int MAX_REPEAT = 3;

	private static final GroupAddress broadcast = new GroupAddress(0);
	// used as default on incoming conn.oriented messages from unknown remote devices
	private final Destination unknownPartner = new Destination(new AggregatorProxy(this),
		new IndividualAddress(0), true);

	private final LogService logger;

	private volatile boolean detached;
	private final KNXNetworkLink lnk;
	private final NetworkLinkListener lnkListener = new NLListener();
	private final List indications = new LinkedList();
	private final EventListeners listeners;

	// holds the mapping of connection destination address to proxy
	private final Map proxies = new HashMap();
	private AggregatorProxy active;
	private volatile int repeated;
	private final Object lock = new Object();

	public TransportLayerImpl(KNXNetworkLink link) throws KNXLinkClosedException
	{
		if (!link.isOpen())
			throw new KNXLinkClosedException();
		lnk = link;
		lnk.addLinkListener(lnkListener);
		logger = LogManager.getManager().getLogger(getName());
		listeners = new EventListeners(logger);
	}

	public void addTransportListener(TransportListener l)
	{
		listeners.add(l);
	}

	public void broadcast(Priority p, byte[] tsdu) throws KNXTimeoutException,
		KNXLinkClosedException
	{
		sendData(broadcast, p, tsdu);
	}

	public void connect(Destination d) throws KNXTimeoutException, KNXLinkClosedException
	{
		final AggregatorProxy p = getProxy(d);
		if (!d.isConnectionOriented()) {
			logger.error("destination not connection oriented: " + d.getAddress());
			return;
		}
		if (d.getState() != Destination.DISCONNECTED)
			return;
		p.setState(Destination.CONNECTING);
		final byte[] tpdu = new byte[] { (byte) CONNECT };
		lnk.sendRequestWait(d.getAddress(), Priority.SYSTEM, tpdu);
		p.setState(Destination.OPEN_IDLE);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.qsl.TransportLayer#createConnectionObject
	 *      (tuwien.auto.calimero.IndividualAddress, boolean)
	 */
	public Destination createDestination(IndividualAddress remote,
		boolean connectionOriented)
	{
		return createDestination(remote, connectionOriented, false, false);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.qsl.TransportLayer#createConnectionObject()
	 */
	public Destination createDestination(IndividualAddress remote,
		boolean connectionOriented, boolean keepAlive, boolean verifyMode)
	{
		if (detached)
			throw new KNXIllegalStateException("TL detached");
		synchronized (proxies) {
			if (proxies.containsKey(remote))
				throw new KNXIllegalArgumentException("destination already created: "
					+ remote);
			final AggregatorProxy p = new AggregatorProxy(this);
			final Destination d = new Destination(p, remote, connectionOriented,
				keepAlive, verifyMode);
			proxies.put(remote, p);
			return d;
		}
	}

	public synchronized KNXNetworkLink detach()
	{
		if (detached)
			return null;
		closeDestinations(false);
		lnk.removeLinkListener(lnkListener);
		detached = true;
		fireDetached();
		return lnk;
	}

	private void fireDetached()
	{
		final DetachEvent e = new DetachEvent(this);
		for (final Iterator i = listeners.iterator(); i.hasNext();) {
			final TransportListener l = (TransportListener) i.next();
			try {
				l.detached(e);
			}
			catch (final RuntimeException rte) {
				removeTransportListener(l);
				logger.error("removed event listener", rte);
			}
		}
	}

	public void disconnect(Destination d) throws KNXLinkClosedException
	{
		if (detached)
			throw new KNXIllegalStateException("TL detached");
		if (d.getState() != Destination.DESTROYED
			&& d.getState() != Destination.DISCONNECTED) {
			disconnectIndicate(getProxy(d), true);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.qsl.TransportLayer#getName()
	 */
	public String getName()
	{
		return "TL " + (detached ? "(detached)" : lnk.getName());
	}

	public void removeTransportListener(TransportListener l)
	{
		listeners.remove(l);
	}

	public void sendData(Destination d, Priority p, byte[] tsdu)
		throws KNXDisconnectException, KNXLinkClosedException
	{
		final AggregatorProxy ap = getProxy(d);
		if (d.getState() == Destination.DISCONNECTED)
			throw new KNXDisconnectException(
				"no connection opened for " + d.getAddress(), d);

		tsdu[0] = (byte) ((tsdu[0] & 0x03) | DATA_CONNECTED | (ap.getSeqSend() << 2));
		// the entry lock guards between send and return (only one at a time)
		synchronized (lock) {
			// on indications we do wait() for incoming messages
			synchronized (indications) {
				try {
					active = ap;
					for (repeated = 0; repeated < MAX_REPEAT + 1; ++repeated) {
						try {
							logger.trace("sending data connected to " + d.getAddress()
								+ ", attempt " + (repeated + 1));
							// set state and timer
							ap.setState(Destination.OPEN_WAIT);
							lnk.sendRequestWait(d.getAddress(), p, tsdu);
							if (waitForAck())
								return;
						}
						catch (final KNXTimeoutException e) {}
						// cancel repetitions if detached or destroyed
						if (detached || d.getState() == Destination.DESTROYED)
							throw new KNXDisconnectException(
								"send data connected failed", d);
					}
				}
				finally {
					active = null;
					repeated = 0;
				}
			}
		}
		disconnectIndicate(ap, true);
		throw new KNXDisconnectException("send data connected failed", d);
	}

	public void sendData(KNXAddress addr, Priority p, byte[] tsdu)
		throws KNXTimeoutException, KNXLinkClosedException
	{
		if (detached)
			throw new KNXIllegalStateException("TL detached");
		tsdu[0] &= 0x03;
		lnk.sendRequestWait(addr, p, tsdu);
	}

	private void closeDestinations(boolean skipSendDisconnect)
	{
		// we can't use proxies default iterator due to concurrent modifications in
		// destroyDestination(), called by d.destroy()
		AggregatorProxy[] allProxies = new AggregatorProxy[proxies.size()];
		synchronized (proxies) {
			allProxies = (AggregatorProxy[]) proxies.values().toArray(allProxies);
		}
		for (int i = 0; i < allProxies.length; i++) {
			final AggregatorProxy p = allProxies[i];
			final Destination d = p.getDestination();
			if (skipSendDisconnect && d.getState() != Destination.DISCONNECTED) {
				p.setState(Destination.DISCONNECTED);
				fireDisconnected(d);
			}
			d.destroy();
		}
	}

	private void disconnectIndicate(AggregatorProxy p, boolean sendDisconnectReq)
		throws KNXLinkClosedException
	{
		p.setState(Destination.DISCONNECTED);
		try {
			if (sendDisconnectReq)
				sendDisconnect(p.getDestination().getAddress());
		}
		finally {
			fireDisconnected(p.getDestination());
		}
	}

	private void fireDisconnected(Destination d)
	{
		for (final Iterator i = listeners.iterator(); i.hasNext();) {
			final TransportListener l = (TransportListener) i.next();
			try {
				l.disconnected(d);
			}
			catch (final RuntimeException rte) {
				removeTransportListener(l);
				logger.error("removed event listener", rte);
			}
		}
	}

	// type 0 = broadcast, 1 = group, 2 = individual, 3 = connected
	private void fireFrameType(CEMI frame, int type)
	{
		final FrameEvent e = new FrameEvent(this, frame);
		for (final Iterator i = listeners.iterator(); i.hasNext();) {
			final TransportListener l = (TransportListener) i.next();
			try {
				if (type == 0)
					l.broadcast(e);
				else if (type == 1)
					l.group(e);
				else if (type == 2)
					l.individual(e);
				else if (type == 3)
					l.connectedData(e);
			}
			catch (final RuntimeException rte) {
				removeTransportListener(l);
				logger.error("removed event listener", rte);
			}
		}
	}

	private void fireLinkClosed(CloseEvent e)
	{
		for (final Iterator i = listeners.iterator(); i.hasNext();) {
			final TransportListener l = (TransportListener) i.next();
			try {
				l.linkClosed(e);
			}
			catch (final RuntimeException rte) {
				removeTransportListener(l);
				logger.error("removed event listener", rte);
			}
		}
	}

	private AggregatorProxy getProxy(Destination d)
	{
		if (detached)
			throw new KNXIllegalStateException("TL detached");
		synchronized (proxies) {
			final AggregatorProxy p = (AggregatorProxy) proxies.get(d.getAddress());
			// check identity, too, to prevent destination with only same address
			if (p == null || p.getDestination() != d)
				throw new KNXIllegalArgumentException("not the owner of " + d.toString());
			return p;
		}
	}

	private void handleConnected(CEMILData frame, AggregatorProxy p)
		throws KNXLinkClosedException, KNXTimeoutException
	{
		final IndividualAddress sender = frame.getSource();
		final byte[] tpdu = frame.getData();
		final int ctrl = tpdu[0] & 0xFF;
		final int seq = (tpdu[0] & 0x3C) >>> 2;

		// on proxy null (no destination found for sender) use 'no partner' placeholder
		final Destination d = p != null ? p.getDestination() : unknownPartner;

		if (ctrl == CONNECT) {
			if (d.getState() == Destination.DISCONNECTED)
				sendDisconnect(sender);
		}
		else if (ctrl == DISCONNECT) {
			if (d.getState() != Destination.DISCONNECTED
				&& sender.equals(d.getAddress())) {
				disconnectIndicate(p, false);
			}
		}
		else if ((ctrl & 0xC0) == DATA_CONNECTED) {
			if (d.getState() == Destination.DISCONNECTED
				|| !sender.equals(d.getAddress()))
				sendDisconnect(sender);
			else {
				p.restartTimeout();
				if (seq == p.getSeqReceive()) {
					lnk.sendRequest(sender, Priority.SYSTEM,
						new byte[] { (byte) (ACK | (p.getSeqReceive() << 2)) });
					p.incSeqReceive();
					fireFrameType(frame, 3);
				}
				else if (seq == ((p.getSeqReceive() - 1) & 0xF)) {
					lnk.sendRequest(sender, Priority.SYSTEM,
						new byte[] { (byte) (ACK | (seq << 2)) });
				}
				else {
					lnk.sendRequest(sender, Priority.SYSTEM,
						new byte[] { (byte) (NACK | (seq << 2)) });
				}
			}
		}
		else if ((ctrl & 0xC3) == ACK) {
			if (d.getState() == Destination.DISCONNECTED
				|| !sender.equals(d.getAddress()))
				sendDisconnect(sender);
			else if (d.getState() == Destination.OPEN_WAIT && seq == p.getSeqSend()) {
				p.incSeqSend();
				p.setState(Destination.OPEN_IDLE);
				logger.trace("positive ack by " + d.getAddress());
			}
			else {
				disconnectIndicate(p, true);
			}
		}
		else if ((ctrl & 0xC3) == NACK) {
			if (d.getState() == Destination.DISCONNECTED
				|| !sender.equals(d.getAddress()))
				sendDisconnect(sender);
			else if (d.getState() == Destination.OPEN_WAIT && seq == p.getSeqSend()
				&& repeated < MAX_REPEAT) {
				// we will send message again
				;
			}
			else {
				disconnectIndicate(p, true);
			}
		}
	}

	private void sendDisconnect(IndividualAddress addr) throws KNXLinkClosedException
	{
		final byte[] tpdu = new byte[] { (byte) DISCONNECT };
		try {
			lnk.sendRequest(addr, Priority.SYSTEM, tpdu);
		}
		catch (final KNXTimeoutException ignore) {
			// do a warning, but otherwise can be ignored
			logger.warn("disconnected not gracefully (timeout)", ignore);
		}
	}

	private boolean waitForAck() throws KNXDisconnectException, KNXLinkClosedException,
		KNXTimeoutException
	{
		long remaining = ACK_TIMEOUT * 1000;
		final long end = System.currentTimeMillis() + remaining;
		final Destination d = active.getDestination();
		while (remaining > 0) {
			try {
				while (indications.size() > 0)
					handleConnected((CEMILData) ((FrameEvent) indications.remove(0))
						.getFrame(), active);
				if (d.getState() == Destination.DISCONNECTED)
					throw new KNXDisconnectException(d.getAddress()
						+ " disconnected while awaiting ACK", d);
				if (d.getState() == Destination.OPEN_IDLE)
					return true;
				indications.wait(remaining);
				if (d.getState() == Destination.DISCONNECTED)
					throw new KNXDisconnectException(d.getAddress()
						+ " disconnected while awaiting ACK", d);
			}
			catch (final InterruptedException e) {}
			remaining = end - System.currentTimeMillis();
		}
		return false;
	}

	/* (non-Javadoc)
	 * @see tuwien.auto.calimero.qsl.TransportLayer#destroyDestination
	 * (tuwien.auto.calimero.qsl.Destination)
	 */
	public void destroyDestination(Destination d)
	{
		// method invocation is idempotent
		synchronized (proxies) {
			final AggregatorProxy p = (AggregatorProxy) proxies.get(d.getAddress());
			if (p == null)
				return;
			if (p.getDestination() == d) {
				proxies.remove(d.getAddress());
				d.destroy();
			}
			else
				logger.warn("not owner of " + d.getAddress());
		}
	}
}
