/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log$
 */

package tuwien.auto.calimero.process;

import tuwien.auto.calimero.GroupAddress;
import tuwien.auto.calimero.KNXAddress;
import tuwien.auto.calimero.dptxlator.DPTXlator;
import tuwien.auto.calimero.dptxlator.PropertyTypes;
import tuwien.auto.calimero.exception.KNXException;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;
import tuwien.auto.calimero.log.LogManager;
import tuwien.auto.calimero.log.LogService;

/**
 * @author bm
 */
public final class DataUnitBuilder
{
	private static final int T_CONNECT = 0x80;
	private static final int T_DISCONNECT = 0x81;
	private static final int T_ACK = 0xC2;
	private static final int T_NAK = 0xC3;
	
	private static final LogService logger =
		LogManager.getManager().getLogger("Data-Unit Builder");

	private DataUnitBuilder()
	{}

	public static byte[] createAPDU(int service, byte[] asdu)
	{
		if (asdu.length > 254)
			throw new KNXIllegalArgumentException(
				"ASDU length exceeds maximum of 254 bytes");
		final byte[] apdu = new byte[2 + asdu.length];
		apdu[0] = (byte) ((service >> 8) & 0x03);
		apdu[1] |= (byte) service;
		for (int i = 0; i < asdu.length; ++i)
			apdu[i + 2] = asdu[i];
		return apdu;
	}

	/**
	 * @param service
	 * @param asdu if no ASDU is required use <code>null</code>
	 */
	public static byte[] createCompactAPDU(int service, byte[] asdu)
	{
		final byte[] apdu =
			new byte[(asdu != null && asdu.length > 0) ? 1 + asdu.length : 2];
		if (apdu.length > 255)
			throw new KNXIllegalArgumentException(
				"APDU length exceeds maximum of 255 bytes");
		apdu[0] = (byte) ((service >> 8) & 0x03);
		apdu[1] = (byte) service;
		if (asdu != null && asdu.length > 0) {
			// maximum of 6 bits in asdu[0] are valid
			apdu[1] |= asdu[0] & 0x3F;
			for (int i = 1; i < asdu.length; ++i)
				apdu[i + 1] = asdu[i];
		}
		return apdu;
	}

	public static int getService(byte[] tpdu)
	{
		// high 4 bits of APCI
		final int apci4 = (tpdu[0] & 0x03) << 2 | (tpdu[1] & 0xC0) >> 6;
		// lowest 6 bits of APCI
		final int apci6 = tpdu[1] & 0x3f;
		// group value codes
		if (apci4 == 0) {
			if (apci6 == 0)
				return 0;
		}
		else if (apci4 == 1)
			return 0x40;
		else if (apci4 == 2)
			return 0x80;
		// individual address codes
		else if (apci4 == 3 || apci4 == 4 || apci4 == 5) {
			if (apci6 == 0)
				return apci4 << 6;
		}
		// ADC codes
		else if (apci4 == 6 || apci4 == 7)
			return apci4 << 6;
		// memory codes
		else if (apci4 == 8 || apci4 == 9 || apci4 == 10) {
			if ((apci6 & 0x30) == 0)
				return apci4 << 6;
		}
		// the rest
		else
			return apci4 << 6 | apci6;
		// unknown codes
		final int code = apci4 << 6 | apci6;
		logger.warn("unknown APCI service code 0x" + Integer.toHexString(code));
		return code;
	}

	/**
	 * Extracts the data array with KNX data type items for translation out of a group
	 * service APDU.
	 * <p>
	 * 
	 * @param apdu application layer protocol data unit byte array, 2 &lt;= apdu.length
	 *        &lt;= 16
	 */
	public static void extractGroupASDU(byte[] apdu, DPTXlator t)
	{
		if (apdu.length < 2)
			throw new KNXIllegalArgumentException("minimum APDU length is 2 bytes");
		t.setData(apdu, apdu.length == 2 ? 1 : 2);
	}

	/**
	 * Creates a group service APDU containing all DPT items of this translator.
	 * <p>
	 * The transport layer bits in the first byte (TL / AL control field) are set 0. The
	 * maximum length used for the ASDU is not checked.<br>
	 * For application service values occupying &lt;= 6 bits in length the optimized group
	 * write / response format structure is used (with the value starting at at bit offset
	 * 4 in the APDU).
	 * 
	 */
	public static byte[] createGroupAPDU(int service, DPTXlator t)
	{
		// check for group read
		if (service == 0x00)
			return new byte[2];
		// only group response and group write are allowed
		if (service != 0x40 && service != 0x80)
			throw new KNXIllegalArgumentException("not a APDU group service");
		// determine if data starts at byte offset 1 (optimized) or 2 (default)
		final int offset = (t.getItems() == 1 && t.getTypeSize() == 0) ? 1 : 2;
		final byte[] buf = new byte[t.getItems() * Math.max(1, t.getTypeSize()) + offset];
		buf[0] = (byte) (service >> 8);
		buf[1] = (byte) service;
		return t.getData(buf, offset);
	}

	/**
	 * @param dataType property data type to get the associated translator for
	 * @param apdu application layer protocol data unit (APDU) of a cEMI property message
	 *        containing PDT items
	 * @return the created DPT translator with the PDT items set
	 * @throws KNXException on PDT not found or translator could not be created
	 */
	public static DPTXlator extractPropertyData(int dataType, byte[] apdu)
		throws KNXException
	{
		final DPTXlator t = PropertyTypes.createTranslator(dataType);
		t.setData(apdu, 6);
		return t;
	}

	public static String decode(byte[] tpdu, KNXAddress dst)
	{
		if (tpdu.length < 1)
			throw new KNXIllegalArgumentException("tpdu length too short");
		final String s = decodeTPCI(tpdu[0] & 0xff, dst);
		if (tpdu.length > 1)
			return s + ", " + decodeAPCI(getService(tpdu));
		return s;
	}

	public static String decodeTPCI(int tpci, KNXAddress dst)
	{
		final int ctrl = tpci & 0xff;
		if ((ctrl & 0xFC) == 0) {
			if (dst == null)
				return "T-broadcast/group/ind";
			if (dst.getRawAddress() == 0)
				return "T-broadcast";
			if (dst instanceof GroupAddress)
				return "T-group";
			return "T-individual";
		}
		if ((ctrl & 0xC0) == 0x40)
			return "T-connected seq " + (ctrl >> 2 & 0xF);
		if (ctrl == T_CONNECT)
			return "T-connect";
		if (ctrl == T_DISCONNECT)
			return "T-disconnect";
		if ((ctrl & 0xC3) == T_ACK)
			return "T-ack seq " + (ctrl >> 2 & 0xF);
		if ((ctrl & 0xC3) == T_NAK)
			return "T-nak seq " + (ctrl >> 2 & 0xF);
		return "unknown TPCI";
	}

	public static String decodeAPCI(int apci)
	{
		switch (apci) {
		case 0x00:
			return "A-group.read";
		case 0x40:
			return "A-group.response";
		case 0x80:
			return "A-group.write";
		case 0x0180:
			return "A-ADC.read";
		case 0x01C0:
			return "A-ADC.response";
		case 0x03D1:
			return "A-authorize.read";
		case 0x03D2:
			return "A-authorize.response";
		case 0x3E0:
			return "A-domain.write";
		case 0x3E1:
			return "A-domain.read";
		case 0x3E2:
			return "A-domain.response";
		case 0x3E3:
			return "A-domain-selective.read";
		case 0x0100:
			return "A-ind.addr.read";
		case 0x0140:
			return "A-ind.addr.response";
		case 0xC0:
			return "A-ind.addr.write";
		case 0x03DC:
			return "A-ind.addr-sno.read";
		case 0x03DD:
			return "A-ind.addr_sno.response";
		case 0x03DE:
			return "A-ind.addr_sno.write";
		case 0x300:
			return "A-device-desc.read";
		case 0x340:
			return "A-device-desc.response";
		case 0x03D3:
			return "A-key.write";
		case 0x03D4:
			return "A-key.response";
		case 0x0200:
			return "A-memory.read";
		case 0x0240:
			return "A-memory.response";
		case 0x0280:
			return "A-memory.write";
		case 0x03D8:
			return "A-property-desc.read";
		case 0x03D9:
			return "A-property-desc.response";
		case 0x03D5:
			return "A-property.read";
		case 0x03D6:
			return "A-property.response";
		case 0x03D7:
			return "A-property.write";
		case 0x0380:
			return "A-restart";
		default:
			return "unknown APCI";
		}
	}
}
