/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: UserLayerImpl.java,v $
 Revision 1.1  2007/09/15 21:20:24  bm
 interface CacheAccess still implemented

 Revision 1.4  2007/03/18 12:53:11  bm
 finish implementation of all interface methods,
 get into conformance to latest KNX standards

 Revision 1.3  2007/03/11 20:09:16  bm
 first prototype implementation of more common used methods

 Revision 1.2  2007/03/03 23:53:41  bm
 added function stubs and constructors

 Revision 1.1  2007/03/03 22:24:37  bm
 added class for further implementation of
 user layer with support for caching

 */

package tuwien.auto.calimero.process;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import tuwien.auto.calimero.CloseEvent;
import tuwien.auto.calimero.FrameEvent;
import tuwien.auto.calimero.GroupAddress;
import tuwien.auto.calimero.KNXAddress;
import tuwien.auto.calimero.Priority;
import tuwien.auto.calimero.cemi.CEMILData;
import tuwien.auto.calimero.datapoint.Datapoint;
import tuwien.auto.calimero.dptxlator.DPTXlator;
import tuwien.auto.calimero.dptxlator.DPTXlator2ByteFloat;
import tuwien.auto.calimero.dptxlator.DPTXlator3BitControlled;
import tuwien.auto.calimero.dptxlator.DPTXlator8BitUnsigned;
import tuwien.auto.calimero.dptxlator.DPTXlatorBoolean;
import tuwien.auto.calimero.dptxlator.DPTXlatorString;
import tuwien.auto.calimero.dptxlator.TranslatorTypes;
import tuwien.auto.calimero.exception.KNXException;
import tuwien.auto.calimero.exception.KNXFormatException;
import tuwien.auto.calimero.exception.KNXIllegalStateException;
import tuwien.auto.calimero.exception.KNXInvalidResponseException;
import tuwien.auto.calimero.exception.KNXRemoteException;
import tuwien.auto.calimero.exception.KNXTimeoutException;
import tuwien.auto.calimero.link.KNXLinkClosedException;
import tuwien.auto.calimero.link.KNXNetworkLink;
import tuwien.auto.calimero.link.event.NetworkLinkListener;
import tuwien.auto.calimero.log.LogManager;
import tuwien.auto.calimero.log.LogService;

/**
 * <p>
 * This implementation of the process communicator interface uses in any case the DPT
 * translators {@link DPTXlatorBoolean}, {@link DPTXlator3BitControlled},
 * {@link DPTXlator8BitUnsigned}, {@link DPTXlator2ByteFloat}, {@link DPTXlatorString}.
 * Other translator types are loaded through {@link TranslatorTypes}.
 * 
 * @author bm
 */
public class ProcessCommunicatorImpl implements ProcessCommunicator
{
	private static final short GROUP_READ = 0x00;
	private static final short GROUP_RESPONSE = 0x40;
	private static final short GROUP_WRITE = 0x80;

	private static final LogService logger = LogManager.getManager().getLogger("UL");

	private final class NLListener implements NetworkLinkListener
	{
		NLListener()
		{}

		public void indication(FrameEvent e)
		{
			final CEMILData f = (CEMILData) e.getFrame();
			final byte[] apdu = f.getTPDU();
			final int svc = DataUnitBuilder.getService(apdu);
			if (svc == GROUP_RESPONSE) {
				if (wait)
					synchronized (indications) {
						indications.add(e);
						indications.notify();
					}
			}
			else {
				final ProcessEvent e2 =
					new ProcessEvent(this, f.getSource(), (GroupAddress) f
						.getDestination(), apdu);
				if (svc == GROUP_READ)
					for (final Iterator i = listeners.iterator(); i.hasNext();)
						((ProcessListener) i.next()).groupRead(e2);
				else if (svc == GROUP_WRITE)
					for (final Iterator i = listeners.iterator(); i.hasNext();)
						((ProcessListener) i.next()).groupWrite(e2);
			}
		}

		public void confirmation(FrameEvent e)
		{}

		public void linkClosed(CloseEvent e)
		{}
	};

	private final KNXNetworkLink lnk;
	private final NetworkLinkListener lnkListener = new NLListener();
	private final List indications = new LinkedList();
	private final EventListeners listeners = new EventListeners(logger);
	private Priority priority = Priority.LOW;
	// maximum wait time in seconds for a response message
	private int responseTimeout = 10;
	private volatile boolean wait;
	private volatile boolean detached;

	// translators are created local, since creation is not expensive
//	private final DPTXlatorBoolean tbool;
//	private final DPTXlator3BitControlled t3bit;
//	private final DPTXlator2ByteFloat tfloat;
//	private final DPTXlatorString tstring;
//	private final DPTXlator8BitUnsigned tubyte;

	public ProcessCommunicatorImpl(KNXNetworkLink link) throws KNXLinkClosedException
	{
		if (!link.isOpen())
			throw new KNXLinkClosedException();
		lnk = link;
		lnk.addLinkListener(lnkListener);
	}

	public KNXNetworkLink detach()
	{
		if (detached)
			return null;
		detached = true;
		lnk.removeLinkListener(lnkListener);
		return lnk;
	}

	public Priority getPriority()
	{
		return priority;
	}

	/**
	 * should read timeout include the sending time before response?
	 * 
	 * @return the read timeout used in read requests
	 */
	public int getResponseTimeout()
	{
		return responseTimeout;
	}

	public String read(Datapoint dp) throws KNXException
	{
		final byte[] apdu = readFromGroup(dp.getMainAddress(), 0, 14);
		final DPTXlator t =
			TranslatorTypes.createTranslator(dp.getMainNumber(), dp.getDPT());
		DataUnitBuilder.extractGroupASDU(apdu, t);
		return t.getValue();
	}

	public boolean readBool(GroupAddress addr) throws KNXTimeoutException,
		KNXRemoteException, KNXLinkClosedException, KNXFormatException
	{
		final byte[] apdu = readFromGroup(addr, 0, 0);
		final DPTXlatorBoolean t = new DPTXlatorBoolean(DPTXlatorBoolean.DPT_BOOL);
		DataUnitBuilder.extractGroupASDU(apdu, t);
		return t.getValueBoolean();
	}

	public byte readControl(GroupAddress addr) throws KNXTimeoutException,
		KNXRemoteException, KNXLinkClosedException, KNXFormatException
	{
		final byte[] apdu = readFromGroup(addr, 0, 0);
		final DPTXlator3BitControlled t =
			new DPTXlator3BitControlled(DPTXlator3BitControlled.DPT_CONTROL_DIMMING);
		DataUnitBuilder.extractGroupASDU(apdu, t);
		return t.getValueSigned();
	}

	public float readFloat(GroupAddress addr) throws KNXTimeoutException,
		KNXRemoteException, KNXLinkClosedException, KNXFormatException
	{
		final byte[] apdu = readFromGroup(addr, 2, 2);
		final DPTXlator2ByteFloat t =
			new DPTXlator2ByteFloat(DPTXlator2ByteFloat.DPT_TEMPERATURE_DIFFERENCE);
		DataUnitBuilder.extractGroupASDU(apdu, t);
		return t.getValueFloat();

	}

	public String readString(GroupAddress addr) throws KNXTimeoutException,
		KNXRemoteException, KNXLinkClosedException, KNXFormatException
	{
		final byte[] apdu = readFromGroup(addr, 0, 14);
		final DPTXlatorString t = new DPTXlatorString(DPTXlatorString.DPT_STRING_8859_1);
		DataUnitBuilder.extractGroupASDU(apdu, t);
		return t.getValue();

	}

	public short readUnsigned(GroupAddress addr, String format)
		throws KNXTimeoutException, KNXRemoteException, KNXLinkClosedException,
		KNXFormatException
	{
		final byte[] apdu = readFromGroup(addr, 1, 1);
		final DPTXlator8BitUnsigned t = new DPTXlator8BitUnsigned(format);
		DataUnitBuilder.extractGroupASDU(apdu, t);
		return t.getValueShort();

	}

	public void setPriority(Priority p)
	{
		priority = p;
	}

	/**
	 * set the read timeout to wait for a read call to complete
	 * 
	 * @param timeout time in seconds
	 */
	public void setResponseTimeout(int timeout)
	{
		responseTimeout = timeout;
	}

	public void write(GroupAddress addr, boolean value) throws KNXTimeoutException,
		KNXLinkClosedException
	{
		try {
			final DPTXlatorBoolean t = new DPTXlatorBoolean(DPTXlatorBoolean.DPT_BOOL);
			t.setValue(value);
			write(addr, priority, t);
		}
		catch (final KNXFormatException ignore) {}
	}

	public void write(GroupAddress addr, boolean control, byte stepcode)
		throws KNXTimeoutException, KNXFormatException, KNXLinkClosedException
	{
		final DPTXlator3BitControlled t =
			new DPTXlator3BitControlled(DPTXlator3BitControlled.DPT_CONTROL_DIMMING);
		t.setValue(control, stepcode);
		write(addr, priority, t);
	}

	public void write(GroupAddress addr, float value) throws KNXTimeoutException,
		KNXFormatException, KNXLinkClosedException
	{
		final DPTXlator2ByteFloat t =
			new DPTXlator2ByteFloat(DPTXlator2ByteFloat.DPT_TEMPERATURE_DIFFERENCE);
		t.setValue(value);
		write(addr, priority, t);
	}

	public void write(GroupAddress addr, short value, byte scale)
		throws KNXTimeoutException, KNXFormatException, KNXLinkClosedException
	{
		final DPTXlator8BitUnsigned t =
			new DPTXlator8BitUnsigned(scale == 1 ? DPTXlator8BitUnsigned.DPT_SCALING
				: scale == 2 ? DPTXlator8BitUnsigned.DPT_ANGLE
					: DPTXlator8BitUnsigned.DPT_VALUE_1_UCOUNT);
		t.setValue(value);
		write(addr, priority, t);
	}

	public void write(GroupAddress addr, String value) throws KNXTimeoutException,
		KNXFormatException, KNXLinkClosedException
	{
		final DPTXlatorString t = new DPTXlatorString(DPTXlatorString.DPT_STRING_8859_1);
		t.setValue(value);
		write(addr, priority, t);
	}

	public void write(Datapoint dp, String value) throws KNXException
	{
		final DPTXlator t =
			TranslatorTypes.createTranslator(dp.getMainNumber(), dp.getDPT());
		t.setValue(value);
		write(dp.getMainAddress(), dp.getPriority(), t);
	}

	private void write(KNXAddress a, Priority p, DPTXlator t) throws KNXTimeoutException,
		KNXLinkClosedException
	{
		if (detached)
			throw new KNXIllegalStateException("process communicator detached");
		lnk.sendRequestWait(a, p, DataUnitBuilder.createGroupAPDU(GROUP_WRITE, t));
	}

	private synchronized byte[] readFromGroup(GroupAddress addr, int minASDULen,
		int maxASDULen) throws KNXTimeoutException, KNXInvalidResponseException,
		KNXLinkClosedException
	{
		if (detached)
			throw new KNXIllegalStateException("process communicator detached");
		try {
			wait = true;
			lnk.sendRequestWait(addr, priority, DataUnitBuilder.createCompactAPDU(
				GROUP_READ, null));
			return waitForResponse(minASDULen + 2, maxASDULen + 2);
		}
		finally {
			wait = false;
		}
	}

	private byte[] waitForResponse(int minAPDU, int maxAPDU)
		throws KNXInvalidResponseException, KNXTimeoutException
	{
		long remaining = responseTimeout * 1000;
		final long end = System.currentTimeMillis() + remaining;
		synchronized (indications) {
			while (remaining > 0) {
				try {
					while (indications.size() > 0) {
						final FrameEvent e = (FrameEvent) indications.remove(0);
						final byte[] d = e.getFrame().getData();
						indications.clear();
						// ok, got response we're waiting for
						if (d.length >= minAPDU && d.length <= maxAPDU)
							return d;
						throw new KNXInvalidResponseException("APDU response length "
							+ d.length + " bytes, expected " + minAPDU + " to " + maxAPDU);
					}
					indications.wait(remaining);
				}
				catch (final InterruptedException e) {}
				remaining = end - System.currentTimeMillis();
			}
		}
		throw new KNXTimeoutException("timeout while waiting for group response");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.qsl.UserLayer#removeEventListener
	 *      (tuwien.auto.calimero.qsl.UserLayerListener)
	 */
	public void removeProcessListener(ProcessListener l)
	{
		listeners.remove(l);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tuwien.auto.calimero.qsl.UserLayer#addEventListener
	 *      (tuwien.auto.calimero.qsl.UserLayerListener)
	 */
	public void addProcessListener(ProcessListener l)
	{
		listeners.add(l);
	}
}
