/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log$
 */

package tuwien.auto.calimero.tools;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import tuwien.auto.calimero.Settings;
import tuwien.auto.calimero.exception.KNXException;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;
import tuwien.auto.calimero.knxnetip.Discoverer;
import tuwien.auto.calimero.knxnetip.KNXnetIPConnection;
import tuwien.auto.calimero.knxnetip.servicetype.DescriptionResponse;
import tuwien.auto.calimero.knxnetip.servicetype.SearchResponse;
import tuwien.auto.calimero.knxnetip.util.DeviceDIB;
import tuwien.auto.calimero.knxnetip.util.ManufacturerDIB;
import tuwien.auto.calimero.knxnetip.util.SuppFamiliesDIB;

/**
 * @author bm
 */
public class Discover
{
	private static final String tool = "Discover";
	private static final String sep = System.getProperty("line.separator");
	private static Discoverer d;

	private Discover()
	{}

	/**
	 * @param args
	 */
	public static void main(String[] args)
	{
		try {
			final Map options = new HashMap();
			if (parseArgs(args, options)) {
				final int lp = ((Integer) options.get("localport")).intValue();
				d = new Discoverer(lp, options.containsKey("nat"));
				if (options.containsKey("desc"))
					description(options);
				else
					discover(options);
			}
		}
		catch (final Throwable t) {
			if (t.getMessage() != null)
				System.out.println(t.getMessage());
		}
	}

	private static void discover(Map options) throws KNXException, InterruptedException
	{
		final int timeout = ((Integer) options.get("timeout")).intValue();
		if (options.containsKey("if"))
			d.startSearch(0, (NetworkInterface) options.get("if"), timeout, false);
		else
			d.startSearch(timeout, false);
		int displayed = 0;
		while (d.isSearching()) {
			Thread.sleep(250);
			final SearchResponse[] res = d.getSearchResponses();
			for (; displayed < res.length; ++displayed)
				printSearchResponse(res[displayed]);
		}
		final SearchResponse[] res = d.getSearchResponses();
		for (; displayed < res.length; ++displayed)
			printSearchResponse(res[displayed]);
	}

	private static void description(Map options) throws KNXException
	{
		final InetSocketAddress server =
			new InetSocketAddress((InetAddress) options.get("desc"), ((Integer) options
				.get("serverport")).intValue());
		final int timeout = ((Integer) options.get("timeout")).intValue();
		final DescriptionResponse res = d.getDescription(server, timeout);
		printDescriptionResponse(res);
	}

	private static boolean parseArgs(String[] args, Map options)
	{
		if (args.length == 0) {
			System.out.println("A tool for KNXnet/IP router discovery & description");
			showVersion();
			System.out.println("type -help for help message");
			return false;
		}
		// add defaults
		options.put("localport", new Integer(0));
		options.put("serverport", new Integer(KNXnetIPConnection.IP_PORT));
		options.put("timeout", new Integer(5));

		int i = 0;
		for (; i < args.length; i++) {
			final String arg = args[i];
			if (isOption(arg, "-help", "-h")) {
				showUsage();
				return false;
			}
			if (isOption(arg, "-version", null)) {
				showVersion();
				return false;
			}
			if (isOption(arg, "-localport", null))
				options.put("localport", Integer.decode(args[++i]));
			else if (isOption(arg, "-nat", "-n"))
				options.put("nat", null);
			else if (isOption(arg, "-interface", "-i"))
				options.put("if", getNetworkIF(args[++i]));
			else if (isOption(arg, "-timeout", "-t")) {
				final Integer timeout = Integer.valueOf(args[++i]);
				// a value of 0 means infinite timeout
				if (timeout.intValue() > 0)
					options.put("timeout", timeout);
			}
			else if (isOption(arg, "-description", "-d")) {
				try {
					options.put("desc", InetAddress.getByName(args[++i]));
				}
				catch (final UnknownHostException e) {
					throw new IllegalArgumentException("failed to read host " + arg);
				}
			}
			else if (isOption(arg, "-serverport", "-p"))
				options.put("serverport", Integer.decode(args[++i]));
			else
				throw new IllegalArgumentException("unknown option " + arg);
		}
		return true;
	}

	private static boolean isOption(String arg, String longOpt, String shortOpt)
	{
		return arg.equals(longOpt) || shortOpt != null && arg.equals(shortOpt);
	}

	private static void showVersion()
	{
		final StringBuffer sb = new StringBuffer();
		sb.append(tool + " version 0.1");
		System.out.println(sb);
		Settings.main(new String[] { "-version" });
	}

	private static void showUsage()
	{
		final StringBuffer sb = new StringBuffer();
		sb.append("usage: " + tool + " [options]" + sep);
		sb.append("options:" + sep);
		sb.append("  -help -h                show this help message" + sep);
		sb.append("  -version                show tool/library version and exit" + sep);
		sb.append("  -localport <number>     local UDP port (default system "
			+ "assigned)" + sep);
		sb.append("  -nat -n                 enable Network Address Translation" + sep);
		sb.append("  -timeout -t             discovery/description response timeout"
			+ sep);
		sb.append("  -interface -i <if-name | ip-address> " + sep
			+ "          local multicast network interface for "
			+ "discovery (default system assigned)" + sep);
		sb.append("  -description -d <host>  query description from host" + sep);
		sb.append("  -serverport -p <number> server UDP port for description (default "
			+ KNXnetIPConnection.IP_PORT + ")" + sep);
		System.out.println(sb);
	}

	private static NetworkInterface getNetworkIF(String id)
	{
		try {
			NetworkInterface nif = NetworkInterface.getByName(id);
			if (nif != null)
				return nif;
			nif = NetworkInterface.getByInetAddress(InetAddress.getByName(id));
			if (nif == null)
				throw new KNXIllegalArgumentException(
					"ip address not bound to network if");
		}
		catch (final UnknownHostException e) {}
		catch (final SocketException e) {}
		throw new KNXIllegalArgumentException("network if not found");
	}

	private static void printSearchResponse(SearchResponse r)
	{
		final StringBuffer buf = new StringBuffer();
		buf.append("control endpoint ").append(r.getControlEndpoint().toString()).append(
			sep);
		extractDeviceInfo(r.getDevice(), buf);
		extractServiceInfo(r.getServiceFamilies(), buf);
		System.out.println(buf);
	}

	private static void printDescriptionResponse(DescriptionResponse r)
	{
		final StringBuffer buf = new StringBuffer();
		extractDeviceInfo(r.getDevice(), buf);
		extractServiceInfo(r.getServiceFamilies(), buf);
		extractManufacturerInfo(r.getManufacturerData(), buf);
		System.out.println(buf);
	}

	private static void extractDeviceInfo(DeviceDIB dib, StringBuffer buf)
	{
		buf.append("device \"").append(dib.getName()).append("\"").append(sep);
		buf.append("  KNX address ").append(dib.getAddress()).append(sep);
		buf.append("  KNX medium ").append(dib.getKNXMediumString()).append(sep);
		buf.append("  installation ").append(dib.getInstallation());
		buf.append(", project ").append(dib.getProject());
		buf.append(" (project-installation-ID ").append(dib.getProjectInstallID())
			.append(")").append(sep);
		try {
			buf.append("  routing multicast address ").append(
				InetAddress.getByAddress(dib.getMulticastAddress())).append(sep);
		}
		catch (final UnknownHostException ignore) {}
		buf.append("  MAC address ").append(dib.getMACAddressString()).append(sep);
		buf.append("  serial number 0x").append(dib.getSerialNumberString()).append(sep);
	}

	private static void extractServiceInfo(SuppFamiliesDIB dib, StringBuffer buf)
	{
		buf.append("supported service families:").append(sep);
		for (final Iterator i = dib.getFamilies().iterator(); i.hasNext();) {
			final Map.Entry entry = (Map.Entry) i.next();
			buf.append("  ").append(
				dib.getFamilyName(((Short) entry.getKey()).shortValue()));
			buf.append(" - version ").append(((Short) entry.getValue()).toString())
				.append(sep);
		}
	}

	private static void extractManufacturerInfo(ManufacturerDIB dib, StringBuffer buf)
	{
		if (dib != null) {
			buf.append("KNX manufacturer ID ");
			buf.append(Integer.toHexString(dib.getID())).append(sep);
			buf.append("manufacturer specific data (hex): ");
			buf.append(toHex(dib.getData())).append(sep);
		}
	}

	private static String toHex(byte[] data)
	{
		final StringBuffer sb = new StringBuffer();
		for (int i = 0; i < data.length; ++i) {
			final int no = data[i] & 0xff;
			if (no < 0x10)
				sb.append('0');
			sb.append(Integer.toHexString(no));
		}
		return sb.toString();
	}
}
