/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log$
 */

package tuwien.auto.calimero.tools;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Map;

import tuwien.auto.calimero.CloseEvent;
import tuwien.auto.calimero.FrameEvent;
import tuwien.auto.calimero.Settings;
import tuwien.auto.calimero.exception.KNXException;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;
import tuwien.auto.calimero.knxnetip.KNXnetIPConnection;
import tuwien.auto.calimero.link.KNXNetworkMonitor;
import tuwien.auto.calimero.link.KNXNetworkMonitorIP;
import tuwien.auto.calimero.link.event.LinkListener;
import tuwien.auto.calimero.link.event.MonitorFrameEvent;
import tuwien.auto.calimero.link.medium.KNXMediumSettings;
import tuwien.auto.calimero.link.medium.PLSettings;
import tuwien.auto.calimero.link.medium.RFSettings;
import tuwien.auto.calimero.link.medium.RawFrame;
import tuwien.auto.calimero.link.medium.RawFrameBase;
import tuwien.auto.calimero.link.medium.TPSettings;
import tuwien.auto.calimero.log.LogLevel;
import tuwien.auto.calimero.log.LogManager;
import tuwien.auto.calimero.log.LogStreamWriter;
import tuwien.auto.calimero.process.DataUnitBuilder;

/**
 * @author bm
 */
public class NetworkMonitor
{
	private static final String tool = "NetworkMonitor";
	private static final String sep = System.getProperty("line.separator");
	private KNXNetworkMonitor m;
	private LinkListener l;

	/**
	 * @param args
	 */
	public static void main(String[] args)
	{
		try {
			final Map options = new HashMap();
			if (parseArgs(args, options)) {
				new NetworkMonitor().run(options);
			}
		}
		catch (final Throwable t) {
			if (t.getMessage() != null)
				System.out.println(t.getMessage());
		}
	}

	private void run(Map options) throws KNXException
	{
		LogManager.getManager().addWriter(null, new ConsoleWriter(true));
		createMonitor(options);
		// TODO we don't shut down properly on Ctrl-C
		synchronized (this) {
			while (m.isOpen())
				try {
					wait();
				}
				catch (final InterruptedException e) {}
		}
	}

	private static boolean parseArgs(String[] args, Map options)
	{
		if (args.length == 0) {
			System.out.println("A tool for monitoring a KNX network");
			showVersion();
			System.out.println("type -help for help message");
			return false;
		}
		// add defaults
		options.put("port", new Integer(KNXnetIPConnection.IP_PORT));
		options.put("medium", TPSettings.TP1);

		int i = 0;
		for (; i < args.length; i++) {
			final String arg = args[i];
			if (isOption(arg, "-help", "-h")) {
				showUsage();
				return false;
			}
			if (isOption(arg, "-version", null)) {
				showVersion();
				return false;
			}
			else if (isOption(arg, "-port", "-p"))
				options.put("port", Integer.decode(args[++i]));
			else if (isOption(arg, "-nat", "-n"))
				options.put("nat", null);
			else if (isOption(arg, "-medium", "-m"))
				options.put("medium", getMedium(args[++i]));
			else {
				if (!options.containsKey("host"))
					try {
						options.put("host", InetAddress.getByName(arg));
					}
					catch (final UnknownHostException e) {
						throw new IllegalArgumentException("failed to read host " + arg);
					}
				else
					throw new IllegalArgumentException("unknown option " + arg);
			}
		}
		return true;
	}

	private static boolean isOption(String arg, String longOpt, String shortOpt)
	{
		return arg.equals(longOpt) || shortOpt != null && arg.equals(shortOpt);
	}

	private static void showVersion()
	{
		final StringBuffer sb = new StringBuffer();
		sb.append(tool + " version 0.1");
		System.out.println(sb);
		Settings.main(new String[] { "-version" });
	}

	private static void showUsage()
	{
		final StringBuffer sb = new StringBuffer();
		sb.append("usage: " + tool + " [options] <host>" + sep);
		sb.append("options:" + sep);
		sb.append("  -help -h                show this help message" + sep);
		sb.append("  -version                show tool/library version and exit" + sep);
		sb.append("  -port <number> -p       UDP port on host (default "
			+ KNXnetIPConnection.IP_PORT + ")" + sep);
		sb.append("  -nat -n                 enable Network Address Translation" + sep);
		sb.append("  -medium -m <id>         KNX medium [tp0|tp1|p110|p132|rf] "
			+ "(default tp1)" + sep);
		System.out.println(sb);
	}

	private static KNXMediumSettings getMedium(String id)
	{
		if (id.equals("tp0"))
			return TPSettings.TP0;
		else if (id.equals("tp1"))
			return TPSettings.TP1;
		else if (id.equals("p110"))
			return new PLSettings(false);
		else if (id.equals("p132"))
			return new PLSettings(true);
		else if (id.equals("rf"))
			return new RFSettings(null);
		else
			throw new KNXIllegalArgumentException("unknown medium");
	}

	private void createMonitor(Map options) throws KNXException
	{
		final InetSocketAddress host =
			new InetSocketAddress((InetAddress) options.get("host"), ((Integer) options
				.get("port")).intValue());
		m =
			new KNXNetworkMonitorIP(null, host, options.containsKey("nat"),
				(KNXMediumSettings) options.get("medium"));
		m.setDecodeRawFrames(true);
		l = new MonitorListener();
		m.addMonitorListener(l);
	}

	private final class MonitorListener implements LinkListener
	{
		public void indication(FrameEvent e)
		{
			final StringBuffer sb = new StringBuffer();
			sb.append(e.getFrame().toString());
			final RawFrame raw = ((MonitorFrameEvent) e).getRawFrame();
			if (raw != null) {
				sb.append(": ").append(raw.toString());
				if (raw instanceof RawFrameBase) {
					final RawFrameBase f = (RawFrameBase) raw;
					sb.append(": ").append(
						DataUnitBuilder.decode(f.getTPDU(), f.getDestination()));
				}
			}
			System.out.println(sb);
		}

		public void linkClosed(CloseEvent e)
		{
			synchronized (NetworkMonitor.this) {
				NetworkMonitor.this.notify();
			}
		}
	}

	private static final class ConsoleWriter extends LogStreamWriter
	{
		ConsoleWriter(boolean verbose)
		{
			super(verbose ? LogLevel.INFO : LogLevel.ERROR, System.out, true);
		}

		public void close()
		{}
	}
}
