/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log$
 */

package tuwien.auto.calimero.tools;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Map;

import tuwien.auto.calimero.GroupAddress;
import tuwien.auto.calimero.Settings;
import tuwien.auto.calimero.datapoint.Datapoint;
import tuwien.auto.calimero.exception.KNXException;
import tuwien.auto.calimero.exception.KNXFormatException;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;
import tuwien.auto.calimero.knxnetip.KNXnetIPConnection;
import tuwien.auto.calimero.link.KNXNetworkLink;
import tuwien.auto.calimero.link.KNXNetworkLinkIP;
import tuwien.auto.calimero.link.medium.KNXMediumSettings;
import tuwien.auto.calimero.link.medium.PLSettings;
import tuwien.auto.calimero.link.medium.RFSettings;
import tuwien.auto.calimero.link.medium.TPSettings;
import tuwien.auto.calimero.log.LogLevel;
import tuwien.auto.calimero.log.LogManager;
import tuwien.auto.calimero.log.LogStreamWriter;
import tuwien.auto.calimero.process.ProcessCommunicator;
import tuwien.auto.calimero.process.ProcessCommunicatorImpl;

/**
 * @author bm
 */
public class ProcComm
{
	private static final String tool = "ProcComm";
	private static final String sep = System.getProperty("line.separator");

	public ProcComm()
	{}

	public static void main(String[] args)
	{
		try {
			final Map options = new HashMap();
			if (parseArgs(args, options))
				new ProcComm().run(options);
		}
		catch (final Throwable t) {
			if (t.getMessage() != null)
				System.out.println(t.getMessage());
		}
	}

	private static boolean parseArgs(String[] args, Map options)
		throws KNXFormatException
	{
		if (args.length == 0) {
			System.out.println("A tool for KNX process communication");
			showVersion();
			System.out.println("type -help for help message");
			return false;
		}
		// add defaults
		options.put("port", new Integer(KNXnetIPConnection.IP_PORT));
		options.put("medium", TPSettings.TP1);

		int i = 0;
		for (; i < args.length; i++) {
			final String arg = args[i];
			if (isOption(arg, "-help", "-h")) {
				showUsage();
				return false;
			}
			if (isOption(arg, "-version", null)) {
				showVersion();
				return false;
			}
			if (isOption(arg, "-read", "-r")) {
				if (i + 2 >= args.length)
					break;
				options.put("read", new GroupAddress(args[++i]));
				options.put("dpt", args[++i]);
			}
			else if (isOption(arg, "-write", "-w")) {
				if (i + 3 >= args.length)
					break;
				options.put("write", new GroupAddress(args[++i]));
				options.put("dpt", args[++i]);
				options.put("value", args[++i]);
			}
			else if (isOption(arg, "-port", "-p"))
				options.put("port", Integer.decode(args[++i]));
			else if (isOption(arg, "-nat", "-n"))
				options.put("nat", null);
			else if (isOption(arg, "-medium", "-m"))
				options.put("medium", getMedium(args[++i]));
			else if (isOption(arg, "-timeout", "-t"))
				options.put("timeout", Integer.decode(args[++i]));
			else if (isOption(arg, "-routing", null))
				options.put("routing", null);
			else {
				if (!options.containsKey("host"))
					try {
						options.put("host", InetAddress.getByName(arg));
					}
					catch (final UnknownHostException e) {
						throw new IllegalArgumentException("failed to read host " + arg);
					}
				else
					throw new IllegalArgumentException("unknown option " + arg);
			}
		}
		if (!options.containsKey("host"))
			throw new IllegalArgumentException("no host specified");
		if (options.containsKey("read") == options.containsKey("write"))
			throw new IllegalArgumentException("do either read or write");
		return true;
	}

	private KNXNetworkLink createLink(Map options) throws KNXException
	{
		final InetSocketAddress host =
			new InetSocketAddress((InetAddress) options.get("host"), ((Integer) options
				.get("port")).intValue());
		final int mode =
			options.containsKey("routing") ? KNXNetworkLinkIP.ROUTER
				: KNXNetworkLinkIP.TUNNEL;
		return new KNXNetworkLinkIP(mode, null, host, options.containsKey("nat"),
			(KNXMediumSettings) options.get("medium"));
	}

	private void run(Map options) throws KNXException
	{
		LogManager.getManager().addWriter(null, new ConsoleWriter(false));
		final KNXNetworkLink lnk = createLink(options);
		ProcessCommunicator pc = null;
		try {
			pc = new ProcessCommunicatorImpl(lnk);
			if (options.containsKey("timeout"))
				pc.setResponseTimeout(((Integer) options.get("timeout")).intValue());
			final boolean read = options.containsKey("read");
			final GroupAddress main = (GroupAddress) options.get(read ? "read" : "write");
			String dptID = (String) options.get("dpt");
			if (dptID.equals("bool"))
				dptID = "1.002";
			else if (dptID.equals("string"))
				dptID = "16.001";
			else if (dptID.equals("float"))
				dptID = "9.002";
			else if (dptID.equals("ucount"))
				dptID = "5.010";
			else if (dptID.equals("angle"))
				dptID = "5.003";

			final Datapoint dp = new Datapoint(main, "", true, 0, dptID);
			if (read)
				System.out.println("read value: " + pc.read(dp));
			else {
				pc.write(dp, (String) options.get("value"));
				System.out.println("write successful");
			}
		}
		finally {
			if (pc != null)
				pc.detach();
			lnk.close();
		}
	}

	private static void showUsage()
	{
		final StringBuffer sb = new StringBuffer();
		sb.append("usage: " + tool + " [options] <host>" + sep);
		sb.append("options:" + sep);
		sb.append("  -help -h                show this help message" + sep);
		sb.append("  -version                show tool/library version and exit" + sep);
		sb.append("  -port -p <number>       UDP port on <host> (default "
			+ KNXnetIPConnection.IP_PORT + ")" + sep);
		sb.append("  -nat -n                 enable Network Address Translation" + sep);
		sb.append("  -routing                use KNX net/IP routing" + sep);
		sb.append("  -medium -m <id>         KNX medium [tp0|tp1|p110|p132|rf] "
			+ "(default tp1)" + sep);
		sb.append("  -read -r <KNX addr> <dpt>           read from group address" + sep);
		sb.append("  -write -w <KNX addr> <dpt> <value>  write to group address" + sep);
		System.out.println(sb);
	}

	private static void showVersion()
	{
		final StringBuffer sb = new StringBuffer();
		sb.append(tool + " version 0.1");
		System.out.println(sb);
		Settings.main(new String[] { "-version" });
	}

	private static boolean isOption(String arg, String longOpt, String shortOpt)
	{
		return arg.equals(longOpt) || shortOpt != null && arg.equals(shortOpt);
	}

	private static KNXMediumSettings getMedium(String id)
	{
		if (id.equals("tp0"))
			return TPSettings.TP0;
		else if (id.equals("tp1"))
			return TPSettings.TP1;
		else if (id.equals("p110"))
			return new PLSettings(false);
		else if (id.equals("p132"))
			return new PLSettings(true);
		else if (id.equals("rf"))
			return new RFSettings(null);
		else
			throw new KNXIllegalArgumentException("unknown medium");
	}

	private static final class ConsoleWriter extends LogStreamWriter
	{
		ConsoleWriter(boolean verbose)
		{
			super(verbose ? LogLevel.INFO : LogLevel.ERROR, System.out, true);
		}

		public void close()
		{}
	}
}
