/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log$
 */

package tuwien.auto.calimero.tools;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import tuwien.auto.calimero.CloseEvent;
import tuwien.auto.calimero.IndividualAddress;
import tuwien.auto.calimero.Settings;
import tuwien.auto.calimero.exception.KNXException;
import tuwien.auto.calimero.exception.KNXFormatException;
import tuwien.auto.calimero.exception.KNXIllegalArgumentException;
import tuwien.auto.calimero.knxnetip.KNXnetIPConnection;
import tuwien.auto.calimero.link.KNXNetworkLink;
import tuwien.auto.calimero.link.KNXNetworkLinkIP;
import tuwien.auto.calimero.link.medium.KNXMediumSettings;
import tuwien.auto.calimero.link.medium.PLSettings;
import tuwien.auto.calimero.link.medium.RFSettings;
import tuwien.auto.calimero.link.medium.TPSettings;
import tuwien.auto.calimero.log.LogLevel;
import tuwien.auto.calimero.log.LogManager;
import tuwien.auto.calimero.log.LogStreamWriter;
import tuwien.auto.calimero.mgmt.KnIPDeviceMgmtAdapter;
import tuwien.auto.calimero.mgmt.PropertyAdapter;
import tuwien.auto.calimero.mgmt.PropertyAdapterListener;
import tuwien.auto.calimero.mgmt.PropertyClient;
import tuwien.auto.calimero.mgmt.PropertyClient.Description;
import tuwien.auto.calimero.mgmt.PropertyClient.Property;
import tuwien.auto.calimero.mgmt.PropertyClient.PropertyKey;
import tuwien.auto.calimero.mgmt.RemotePropertyServiceAdapter;

/**
 * A console tool showing features of the {@link PropertyClient} interface for KNX
 * property access.
 * <p>
 *
 * @author bm
 */
public class PropClient
{
	private static final String tool = "PropClient";
	private static final String sep = System.getProperty("line.separator");

	// for now, local device address always 0, domain is broadcast

	private PropertyClient pc;
	private Map definitions;
	private PropertyListener l;

	/**
	 * @param args
	 */
	public static void main(String[] args)
	{
		try {
			final Map options = new HashMap();
			if (parseArgs(args, options))
				new PropClient().run(options);
		}
		catch (final Throwable t) {
			if (t.getMessage() != null)
				System.out.println(t.getMessage());
		}
	}

	public PropClient()
	{}

	private void run(Map options) throws KNXException, IOException
	{
		final Map commands = new HashMap();
		commands.put("get", new GetProperty());
		commands.put("set", new SetProperty());
		commands.put("scan", new ScanProperties());
		commands.put("desc", new GetDescription());
		commands.put("?", new Help());
		commands.put("quit", new Quit());

		try {
			pc = new PropertyClient(create(options));
			if (options.containsKey("defs"))
				PropertyClient.loadDefinitions((String) options.get("defs"));
			definitions = PropertyClient.getDefinitions();

			((Command) commands.get("?")).execute(null);
			final BufferedReader r = new BufferedReader(new InputStreamReader(System.in));
			String args[];
			while ((args = readLine(r)) != null) {
				if (args.length > 0) {
					final Command c = (Command) commands.get(args[0]);
					if (c != null && args.length > 1 && args[1].equals("?"))
						c.printHelp();
					else if (c != null) {
						try {
							if (!c.execute(args))
								break;
						}
						catch (final KNXException e) {
							if (!pc.isOpen())
								throw e;
							System.out.println(e.getMessage());
						}
						catch (final KNXIllegalArgumentException e) {
							System.out.println(e.getMessage());
						}
						catch (final NumberFormatException e) {
							System.out.println("invalid number (" + e.getMessage() + ")");
						}
					}
					else
						System.out.println("unknown command, type ? for help");
				}
			}
		}
		finally {
			if (pc != null)
				pc.close();
		}
	}

	private PropertyAdapter create(Map options) throws KNXException
	{
		LogManager.getManager().addWriter(null,
			new ConsoleWriter(options.containsKey("verbose")));
		l = new PropertyListener();
		final InetSocketAddress addr =
			new InetSocketAddress((InetAddress) options.get("host"), ((Integer) options
				.get("port")).intValue());
		if (options.containsKey("localDM")) {
			return new KnIPDeviceMgmtAdapter(null, addr, options.containsKey("nat"), l,
				options.containsKey("emulatewrite"));
		}
		final int mode =
			options.containsKey("routing") ? KNXNetworkLinkIP.ROUTER
				: KNXNetworkLinkIP.TUNNEL;
		final KNXNetworkLink lnk =
			new KNXNetworkLinkIP(mode, null, addr, options.containsKey("nat"),
				(KNXMediumSettings) options.get("medium"));

		final IndividualAddress remote = (IndividualAddress) options.get("remote");
		final byte[] authKey = (byte[]) options.get("authorize");
		if (authKey != null)
			return new RemotePropertyServiceAdapter(lnk, true, remote, authKey);
		return new RemotePropertyServiceAdapter(lnk, true, remote, options
			.containsKey("connect"));
	}

	private String[] readLine(BufferedReader r) throws IOException
	{
		System.out.print("> ");
		final String line = r.readLine();
		return line != null ? split(line) : null;
	}

	private static boolean parseArgs(String[] args, Map options)
		throws KNXFormatException
	{
		if (args.length == 0) {
			System.out.println("A tool for KNX property access");
			showVersion();
			System.out.println("type -help for help message");
			return false;
		}
		// add defaults
		options.put("port", new Integer(KNXnetIPConnection.IP_PORT));
		options.put("medium", TPSettings.TP1);

		for (int i = 0; i < args.length; i++) {
			final String arg = args[i];
			if (isOption(arg, "-help", "-h")) {
				showUsage();
				return false;
			}
			if (isOption(arg, "-version", null)) {
				showVersion();
				return false;
			}
			if (isOption(arg, "-local", "-l"))
				options.put("localDM", null);
			else if (isOption(arg, "-remote", "-r"))
				options.put("remote", new IndividualAddress(args[++i]));
			else if (isOption(arg, "-definitions", "-d"))
				options.put("defs", args[++i]);
			else if (isOption(arg, "-verbose", "-v"))
				options.put("verbose", null);
			else if (isOption(arg, "-port", "-p"))
				options.put("port", Integer.decode(args[++i]));
			else if (isOption(arg, "-nat", "-n"))
				options.put("nat", null);
			else if (isOption(arg, "-medium", "-m"))
				options.put("medium", getMedium(args[++i]));
			else if (isOption(arg, "-emulatewriteenable", "-e"))
				options.put("emulatewrite", null);
			else if (isOption(arg, "-connect", "-c"))
				options.put("connect", null);
			else if (isOption(arg, "-authorize", "-a"))
				options.put("authorize", getAuthorizeKey(args[++i]));
			else if (isOption(arg, "-routing", null))
				options.put("routing", null);
			else {
				if (!options.containsKey("host"))
					try {
						options.put("host", InetAddress.getByName(arg));
					}
					catch (final UnknownHostException e) {
						throw new IllegalArgumentException("failed to read host " + arg);
					}
				else
					throw new IllegalArgumentException("unknown option " + arg);
			}
		}
		if (!options.containsKey("localDM") && !options.containsKey("remote"))
			throw new IllegalArgumentException("no connection category specified");
		if (!options.containsKey("host"))
			throw new IllegalArgumentException("no host specified");
		return true;
	}

	private static void showUsage()
	{
		final StringBuffer sb = new StringBuffer();
		sb.append("usage: " + tool + " [options] <host>" + sep);
		sb.append("options:" + sep);
		sb.append("  -help -h                show this help message" + sep);
		sb.append("  -version                show tool/library version and exit" + sep);
		sb.append("  -verbose -v             enable verbose status output" + sep);
		sb.append("  -local -l               local device management" + sep);
		sb.append("  -remote -r <KNX addr>   remote property service" + sep);
		sb.append("  -definitions -d <file>  use property definition file" + sep);
		sb.append("  -port -p <number>       UDP port on <host> (default "
			+ KNXnetIPConnection.IP_PORT + ")" + sep);
		sb.append("  -nat -n                 enable Network Address Translation" + sep);
		sb.append(" local DM only:" + sep);
		sb.append("  -emulatewriteenable -e  check write-enable of a property" + sep);
		sb.append(" remote property service only:" + sep);
		sb.append("  -routing                use KNXnet/IP routing" + sep);
		sb.append("  -medium -m <id>         KNX medium [tp0|tp1|p110|p132|rf] "
			+ "(default tp1)" + sep);
		sb.append("  -connect -c             connection oriented mode" + sep);
		sb.append("  -authorize -a <key>     authorize key to access KNX device" + sep);
		System.out.println(sb);
	}

	private static void showVersion()
	{
		final StringBuffer sb = new StringBuffer();
		sb.append(tool + " version 0.1");
		System.out.println(sb);
		Settings.main(new String[] { "-version" });
	}

	private static boolean isOption(String arg, String longOpt, String shortOpt)
	{
		return arg.equals(longOpt) || shortOpt != null && arg.equals(shortOpt);
	}

	private static byte[] getAuthorizeKey(String key)
	{
		final long value = Long.decode(key).longValue();
		if (value < 0 || value > 0xFFFFFFFFL)
			throw new KNXIllegalArgumentException("invalid authorize key");
		return new byte[] { (byte) (value >> 24), (byte) (value >> 16),
			(byte) (value >> 8), (byte) value };
	}

	private static KNXMediumSettings getMedium(String id)
	{
		if (id.equals("tp0"))
			return TPSettings.TP0;
		else if (id.equals("tp1"))
			return TPSettings.TP1;
		else if (id.equals("p110"))
			return new PLSettings(false);
		else if (id.equals("p132"))
			return new PLSettings(true);
		else if (id.equals("rf"))
			return new RFSettings(null);
		else
			throw new KNXIllegalArgumentException("unknown medium");
	}

	private static void printHex(byte[] data, int elements)
	{
		final StringBuffer sb = new StringBuffer();
		sb.append("hex: [");
		for (int i = 0; i < data.length; ++i) {
			final int no = data[i] & 0xff;
			if (no < 0x10)
				sb.append('0');
			sb.append(Integer.toHexString(no));
			if (i > 0 && (i % (data.length / elements)) == 0)
				sb.append(", ");
		}
		sb.append("]");
		System.out.println(sb);
	}

	private static String[] split(String text)
	{
		final StringTokenizer st = new StringTokenizer(text, " \t");
		final String[] tokens = new String[st.countTokens()];
		for (int i = 0; i < tokens.length; ++i)
			tokens[i] = st.nextToken();
		return tokens;
	}

	private static int toInt(String number)
	{
		return Integer.decode(number).intValue();
	}

	private void printDescription(Description d)
	{
		final StringBuffer buf = new StringBuffer();
		buf.append(d.getPropIndex());
		buf.append(" OT " + d.getObjectType());
		buf.append(", OI " + d.getObjectIndex());
		buf.append(", PID " + d.getPID());
		if (definitions != null) {
			Property p;
			if ((p =
				(Property) definitions.get(new PropertyClient.PropertyKey(d
					.getObjectType(), d.getPID()))) != null)
				buf.append(" (" + p.getName() + ")");
			else if ((p =
				(Property) definitions.get(new PropertyClient.PropertyKey(
					PropertyKey.GLOBAL_OBJTYPE, d.getPID()))) != null)
				buf.append(" (" + p.getName() + ")");
		}
		buf.append(", PDT " + (d.getPDT() == -1 ? "-" : Integer.toString(d.getPDT())));
		buf.append(", curr. elems " + d.getCurrentElements());
		buf.append(", max. " + d.getMaxElements());
		buf.append(", r/w access " + d.getReadLevel() + "/" + d.getWriteLevel());
		buf.append(d.isWriteEnabled() ? ", w.enabled" : ", r.only");
		System.out.println(buf);
	}

	private static final class PropertyListener implements PropertyAdapterListener
	{
		boolean closed;

		public void adapterClosed(CloseEvent e)
		{
			closed = true;
		}
	}

	private static abstract class Command
	{
		String help;

		abstract boolean execute(String[] args) throws KNXException;

		final void printHelp()
		{
			System.out.println(help);
		}
	}

	private final class GetProperty extends Command
	{
		GetProperty()
		{
			help = "get object-idx pid [start-idx elements]";
		}

		boolean execute(String[] args) throws KNXException
		{
			if (args.length < 3 || args.length > 5) {
				System.out.println("sorry, wrong number of arguments");
				return true;
			}
			final int oi = toInt(args[1]);
			final int pid = toInt(args[2]);
			try {
				if (args.length == 3)
					System.out.println(pc.getProperty(oi, pid));
			}
			catch (final KNXException e) {
				printHex(pc.getProperty(oi, pid, 1, 1), 1);
			}
			try {
				if (args.length == 5)
					System.out.println(Arrays.asList(
						pc.getPropertyTranslated(oi, pid, toInt(args[3]), toInt(args[4]))
							.getAllValues()).toString());
			}
			catch (final KNXException e) {
				printHex(pc.getProperty(oi, pid, toInt(args[3]), toInt(args[4])),
					toInt(args[4]));
			}
			return true;
		}
	}

	private final class GetDescription extends Command
	{
		GetDescription()
		{
			help = "desc object-idx pid" + sep + "desc object-idx \"i\" prop-idx";
		}

		boolean execute(String[] args) throws KNXException
		{
			if (args.length == 3)
				printDescription(pc.getDescription(toInt(args[1]), toInt(args[2])));
			else if (args.length == 4 && args[2].equals("i"))
				printDescription(pc.getDescriptionByIndex(toInt(args[1]), toInt(args[3])));
			else
				System.out.println("sorry, wrong number of arguments");
			return true;
		}
	}

	private final class SetProperty extends Command
	{
		SetProperty()
		{
			help = "set object-idx pid [start-idx] string-value" + sep
				+ "set object-idx pid start-idx elements [\"0x\"|\"0\"|\"b\"]data" + sep
				+ "(use hexadecimal format for more than 8 byte data or leading zeros)";
		}

		boolean execute(String[] args) throws KNXException
		{
			if (args.length < 4 || args.length > 6) {
				System.out.println("sorry, wrong number of arguments");
				return true;
			}
			final int cnt = args.length;
			final int oi = toInt(args[1]);
			final int pid = toInt(args[2]);
			if (cnt == 4)
				pc.setProperty(oi, pid, 1, args[3]);
			else if (cnt == 5)
				pc.setProperty(oi, pid, toInt(args[3]), args[4]);
			else if (cnt == 6) {
				pc.setProperty(oi, pid, toInt(args[3]), toInt(args[4]),
					toByteArray(args[5]));
			}
			return true;
		}

		private byte[] toByteArray(String s)
		{
			// use of BigXXX equivalent is a bit awkward, for now it's sufficient..
			long l = 0;
			if (s.startsWith("0x") || s.startsWith("0X")) {
				final byte[] d = new byte[(s.length() - 1) / 2];
				int k = (s.length() & 0x01) != 0 ? 3 : 4;
				for (int i = 2; i < s.length(); i = k, k += 2)
					d[(i - 1) / 2] = (byte) Short.parseShort(s.substring(i, k), 16);
				return d;
			}
			else if (s.length() > 1 && s.startsWith("0"))
				l = Long.parseLong(s, 8);
			else if (s.startsWith("b"))
				l = Long.parseLong(s.substring(1), 2);
			else
				l = Long.parseLong(s);
			int i = 0;
			for (long test = l; test != 0; ++i)
				test /= 0x100;
			final byte[] d = new byte[i == 0 ? 1 : i];
			for (; i-- > 0; l /= 0x100)
				d[i] = (byte) (l & 0xff);
			return d;
		}
	}

	private final class ScanProperties extends Command
	{
		ScanProperties()
		{
			help = "scan [object-idx] [\"all\" for all object properties]";
		}

		boolean execute(String[] args) throws KNXException
		{
			final int cnt = args.length;
			List l = Collections.EMPTY_LIST;
			if (cnt == 1)
				l = pc.scanProperties(false);
			else if (cnt == 2) {
				if (args[1].equals("all"))
					l = pc.scanProperties(true);
				else
					l = pc.scanProperties(toInt(args[1]), false);
			}
			else if (cnt == 3 && args[2].equals("all"))
				l = pc.scanProperties(toInt(args[1]), true);
			else
				System.out.println("sorry, wrong number of arguments");
			for (final Iterator i = l.iterator(); i.hasNext();) {
				final Description d = (Description) i.next();
				printDescription(d);
			}
			return true;
		}
	}

	private final class Help extends Command
	{
		Help()
		{
			help = "show command help";
		}

		boolean execute(String[] args)
		{
			final StringBuffer buf = new StringBuffer();
			buf.append("commands: get | set | desc | scan | quit (append ? for help)"
				+ sep);
			buf.append("get - read property value(s)" + sep);
			buf.append("set - write property value(s)" + sep);
			buf.append("desc - read one property description" + sep);
			buf.append("scan - read property descriptions" + sep);
			buf.append("quit - quit this program" + sep);
			System.out.print(buf);
			return true;
		}
	}

	private final class Quit extends Command
	{
		Quit()
		{
			help = "close property client and quit";
		}

		boolean execute(String[] args)
		{
			return false;
		}
	}

	private static final class ConsoleWriter extends LogStreamWriter
	{
		ConsoleWriter(boolean verbose)
		{
			super(verbose ? LogLevel.INFO : LogLevel.ERROR, System.out, true);
		}

		public void close()
		{}
	}
}
