/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: Element.java,v $
 Revision 1.3  2007/04/13 13:37:17  bm
 added API specification comments
 code cleanup

 Revision 1.2  2007/03/01 20:35:32  bm
 minor stylistic changes

 Revision 1.1  2007/02/09 22:56:15  bm
 Adding basic functionality to read and write XML
 (no import of org.w3c.dom.* classes) in conformance
 for the port to Java ME cdc foundation profile

 */

package tuwien.auto.calimero.xml.def;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import tuwien.auto.calimero.xml.Attribute;
import tuwien.auto.calimero.xml.Element;

/**
 * Holds attributes, element information and textual content of a XML element.
 * <p>
 * 
 * @author bm
 */
public class DefaultElement implements Element
{
	private final String name;
	private final List attributes;
	private String content;
	private boolean empty;

	/**
	 * Creates a new {@link DefaultElement} <code>name</code> with <code>id</code>.
	 * <p>
	 * 
	 * @param name name of element
	 */
	public DefaultElement(String name)
	{
		this.name = name;
		attributes = new ArrayList();
	}

	/**
	 * Adds <code>att</code> to the attribute list.
	 * <p>
	 * 
	 * @param a an {@link Attribute}
	 */
	public final void addAttribute(Attribute a)
	{
		synchronized (attributes) {
			attributes.add(a);
		}
	}

	/**
	 * Returns the attribute value of the attribute <code>name</code>.
	 * <p>
	 * 
	 * @param name name of attribute
	 * @return value as String, or <code>null</code> if <code>name</code> not found
	 */
	public final String getAttribute(String name)
	{
		synchronized (attributes) {
			for (final Iterator i = attributes.iterator(); i.hasNext();) {
				final Attribute a = (Attribute) i.next();
				if (a.getName().equals(name))
					return a.getValue();
			}
		}
		return null;
	}

	/**
	 * Returns the data body of this element as text.
	 * <p>
	 * No formatting is done on the content.
	 * 
	 * @return content as String or <code>null</code> if no data available
	 */
	public final String getContent()
	{
		return content;
	}

	/**
	 * Returns if this is an empty element (no content).
	 * <p>
	 * 
	 * @return <code>true</code> if empty, <code>false</code> otherwise
	 */
	public final boolean isEmptyElement()
	{
		return empty;
	}

	/**
	 * Sets the data body of this element.
	 * <p>
	 * 
	 * @param data data of this element
	 */
	public final void setContent(String data)
	{
		content = data;
		if (content != null)
			empty = false;
	}

	/**
	 * Determines if this element has a body or not.
	 * <p>
	 * An empty element is of the form
	 * <code>{@literal &lt;element-tag [attributes] /&gt;}</code>.
	 * 
	 * @param empty <code>true</code> if empty, <code>false</code> otherwise
	 */
	public final void setEmptyElement(boolean empty)
	{
		this.empty = empty;
		if (empty)
			content = null;
	}

	/**
	 * Returns element name.
	 * <p>
	 * 
	 * @return element name
	 */
	public final String getName()
	{
		return name;
	}
}
