/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: XmlReader.java,v $
 Revision 1.2  2007/03/01 20:35:32  bm
 minor stylistic changes

 Revision 1.1  2007/02/09 22:56:15  bm
 Adding basic functionality to read and write XML
 (no import of org.w3c.dom.* classes) in conformance
 for the port to Java ME cdc foundation profile

 */

package tuwien.auto.calimero.xml.def;

import java.io.IOException;
import java.io.Reader;
import java.util.Stack;

import tuwien.auto.calimero.xml.Attribute;
import tuwien.auto.calimero.xml.Element;
import tuwien.auto.calimero.xml.KNXMLException;
import tuwien.auto.calimero.xml.XMLReader;

/**
 * Xml reader for for xml resources, implemented as pull parser, non-validating, resolves
 * predefined XML entities.
 * <p>
 * 
 * @author bm
 * @see DefaultElement
 */
public class DefaultXMLReader implements XMLReader
{
	private Reader r;
	private boolean closeReader;
	private Element elem;
	private final Stack openElems = new Stack();
	private int pos;
	private int line;

	/**
	 * 
	 */
	public DefaultXMLReader()
	{}

	/**
	 * Creates a new {@link DefaultXMLReader} with input <code>r</code>.
	 * <p>
	 * The {@link Reader} should already be buffered or wrapped with a buffered reader, if
	 * necessary (e.g. when reading from a file). Note that <code>r</code> will not be
	 * closed after usage.
	 * 
	 * @param r a {@link Reader} for input
	 */
	public DefaultXMLReader(Reader r, boolean close)
	{
		reset();
		setInput(r, close);
	}

	private String read(char delimiter) throws KNXMLException
	{
		final StringBuffer buf = new StringBuffer(50);
		try {
			for (int c = 0; (c = r.read()) != -1 && c != delimiter;) {
				if (Character.getType((char) c) == Character.LINE_SEPARATOR)
					++line;
				buf.append((char) c);
			}
		}
		catch (final IOException e) {
			throw new KNXMLException(e.getMessage(), buf.toString(), line);
		}
		return buf.toString();
	}

	/**
	 * Reads input until end of element <code>e</code>.
	 * <p>
	 * All gathered information will be stored into <code>e</code>.
	 * 
	 * @param e the {@link DefaultElement} which should be completed
	 * @throws KNXMLException if end of element is not reachable in xml stream, or end of
	 *         parent element of <code>e</code> was reached
	 */
	public void complete(Element e) throws KNXMLException
	{
		if (e.isEmptyElement())
			return;
		final int index = openElems.lastIndexOf(e.getName());
		if (index == -1)
			throw new KNXMLException("element tag not read before", e.getName(), line);
		String end = null;
		final StringBuffer content = new StringBuffer(50);
		while (canRead()) {
			// read text content
			content.append(read('<'));
			// read a possible end tag
			end = read('>');
			if (skipComment(end))
				;
			else if (end.length() > 0 && end.charAt(0) == '/') {
				// actually, no white space is allowed between '/' and tag name
				// but there might be some between end of name and '>'
				final String tag = end.substring(1).trim();
				// got end tag?
				if (!tag.equals(openElems.peek()))
					throw new KNXMLException("element end tag does not match start tag",
						tag, line);
				openElems.pop();
				if (tag.equals(e.getName())) {
					e.setContent(References.substitute(content.toString(), false));
					pos = END_TAG;
					return;
				}
			}
			else if (end.length() > 0) {
				if (end.charAt(end.length() - 1) != '/') {
					final String tag = splitOnSpace(end);
					if (tag.length() > 0)
						openElems.push(tag);
				}
			}
			content.append('<').append(end).append('>');
		}
		// e.setContent(References.substitute(content.toString(), false));
		throw new KNXMLException("end of XML input with elements left open", end, line);
	}

	/**
	 * Returns the current element read by the last call of {@link #read()}.
	 * <p>
	 * 
	 * @return current {@link DefaultElement} or <code>null</code> if no element is
	 *         available
	 */
	public final Element getCurrent()
	{
		return elem;
	}

	public final int getLineNumber()
	{
		return line;
	}

	public final int getPosition()
	{
		return pos;
	}

	public void setInput(Reader r, boolean close)
	{
		if (this.r != null)
			reset();
		this.r = r;
		closeReader = close;
	}

	private void reset()
	{
		elem = null;
		openElems.clear();
		pos = START_DOC;
		line = 0;
	}

	/**
	 * Reads in the next recognized element tag.
	 * <p>
	 * If in open elements, they will be discarded.
	 * 
	 * @return the next {@link DefaultElement} or <code>null</code> if no more known
	 *         elements
	 * @throws KNXMLException on read error or malformed XML structure
	 */
	public int read() throws KNXMLException
	{
		while (canRead()) {
			// init line counter on first read
			if (line == 0)
				line = 1;

			read('<');
			String str = read('>');
			// if no comment and no closing tag
			if (str.length() == 0 || skipComment(str) || skipInstruction(str))
				continue;
			str = str.trim();
			// extract element name
			final String name = splitOnSpace(str);
			if (name.charAt(0) == '/') {
				if (!name.substring(1).equals(openElems.peek()))
					throw new KNXMLException("element end tag does not match start tag",
						name.substring(1), line);
				elem = new DefaultElement((String) openElems.pop());
				pos = END_TAG;
				return pos;
			}
			elem = new DefaultElement(name);
			extractAttributes(str.substring(name.length()));
			if (!elem.isEmptyElement())
				openElems.push(name);
			pos = START_TAG;
			return pos;
		}
		if (!openElems.empty())
			throw new KNXMLException("end of XML input with elements left open");
		pos = END_DOC;
		return pos;
	}

	public Element readComplete() throws KNXMLException
	{
		if (getPosition() != START_TAG)
			read();
		complete(elem);
		return elem;
	}

	private boolean canRead()
	{
		try {
			return r.ready();
		}
		catch (final IOException e) {
			return false;
		}
	}

	// adds all available attributes to current element
	private void extractAttributes(String attributes)
	{
		String s = attributes.trim();
		if (s.length() > 0 && s.charAt(s.length() - 1) == '/')
			elem.setEmptyElement(true);

		while (s.length() != 0) {
			final int equal = s.indexOf('=');
			if (equal == -1 || equal == s.length() - 1)
				break;
			// cut off attribute name
			final String att = s.substring(0, equal).trim();
			s = s.substring(equal + 1).trim();
			if (s.length() == 0)
				break;
			final boolean quote = s.charAt(0) == '\'' || s.charAt(0) == '\"';
			if (att.length() > 0 && quote) {
				String value = s.substring(1);
				final int end = value.indexOf(s.charAt(0));
				if (end >= 0)
					value = value.substring(0, end);
				elem
					.addAttribute(new Attribute(att, References.substitute(value, false)));
			}
			final int i = s.indexOf(quote ? s.charAt(0) : ' ', 1);
			s = i == -1 ? "" : s.substring(i + 1);
		}
	}

	// checks if '<' marks begin of a comment, and if so skips over it
	private boolean skipComment(String s) throws KNXMLException
	{
		if (s.startsWith("!--")) {
			String comment = s;
			while (canRead() && !comment.endsWith("--"))
				comment = read('>');
			return true;
		}
		return false;
	}

	// skips input to end of element
	// tag is the content of the whole element tag
//	private boolean skipElement(String tag)
//	{
//		// is this an empty element
//		if (tag.length() == 0 || tag.charAt(tag.length() - 1) == '/')
//			return false;
//		// is this a processing instruction
//		if (skipInstruction(tag))
//			return true;
//		final String name = splitOnSpace(tag);
//		String end;
//		while (canRead()) {
//			read(r, '<');
//			end = read(r, '>');
//			if (!skipComment(end) && end.length() != 0 && end.charAt(0) == '/') {
//				end = end.substring(1).trim();
//				// got end tag?
//				if (end.equals(name))
//					return true;
//			}
//		}
//		return false;
//	}

	private boolean skipInstruction(String tag)
	{
		// is this a processing instruction
		if (tag.charAt(0) == '?' && tag.charAt(tag.length() - 1) == '?')
			return true;
		return false;
	}

	private String splitOnSpace(String s)
	{
		for (int i = 0; i < s.length(); ++i)
			if (Character.isSpaceChar(s.charAt(i)))
				return s.substring(0, i);
		return s;
	}

	/* (non-Javadoc)
	 * @see tuwien.auto.calimero.xml.XMLReader#close()
	 */
	public void close() throws KNXMLException
	{
		if (closeReader)
			try {
				r.close();
			}
			catch (final IOException e) {
				throw new KNXMLException(e.getMessage());
			}
	}
}
