/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: XmlWriter.java,v $
 Revision 1.2  2007/03/01 20:35:32  bm
 minor stylistic changes

 Revision 1.1  2007/02/09 22:56:15  bm
 Adding basic functionality to read and write XML
 (no import of org.w3c.dom.* classes) in conformance
 for the port to Java ME cdc foundation profile

 */

package tuwien.auto.calimero.xml.def;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.Iterator;
import java.util.List;
import java.util.Stack;

import tuwien.auto.calimero.exception.KNXIllegalStateException;
import tuwien.auto.calimero.xml.Attribute;
import tuwien.auto.calimero.xml.KNXMLException;
import tuwien.auto.calimero.xml.XMLWriter;

/**
 * A basic xml writer.
 * <p>
 */
public class DefaultXMLWriter implements XMLWriter
{
	// default indentation
	private static final int indentWidth = 4;
	private static final String quote = "\"";
	private BufferedWriter w;
	private boolean closeWriter;
	// xml layout stack
	private Stack layout;
	// current layout indentation
	private int indent;
	private boolean newTag;

	/**
	 * 
	 */
	public DefaultXMLWriter()
	{}

	/**
	 * Creates a new {@link DefaultXMLWriter} with output <code>w</code>.
	 * <p>
	 * The writer will be buffered by this {@link DefaultXMLWriter}. Note that
	 * <code>w</code> will not be closed after usage.
	 * 
	 * @param w the output {@link Writer}
	 */
	public DefaultXMLWriter(Writer w, boolean close)
	{
		setOutput(w, close);
	};

	private void reset()
	{
		layout = new Stack();
		indent = 0;
		newTag = false;
	}

	/**
	 * Closes the current most inner open element node.
	 * 
	 */
	public void endElement() throws KNXMLException
	{
		if (layout.empty())
			throw new KNXIllegalStateException("no elements to end");
		try {
			((Tag) layout.pop()).endTag();
			if (layout.empty())
				w.flush();
		}
		catch (final IOException e) {
			throw new KNXMLException(e.getMessage());
		}
	}

	/**
	 * Closes all open elements and writes all buffered data to output.
	 * <p>
	 * Call this when finished with writing elements.
	 * 
	 */
	public void endAllElements() throws KNXMLException
	{
		try {
			while (!layout.empty())
				((Tag) layout.pop()).endTag();
			w.flush();
		}
		catch (final IOException e) {
			throw new KNXMLException(e.getMessage());
		}
	}

	public void setOutput(Writer w, boolean close)
	{
		reset();
		this.w = new BufferedWriter(w);
		closeWriter = close;
	}

	public void writeDeclaration(boolean standalone, String encoding)
		throws KNXMLException
	{
		try {
			w.write("<?xml version=" + quote + "1.0" + quote);
			w.write(" standalone=" + quote + (standalone ? "yes" : "no") + quote);
			if (encoding != null && encoding.length() > 0)
				w.write(" encoding=" + quote + encoding + quote);
			w.write("?>");
			w.newLine();
		}
		catch (final IOException e) {
			throw new KNXMLException(e.getMessage());
		}
	}

	/**
	 * Writes a new element.
	 * <p>
	 * Note, if an empty element is written (which has the format &lt;element-name
	 * attribute-list /&gt; ), endElement() must not be called for this element, since it
	 * is ended immediately anyway and not left open.
	 * 
	 * @param name element name of node
	 * @param att list of attributes in start tag, use empty list or <code>null</code>
	 *        for no attributes
	 * @param content plain text in element node (may be null)
	 */
	public void writeElement(String name, List att, String content) throws KNXMLException
	{
		try {
			final Tag tag = new Tag(name, att, content, false);
			layout.push(tag);
		}
		catch (final IOException e) {
			throw new KNXMLException(e.getMessage());
		}
	}

	public void writeEmptyElement(String name, List att) throws KNXMLException
	{
		try {
			final Tag tag = new Tag(name, att, null, true);
			tag.endTag();
		}
		catch (final IOException e) {
			throw new KNXMLException(e.getMessage());
		}
	}

	public void writeComment(String comment) throws KNXMLException
	{
		try {
			w.write("<!--");
			w.write(comment);
			w.write("-->");
			w.newLine();
		}
		catch (final IOException e) {
			throw new KNXMLException(e.getMessage());
		}
	}

	private BufferedWriter indent() throws IOException
	{
		for (int i = 0; i < indent; ++i)
			w.write(' ');
		return w;
	}

	// helper class for tag writing
	private final class Tag
	{
		private static final String lt = "<";
		private static final String gt = ">";
		private static final String equal = "=";
		private static final String slash = "/";
		private static final String space = " ";
		private String name;

		Tag(String name, List att, String cnt, boolean empty) throws IOException
		{
			if (newTag)
				w.newLine();
			indent().write(lt + name);
			if (att != null) {
				for (final Iterator i = att.iterator(); i.hasNext();) {
					final Attribute a = (Attribute) i.next();
					w.write(space + a.getName() + equal + quote
						+ References.substitute(a.getValue(), true) + quote);
				}
			}
			if (empty) {
				w.write(space + slash + gt);
				w.newLine();
				newTag = false;
			}
			else {
				newTag = true;
				if (cnt == null || cnt.length() == 0)
					w.write(gt);
				else
					w.write(gt + References.substitute(cnt, true));
				this.name = name;
				indent += indentWidth;
			}
		}

		void endTag() throws IOException
		{
			if (name != null) {
				indent -= indentWidth;
				if (!newTag)
					indent();
				newTag = false;
				w.write(lt + slash + name + gt);
				w.newLine();
			}
		}
	}

	/* (non-Javadoc)
	 * @see tuwien.auto.calimero.xml.XMLWriter#close()
	 */
	public void close() throws KNXMLException
	{
		if (!layout.isEmpty())
			endAllElements();
		if (closeWriter)
			try {
				w.close();
			}
			catch (final IOException e) {
				throw new KNXMLException(e.getMessage());
			}
	}
}
