/*
  This file is part of Calimero, a pure Java library
  for KNX network access and management.
  http://calimero.sourceforge.net

  Calimero was developed at the Vienna University of Technology,
  Automation Systems Group.
  http://www.auto.tuwien.ac.at/a-lab

  Copyright 2007 Boris Malinowsky

  This version is made available for evaluation purposes only.
  Any commercial use is prohibited. This code is preliminary and
  subject to change at any time without further notice.
  All rights reserved.
*/

/*
 $Log: References.java,v $
 Revision 1.3  2007/04/13 13:37:17  bm
 added API specification comments
 code cleanup

 Revision 1.2  2007/03/01 20:35:32  bm
 minor stylistic changes

 Revision 1.1  2007/02/09 22:56:15  bm
 Adding basic functionality to read and write XML
 (no import of org.w3c.dom.* classes) in conformance
 for the port to Java ME cdc foundation profile

 */

package tuwien.auto.calimero.xml.def;

import java.util.HashMap;
import java.util.Map;

/**
 * Substitutes predefined XML entities.
 * <p>
 * Does the replacement to and from references in strings.<br>
 * Predefined entities exist for
 * {@literal '&', '<', '>', '"' (quote) and ''' (single quote)}.
 * 
 * @author bm
 */
final class References
{
	// entities predefined in XML:
	// char | entity ref | num ref | hex ref
	// &      &amp;        &#38;     &#x26;
	// <      &lt;         &#60;     &#x3C;
	// >      &gt;         &#62;     &#x3E;
	// "      &quot;       &#34;     &#x22;
	// '      &apos;       &#39;     &#x27;

	// substitution table
	private static final String[][] entityTable = {
		{ "&amp;", "&#38;", "&#x26;", "&" },
		{ "&lt;", "&#60;", "&#x3C;", "<" },
		{ "&gt;", "&#62;", "&#x3E;", ">" },
		{ "&quot;", "&#34;", "&#x22;", "\"" },
		{ "&apos;", "&#39;", "&#x27;", "'" },
	};
	// references to entity mapping
	private static final Map map;
	
	static
	{
		map = new HashMap(25);
		for (int i = 0; i < 5; ++i) {
			for (int k = 0; k < 3; ++k)
				map.put(entityTable[i][k], entityTable[i][3]);
		}
	}
	
	private References()
	{}

	static String replaceFromRef(String text)
	{
		final StringBuffer mod = new StringBuffer(text.length());
		int pos = 0;
		int oldpos = 0;
		while ((pos = text.indexOf('&', pos)) != -1) {
			final int end = text.indexOf(';', pos + 1);
			if (end == -1)
				break;
			final String entity = (String) map.get(text.substring(pos, end + 1));
			if (entity != null) {
				mod.append(text.substring(oldpos, pos));
				mod.append(entity);
				oldpos = end + 1;
				pos = end;
			}
			++pos;
		}
		if (mod.length() != 0)
			return mod.append(text.substring(oldpos)).toString();
		return text;
	}

	static String replaceWithRef(String text)
	{
		final StringBuffer mod = new StringBuffer((int)(1.5f * text.length()));
		for (int i = 0; i < text.length(); ++i) {
			final char c = text.charAt(i);
			if (c == '&')
				mod.append("&amp;");
			else if (c == '<')
				mod.append("&lt;");
			else if (c == '>')
				mod.append("&gt;");
			else if (c == '"')
				mod.append("&quot;");
			else if (c == '\'')
				mod.append("&apos;");
			else
				mod.append(c);
		}
		return mod.toString();
	}
	
	// wrapper for reference substitution
	static String substitute(String text, boolean toReference)
	{
		return toReference ? replaceWithRef(text) : replaceFromRef(text);
	}
}
